/* Obtain associations from integrated datasets.
   For each gene, identify the input aberrations that cannot be explained away by other associations.
   Extract pairwise associations from pre-computed lists.
   Then apply three-way tests to filter out weaker associations.
   Difference from compute_integrated_associations.c: (1)If candidate aberrations contain mutation, CNV, DNA methylation, then perform model selection on these aberrations and retain the selected ones.  Add mir and TF on top of the selected models, (2)Use different thresholds for (CNV,mutation,DNA methylation) group and (mir,TF) group.
   Difference from compute_integrated_associations2.c: (1)Allow inter segment CNVs on the same chromosome as the gene.  (2)Build a look up table of the TFs associated with each gene in the front.  (3)Intra segment CNVs and mutations have higher priorities than inter segment CNVs and methylations, since the latter categories have more associations.  (4)Avoid associating the TF expressions themselves.  (5)Separate TF and mir correlation thresholds.  (6)Separate inter segment CNV and methylation correlation thresholds.  (7)Separate mutation correlation threshold.
   Difference from compute_integrated_associations3.c: a cluster version.
   Difference from compute_integrated_associations3_cluster.c: (1)Correct a bug in compute_integrated_associations3: do not add existing candidates on the list. (2)Only run the previously crashed jobs.
   Difference from compute_integrated_associations3_cluster.c and compute_integrated_associations3_cluster_correct1.c: (1)Sequentially execute 3 levels of associations.  Level 1: local aberrations, intra segment CNVs.  Level 2: inter segment CNVs, mutations, methylations.  Level 3: mirs, TFs that are explained by level 1 and level 2 modes.  Level 4: mirs, TFs that are not explained by level 1 and level 2 models.  (2)Simplify the methods of filtering candidate drivers.  (3)DNA methylation and mir drivers are individual genes or mirs instead of clusters.
   Difference from compute_integrated_associations3_cluster.c and compute_integrated_associations3_cluster_correct1.c: (1)Sequentially execute 3 levels of associations.  Level 1: local aberrations, intra segment CNVs.  Level 2: inter segment CNVs, mutations, methylations.  Level 3: mirs that are not explained by level 1 and level 2 models.  (2)Simplify the methods of filtering candidate drivers.  (3)DNA methylation and mir drivers are individual genes or mirs instead of clusters.
   Difference from compute_integrated_associations4_cluster.c: include SNP data in associations.
   Aberrations that explain the mRNA or mir expressions:
   Level 1: local mutations and methylations, intra segment CNVs, local SNPs.  
   Level 2: inter segment CNVs, mutations, methylations, SNPs.
   Level 3: mirs that are not explained by level 1 and level 2 models.
   Level 4: TFs (regulators in general) that are not explained by level 1, 2 and 3 models.
   Difference from compute_integrated_associations5_cluster.c: 
   (1)Correct an error in evaluating likelihoods when the lambda coefficients <= 0.
   (2)Follow a strictly priority list of model incorporation: cis-acting CNV > local mutation > local negative methylation > positive/negative local SNP > positive trans-acting CNV > positive/negative nonlocal mutation > negative nonlocal methylation > positive/negative nonlocal SNP > negative mir > positive nonlocal TF > ( negative trans-acting CNV > positive nonlocal methylation > positive mir > negative nonlocal TF ) > combinatorial effects of chromosome CNVs and mutations.
   (3)For inter segment CNV associations, only consider the chromosomes that accommodate sufficient number of associations.  This includes chromosomes 7, 9, 10, 11, 14, 19, 20.
   (4)Combine associations with segments to associations with chromosomes.
   (5)A valid association with DNA methylation requires (1)the methylation state and the gene are negatively associated, (2)the methylated genes are associated with sufficient number of genes, (3)among the associations with genes in a cluster, report the representative one.
   (6)A valid association with mir expression requires (1)the mir expression is not explained by segment CNVs, mutations, DNA methylations and SNPs, (2)the mir is associated with sufficient number of genes, (3)among the associations with mirs in a cluster, report the representative one.
   (7)A valid association with TF expression requires (1)the TF expression is not explained by segment CNVs, mutations, DNA methylations and SNPs, (2)the TF is associated with sufficient number of genes, (3)among the associations with TFs in a cluster, report the representative one. 
   (8)Include combinatorial effects of segment CNVs and mutations.  Consider combinatorial effects when the marginal effects are small.
   Level 1: local mutations and methylations, intra segment CNVs, local SNPs.
   Level 2: positive inter segment CNVs, mutations, negative methylations, SNPs.
   Level 3: negative mirs that are not explained by level 1 and level 2 models.
   Level 4: negative inter segment CNVs, positive methylations.
   Level 5: positive mirs that are not explained by level 1 and level 4 models.
   Level 6: pairwise combinatorial interactions of segment CNVs and mutations.
   Level 7: positive TFs (regulators in general) that are not explained by lower level models.
   Level 8: negative TFs (regulators in general) that are not explained by lower level models.
   Difference from compute_integrated_associations6_cluster.c: 
   Alter the definition of levels.
   Level 1: local mutations and methylations, intra segment CNVs, local SNPs.
   Level 2: positive inter segment CNVs, mutations, negative methylations, SNPs.
   Level 3: negative inter segment CNVs and positive methylations not explained by level 1-2 models.
   Level 4: negative mirs not explained by level 1-3 models.
   Level 5: positive mirs not explained by level 1-4 models.
   Level 6: pairwise combinatorial interactions of segment CNVs and mutations.      
   Difference from compute_integrated_associations7_cluster.c: 
   Alter the definition of levels.
   Level 1: local mutations and methylations, intra segment CNVs, local SNPs.
   Level 2: positive inter segment CNVs, mutations, negative methylations, SNPs.
   Level 3: negative inter segment CNVs not explained by level 1-2 models. 
   Level 4: negative mirs not explained by level 1-3 models.
   Level 5: pairwise combinatorial interactions of segment CNVs and mutations.  
*/


# include <stdio.h>
# include <stdlib.h>
# include <string.h>
# include <math.h>
# include <time.h>
# include "models.h"
# include "motif.h"
# define MAXLINEWIDTH 1000000
# define NAMELENGTH 30
# define LONGNAMELENGTH 100
# define max(i,j) (i>j ? i : j)
# define min(i,j) (i<j ? i : j)
# define argmax(i,j) (j>i ? 2 : 1)

extern int pair_cmp(struct pair *p1, struct pair *p2);
extern int pair_cmp_inv(struct pair *p1, struct pair *p2);
extern char * getitemval2(char *s, int ind, int *length);
extern char *i2a(long num, int *length);
extern double atof2(char *s, int length);
extern char * getitemval3(char *s, int ind, int *length, char sepch);
extern double * ml_pairwise_model(int nsamples, double *x, double *y, int mode);
extern double * ml_pairwise_model2(int nsamples, double *x, double *y, int mode);
extern double new_missing_corrcoef(int nsamples, double *x, double *y);
extern double * additional_power(int nsamples, int noldvars, double *olddata, int *oldnstates, int *oldsigns, double *newvardata, int newnstates, int newsign);
extern double * additional_power_corrected(int nsamples, int noldvars, double *olddata, int *oldnstates, int *oldsigns, double *newvardata, int newnstates, int newsign);
extern struct matdata * load_highRes_data(char *filename, int *nprobespnt, int *nsamplespnt);
extern struct matdata * load_highRes_data_range(char *filename, int startindex, int endindex, int *nprobespnt, int *nsamplespnt);
extern double * load_matrix(char *filename, int *nrowspnt, int *ncolspnt);
extern double mat_det(int ndim, double *A);
extern int inlist(int Nv, int *v, int target);
extern int inlistind(int Nv, int *v, int target);
extern struct component * find_conn_comps(int nnodes, int *G, int *nconncomps_pnt);
extern struct matdata * load_highRes_data2(char *filename, int nsubprobes, int *probeinds, int *nsamplespnt);
extern double * ml_pairwise_model_corrected(int nsamples, double *x, double *y, int mode, int nstates);
extern double * ml_pairwise_interaction_model_corrected(int nsamples, double *inputvec1, double *inputvec2, double *outputvec, int nstates1, int nstates2, int *direction);
extern struct integerset * load_cluster(char *filename, int *nclus_int);


int nintervals=100;
int nmaxiterations=1000, nperm=10000;
//int nmaxiterations=1000, nperm=100;
double NAval=-1e6, epsilon=0.001, infval=1e20;
double high_thre=0.6, low_thre=0.4;
int ndigits=10;

struct intset {
  int nmembers;
  int *members1;
  int *members2;
};

int main(int argc, char *argv[])
{
  // *(argv+1) -- mode=0: fit the mRNA expressions, mode=1: fit the microRNA expressions.
  // *(argv+2) -- mRNA expression data.
  // *(argv+3) -- microRNA expression data.
  // *(argv+4) -- segment CNV data.
  // *(argv+5) -- mutation data. 
  // *(argv+6) -- DNA methylation data of individual genes.
  // *(argv+7) -- TF expression data.
  // *(argv+8) -- complete SNP data.
  // *(argv+9) -- intra-segment CNV association list.
  // *(argv+10) -- inter-segment CNV association list.
  // *(argv+11) -- mutation association list.
  // *(argv+12) -- DNA methylation association list.
  // *(argv+13) -- microRNA expression association list.
  // *(argv+14) -- TF expression association list.
  // *(argv+15) -- selected SNP-mRNA association list.
  // *(argv+16) -- selected SNP-mir association list.
  // *(argv+17) -- clusters of DNA methylation data.
  // *(argv+18) -- clusters of mir expression data.
  // *(argv+19) -- clusters of TF expression data.
  // *(argv+20) -- selected chromosomes for external segment associations.
  // *(argv+21) -- selected methylated genes to be considered as drivers.
  // *(argv+22) -- selected mirs to be consired as drivers.
  // *(argv+23) -- selected TFs to be considered as drivers.
  // *(argv+24) -- association level (1-5).  
  // *(argv+25) -- input file of llratio and pvalue and correlation thresholds.       
  // *(argv+26) -- start gene index.
  // *(argv+27) -- increment size
  // *(argv+28) -- input integrated association list of the same type of data as the output list (mRNA or mir).
  // *(argv+29) -- input integrated association list of the different type of data from the output list (mir or mRNA).  
  // *(argv+30) -- output integrated association list.

  int i, j, k, kk, l, m, n, p, mode, level, ngenes, nsamples, nsegments, nmirs, nmethlgenes, nmuts, nTFs, nsnps, length, noutputs;
  double val, val2, *outdata, *mrnadata, *mirdata, *segcnvdata, *muttable, *methldata, *TFdata, *snptable;
  char *s, *item, ch, *samples;
  struct genemirseginfo *genes, *segments, *mutgenes, *methlgenes, *mirs, *TFs, *snps;
  int nintraseglist, ninterseglist, nmutlist, nmethllist, nmirlist, nTFlist, nsnplist, *snpinds, nmirsnplist;
  int nassociations, ncols, geneind, curind, incsize;
  long startindex, endindex;
  double *intraseglist, *interseglist, *mutlist, *methllist, *mirlist, *TFlist, *snplist, *mirsnplist;
  double *llthres, *pvalthres, *corrthres;  
  struct intassociation *associations, *preassociations, *preassociations2, *newassociations;
  struct matdata *template;
  FILE *fp, *fout;
  time_t *tp;  
  struct intset *geneTFs;
  char *filename, *prefix;
  int *mut2gene, *methl2gene, *TF2gene, wrote=0, *snplistinds, *mirsnplistinds, *gene2TF;
  struct pair *pairs;  
  int nmethlclus, nmirclus, nTFclus, interaction, funcsign, *mutvalid;
  int nselchrs, *selchrs, nselmethlgenes, *selmethlgenes, nselmirs, *selmirs, nselTFs, *selTFs;
  struct integerset *methlclus, *mirclus, *TFclus;
  char *str, *str2;
  
  s=malloc(sizeof(char)*MAXLINEWIDTH); 

  // Load the molecular aberration and expression data.
  mode=atoi(*(argv+1));
     
  // mRNA expression data.
  template=load_highRes_data(*(argv+2),&ngenes,&nsamples);
  genes=template->probes;
  samples=template->samples;
  mrnadata=template->data;

  // microRNA expression data.
  template=load_highRes_data(*(argv+3),&nmirs,&nsamples);
  mirs=template->probes;
  samples=template->samples;
  mirdata=template->data;

  if (mode==0)
    {
      noutputs=ngenes;
      outdata=mrnadata;
    }
  else
    {
      noutputs=nmirs;
      outdata=mirdata;
    }

  // segment CNV data.
  template=load_highRes_data(*(argv+4),&nsegments,&nsamples);
  segments=template->probes;
  segcnvdata=template->data;

  // mutation data.
  template=load_highRes_data(*(argv+5),&nmuts,&nsamples);
  mutgenes=template->probes;
  muttable=template->data;
 
  // Convert the mutation data into binary numbers.
  for (n=0; n<=(nmuts-1); n++)
    {
      for (m=0; m<=(nsamples-1); m++)
	{
	  val=*(muttable+n*nsamples+m);
	  if (val>NAval)
	    {
	      if (val>0)
		*(muttable+n*nsamples+m)=1;
	      else
		*(muttable+n*nsamples+m)=0;
	    }
	}
    }

  // Find valid mutated genes.  A valid gene contains at least 10 mutations.
  mutvalid=malloc(sizeof(int)*nmuts);
  for (n=0; n<=(nmuts-1); n++)
    {
      k=0;
      for (i=0; i<=(nsamples-1); i++)
	if (*(muttable+n*nsamples+i)==1)
	  k++;
      if (k>=10)
	*(mutvalid+n)=1;
      else
	*(mutvalid+n)=0;
    }
  

  // DNA methylation data.
  template=load_highRes_data(*(argv+6),&nmethlgenes,&nsamples);
  methlgenes=template->probes;
  methldata=template->data;

  // TF expression data.
  template=load_highRes_data(*(argv+7),&nTFs,&nsamples);
  TFs=template->probes;
  TFdata=template->data;

  // Load pairwise association list data.
  intraseglist=load_matrix(*(argv+9),&nintraseglist,&ncols);

  interseglist=load_matrix(*(argv+10),&ninterseglist,&ncols);

  mutlist=load_matrix(*(argv+11),&nmutlist,&ncols);

  methllist=load_matrix(*(argv+12),&nmethllist,&ncols);

  // Load the mir association list for mRNA data only.
  if (mode==0)
    {
      mirlist=load_matrix(*(argv+13),&nmirlist,&ncols);
    }
  else
    nmirlist=0;

  TFlist=load_matrix(*(argv+14),&nTFlist,&ncols);
     
  snplist=load_matrix(*(argv+15),&nsnplist,&ncols);

  
  // For mode=1, load the SNP-mir association list.
  // For level 1 models, load selected SNP-mir associations.
  // For level 2 models, load the entire SNP-mir association list.

  if (mode==1)
    mirsnplist=load_matrix(*(argv+16),&nmirsnplist,&ncols);
  else
    nmirsnplist=0;

  // Load the SNP data of selected probes.
  nsnps=0; snpinds=malloc(sizeof(int));
  for (n=0; n<=(nsnplist-1); n++)
    {
      k=(int)(*(snplist+n*ncols+0)-1);
      if (inlist(nsnps,snpinds,k)==0)
	{
	  nsnps++; snpinds=realloc(snpinds,sizeof(int)*nsnps);
	  *(snpinds+nsnps-1)=k;
	}
    }
  for (n=0; n<=(nmirsnplist-1); n++)
    {
      k=(int)(*(mirsnplist+n*ncols+0)-1);
      if (inlist(nsnps,snpinds,k)==0)
	{
	  nsnps++; snpinds=realloc(snpinds,sizeof(int)*nsnps);
	  *(snpinds+nsnps-1)=k;
	}
    }


  // Sort SNP probe indices.
  pairs=malloc(sizeof(struct pair)*nsnps);
  for (n=0; n<=(nsnps-1); n++)
    {
      (pairs+n)->i=n; (pairs+n)->j=*(snpinds+n);
    }
  qsort(pairs,nsnps,sizeof(struct pair),(* pair_cmp_inv));
  for (n=0; n<=(nsnps-1); n++)
    *(snpinds+n)=(pairs+n)->j;
 
  // Load the SNP data of selected probes.
  template=load_highRes_data2(*(argv+8),nsnps,snpinds,&nsamples);
  snps=template->probes;
  snptable=template->data;

  // Convert the SNP data into cdf values 0, 0.5 or 1.
  // 0: homozygous of minor allele.
  // 1: homozygous of major allele.
  // 0.5: heterozygous.
  for (n=0; n<=(nsnps-1); n++)
    {
      for (m=0; m<=(nsamples-1); m++)
	{
	  val=*(snptable+n*nsamples+m);
	  if (val>NAval)
	    {
	      if (val<=0)
		*(snptable+n*nsamples+m)=0;
	      else if (val>1)
		*(snptable+n*nsamples+m)=1;
	      else
		*(snptable+n*nsamples+m)=0.5;
	    }
	}
    }

  // Convert the SNP probe indices on snplist.
  // Retain the original SNP probe indices in snplistinds.
  snplistinds=malloc(sizeof(int)*nsnplist);
  for (n=0; n<=(nsnplist-1); n++)
    {
      k=(int)(*(snplist+n*ncols+0)-1);
      l=inlistind(nsnps,snpinds,k);
      *(snplist+n*ncols+0)=l+1;
      *(snplistinds+n)=k;
    }
 
  mirsnplistinds=malloc(sizeof(int)*nmirsnplist);
  for (n=0; n<=(nmirsnplist-1); n++)
    {
      k=(int)(*(mirsnplist+n*ncols+0)-1);
      l=inlistind(nsnps,snpinds,k);
      *(mirsnplist+n*ncols+0)=l+1;
      *(mirsnplistinds+n)=k;
    }
  

  // Load methylation cluster data.
  methlclus=load_cluster(*(argv+17),&nmethlclus);

  // Load mir cluster data.
  mirclus=load_cluster(*(argv+18),&nmirclus);

  // Load TF cluster data.
  TFclus=load_cluster(*(argv+19),&nTFclus);

  // Load selected chromosomes.
  if ((fp=fopen(*(argv+20),"r"))==NULL)
    {
      printf("Cannot open %s.\n",*(argv+20));
      return -1;
    }
  nselchrs=0; selchrs=malloc(sizeof(int)*24);
  while ((fgets(s,MAXLINEWIDTH,fp))!=NULL)
    {
      item=getitemval3(s,0,&length,0x09);
      val=atof2(item,length); k=(int)val;
      free(item);
      nselchrs+=1; *(selchrs+nselchrs-1)=k;
    }
  fclose(fp);


  // Load selected methylated genes as drivers.
  if ((fp=fopen(*(argv+21),"r"))==NULL)
    {
      printf("Cannot open %s.\n",*(argv+21));
      return -1;
    }
  nselmethlgenes=0; selmethlgenes=malloc(sizeof(int)*nmethlgenes);
  while ((fgets(s,MAXLINEWIDTH,fp))!=NULL)
    {
      item=getitemval3(s,0,&length,0x09);
      val=atof2(item,length); j=(int)val-1;      
      free(item); 
      nselmethlgenes+=1; *(selmethlgenes+nselmethlgenes-1)=j;      
    }
  fclose(fp);

  // Load selected mirs as drivers.
  if ((fp=fopen(*(argv+22),"r"))==NULL)
    {
      printf("Cannot open %s.\n",*(argv+22));
      return -1;
    }
  nselmirs=0; selmirs=malloc(sizeof(int)*nmirs);
  while ((fgets(s,MAXLINEWIDTH,fp))!=NULL)
    {
      item=getitemval3(s,0,&length,0x09);
      val=atof2(item,length); j=(int)val-1;     
      free(item); 
      nselmirs+=1; *(selmirs+nselmirs-1)=j;      
    }
  fclose(fp);

  // Load selected TFs as drivers.
  if ((fp=fopen(*(argv+23),"r"))==NULL)
    {
      printf("Cannot open %s.\n",*(argv+23));
      return -1;
    }
  nselTFs=0; selTFs=malloc(sizeof(int)*nTFs);
  while ((fgets(s,MAXLINEWIDTH,fp))!=NULL)
    {
      item=getitemval3(s,0,&length,0x09);
      val=atof2(item,length); j=(int)val-1;      
      free(item); 
      nselTFs+=1; *(selTFs+nselTFs-1)=j;      
    }
  fclose(fp); 
 
  // Load the level of integrated associations.
  level=atoi(*(argv+24));

  
  // Build a lookup table of associated TFs for each gene/mir.
  // Is useful only for levels 2 and 3 associations.
  if ((level==2)||(level==3))
    {            
      geneTFs=malloc(sizeof(struct intset)*noutputs);
      for (n=0; n<=(noutputs-1); n++)
	{
	  (geneTFs+n)->nmembers=0;
	  (geneTFs+n)->members1=malloc(sizeof(int));
	  (geneTFs+n)->members2=malloc(sizeof(int));
	}
      for (k=0; k<=(nTFlist-1); k++)
	{
	  int ind1, ind2, ind3, ff=0;
	  double val1, val2, val3;
	  ind1=(int)(*(TFlist+k*6+0)-1); ind2=(int)(*(TFlist+k*6+1)-1); 
	  val1=*(TFlist+k*6+4);	 
	  (geneTFs+ind2)->nmembers+=1;
	  (geneTFs+ind2)->members1=realloc((geneTFs+ind2)->members1,sizeof(int)*(geneTFs+ind2)->nmembers);
	  *((geneTFs+ind2)->members1+(geneTFs+ind2)->nmembers-1)=ind1;
	  (geneTFs+ind2)->members2=realloc((geneTFs+ind2)->members2,sizeof(int)*(geneTFs+ind2)->nmembers);
	  *((geneTFs+ind2)->members2+(geneTFs+ind2)->nmembers-1)=k;
	    
	}   
    }

  // Build the maps of mutated genes, methylated genes and TFs and snps to genes. 
  mut2gene=malloc(sizeof(int)*nmuts);
  for (n=0; n<=(nmuts-1); n++)
    {
      *(mut2gene+n)=-1;
      i=0; 
      while ((i<=(ngenes-1))&&(*(mut2gene+n)<0))
	{
	  if (strncmp((mutgenes+n)->name,(genes+i)->name,max(strlen((mutgenes+n)->name),strlen((genes+i)->name)))==0)
	    *(mut2gene+n)=i;
	  i++;
	}
    }

  methl2gene=malloc(sizeof(int)*nmethlgenes);
  for (n=0; n<=(nmethlgenes-1); n++)
    {
      *(methl2gene+n)=-1;
      i=0; 
      while ((i<=(ngenes-1))&&(*(methl2gene+n)<0))
	{
	  if (strncmp((methlgenes+n)->name,(genes+i)->name,max(strlen((methlgenes+n)->name),strlen((genes+i)->name)))==0)
	    *(methl2gene+n)=i;
	  i++;
	}
    }

  TF2gene=malloc(sizeof(int)*nTFs); gene2TF=malloc(sizeof(int)*ngenes);
  for (n=0; n<=(ngenes-1); n++)
    *(gene2TF+n)=-1;
  for (n=0; n<=(nTFs-1); n++)
    {
      *(TF2gene+n)=-1;
      i=0; 
      while ((i<=(ngenes-1))&&(*(TF2gene+n)<0))
	{
	  if (strncmp((TFs+n)->name,(genes+i)->name,max(strlen((TFs+n)->name),strlen((genes+i)->name)))==0)
	    {
	      *(TF2gene+n)=i; *(gene2TF+i)=n;
	    }
	  i++;
	}
    }
  
  // Load llratio, p-value and correlation thresholds.
  // Line 1: level 1 intra segment CNV associations.  
  // Line 2: level 1 snp associations.
  // Line 3: level 2/3 inter segment CNV associations.
  // Line 4: level 2 mutation associations.
  // Line 5: level 2/3 methylation associations.
  // Line 6: level 2 SNP associations. 
  // Line 7: level 4 mir associations.
  // Line 8: level 7/8 TF associations.  Not used here.
  // Line 9: joint additive model associations.
  // Line 10: joint combinatorial model associations.

  if ((fp=fopen(*(argv+25),"r"))==NULL)
    {
      printf("Cannot open %s.\n",*(argv+25));
      return -1;
    }  
  llthres=malloc(sizeof(double)*10);
  pvalthres=malloc(sizeof(double)*10);
  corrthres=malloc(sizeof(double)*10);
  for (n=0; n<=9; n++)
    {
      fgets(s,MAXLINEWIDTH,fp);
      item=getitemval3(s,0,&length,0x09);
      val=atof2(item,length);
      free(item);
      *(llthres+n)=val;
      item=getitemval3(s,1,&length,0x09);
      val=atof2(item,length);
      free(item);
      *(pvalthres+n)=val;
      item=getitemval3(s,2,&length,0x09);
      val=atof2(item,length);
      free(item);
      *(corrthres+n)=val;
    }
  fclose(fp);


  // Load the range of genes to consider.
  curind=atoi(*(argv+26)); incsize=atoi(*(argv+27)); 
  startindex=curind*incsize; startindex=min(startindex,noutputs-1);
  endindex=min(startindex+incsize-1,noutputs-1);

  // Load the integrated association list established in the previous step.
  // Only consider the passenger genes within the range.
  // The association list gene/aberration indices are shifted by 1.
  if (mode==0)
    noutputs=ngenes;
  else
    noutputs=nmirs;
  preassociations=malloc(sizeof(struct intassociation)*noutputs);
  for (n=0; n<=(noutputs-1); n++)
    {
      (preassociations+n)->ndrivers=0;
      (preassociations+n)->drivertypes=malloc(sizeof(char));
      (preassociations+n)->driverinds=malloc(sizeof(int));
      (preassociations+n)->directions=malloc(sizeof(int));
      (preassociations+n)->nregulators=0;
      (preassociations+n)->regulatorinds=malloc(sizeof(int));
      (preassociations+n)->ninteractions=0;
      (preassociations+n)->intdrivertypes=malloc(sizeof(char));
      (preassociations+n)->intdriverinds=malloc(sizeof(int));
      (preassociations+n)->intdirections=malloc(sizeof(int));
      (preassociations+n)->intregulatorinds=malloc(sizeof(int));      
      (preassociations+n)->intlogratios=malloc(sizeof(double));
      (preassociations+n)->intpvals=malloc(sizeof(double));      
      
      (preassociations+n)->npassengers=1;
      (preassociations+n)->passengertypes=malloc(sizeof(char)*LONGNAMELENGTH);
      if (mode==0)
	{
	  strncpy((preassociations+n)->passengertypes,"mRNA",4);
	  *((preassociations+n)->passengertypes+4)='\0';
	}
      else
	{
	  strncpy((preassociations+n)->passengertypes,"mir",3);
	  *((preassociations+n)->passengertypes+3)='\0';
	}      
      (preassociations+n)->passengerinds=malloc(sizeof(int));
      *((preassociations+n)->passengerinds)=n;
      (preassociations+n)->logratios=malloc(sizeof(double));
      (preassociations+n)->pvals=malloc(sizeof(double));
      (preassociations+n)->representatives=malloc(sizeof(int));
    }

  if (level>1)
    {      
      if ((fp=fopen(*(argv+28),"r"))==NULL)
	{
	  printf("Cannot open %s.\n",*(argv+28));
	  return -1;
	}
      s=malloc(sizeof(char)*MAXLINEWIDTH); 
      while ((fgets(s,MAXLINEWIDTH,fp))!=NULL)
	{
	  int ind1, ind2, interaction=0;

	  if (strncmp(s,"interaction",11)!=0)
	    {	  
	      item=getitemval3(s,3,&length,0x09);
	      val=atof2(item,length);
	      free(item);
	      ind2=(int)val-1;
	    }
	  else
	    {
	      interaction=1;
	      item=getitemval3(s,7,&length,0x09);
	      val=atof2(item,length);
	      free(item);
	      ind2=(int)val-1;
	    }

	  //if ((ind2>=startindex)&&(ind2<=endindex))
	    {
	      int ff=1;

	      // Load all the associations from the prior level.  But only use the representatives in constructing the null model.
	      ff=0;
	      	      
	      if (ff==0)
		{	      	      		  
		  // Load additive terms.
		  if (interaction==0)
		    {
		      (preassociations+ind2)->ndrivers+=1;
		      (preassociations+ind2)->drivertypes=realloc((preassociations+ind2)->drivertypes,sizeof(char)*(preassociations+ind2)->ndrivers*LONGNAMELENGTH);
		      item=getitemval3(s,0,&length,0x09);
		      strncpy((preassociations+ind2)->drivertypes+((preassociations+ind2)->ndrivers-1)*LONGNAMELENGTH,item,length);
		      *((preassociations+ind2)->drivertypes+((preassociations+ind2)->ndrivers-1)*LONGNAMELENGTH+length)='\0';
		      free(item);
		      (preassociations+ind2)->driverinds=realloc((preassociations+ind2)->driverinds,sizeof(int)*(preassociations+ind2)->ndrivers);
		      item=getitemval3(s,1,&length,0x09);
		      val=atof2(item,length); free(item);
		      *((preassociations+ind2)->driverinds+(preassociations+ind2)->ndrivers-1)=(int)val-1;
		      (preassociations+ind2)->directions=realloc((preassociations+ind2)->directions,sizeof(int)*(preassociations+ind2)->ndrivers);
		      item=getitemval3(s,4,&length,0x09);
		      val=atof2(item,length); free(item);
		      *((preassociations+ind2)->directions+(preassociations+ind2)->ndrivers-1)=(int)val;
		      item=getitemval3(s,2,&length,0x09);
		      if (strncmp(item,"-",1)!=0)
			{
			  val=atof2(item,length);
			  (preassociations+ind2)->nregulators+=1;
			  (preassociations+ind2)->regulatorinds=realloc((preassociations+ind2)->regulatorinds,sizeof(int)*(preassociations+ind2)->nregulators);
			  *((preassociations+ind2)->regulatorinds+(preassociations+ind2)->nregulators-1)=(int)val-1;
			}
		      free(item);	  
		      (preassociations+ind2)->logratios=realloc((preassociations+ind2)->logratios,sizeof(double)*(preassociations+ind2)->ndrivers);
		      item=getitemval3(s,5,&length,0x09);
		      val=atof2(item,length); free(item);
		      *((preassociations+ind2)->logratios+(preassociations+ind2)->ndrivers-1)=val;
		      (preassociations+ind2)->pvals=realloc((preassociations+ind2)->pvals,sizeof(double)*(preassociations+ind2)->ndrivers);
		      item=getitemval3(s,6,&length,0x09);
		      val=atof2(item,length); free(item);
		      *((preassociations+ind2)->pvals+(preassociations+ind2)->ndrivers-1)=val;
		      (preassociations+ind2)->representatives=realloc((preassociations+ind2)->representatives,sizeof(int)*(preassociations+ind2)->ndrivers);
		      item=getitemval3(s,7,&length,0x09);
		      val=atof2(item,length); free(item);
		      *((preassociations+ind2)->representatives+(preassociations+ind2)->ndrivers-1)=(int)val;
		    }

		  // Load interaction terms.
		  else
		    {
		      (preassociations+ind2)->ninteractions+=1;
		      (preassociations+ind2)->intdrivertypes=realloc((preassociations+ind2)->intdrivertypes,sizeof(char)*(preassociations+ind2)->ninteractions*2*LONGNAMELENGTH);
		      item=getitemval3(s,1,&length,0x09);
		      strncpy((preassociations+ind2)->intdrivertypes+(((preassociations+ind2)->ninteractions-1)*2+0)*LONGNAMELENGTH,item,length);
		      *((preassociations+ind2)->intdrivertypes+(((preassociations+ind2)->ninteractions-1)*2+0)*LONGNAMELENGTH+length)='\0';
		      free(item);
		      item=getitemval3(s,4,&length,0x09);
		      strncpy((preassociations+ind2)->intdrivertypes+(((preassociations+ind2)->ninteractions-1)*2+1)*LONGNAMELENGTH,item,length);
		      *((preassociations+ind2)->intdrivertypes+(((preassociations+ind2)->ninteractions-1)*2+1)*LONGNAMELENGTH+length)='\0';
		      free(item);
		      (preassociations+ind2)->intdriverinds=realloc((preassociations+ind2)->intdriverinds,sizeof(int)*(preassociations+ind2)->ninteractions*2);
		      item=getitemval3(s,2,&length,0x09);
		      val=atof2(item,length); free(item);
		      *((preassociations+ind2)->intdriverinds+((preassociations+ind2)->ninteractions-1)*2+0)=(int)val-1;
		      item=getitemval3(s,5,&length,0x09);
		      val=atof2(item,length); free(item);
		      *((preassociations+ind2)->intdriverinds+((preassociations+ind2)->ninteractions-1)*2+1)=(int)val-1;
		      (preassociations+ind2)->intregulatorinds=realloc((preassociations+ind2)->intregulatorinds,sizeof(int)*(preassociations+ind2)->ninteractions*2);
		      item=getitemval3(s,3,&length,0x09);
		      if (strncmp(item,"-",1)!=0)
			{
			  val=atof2(item,length);			  
			  *((preassociations+ind2)->intregulatorinds+((preassociations+ind2)->ninteractions-1)*2+0)=(int)val-1;
			}
		      else
			*((preassociations+ind2)->intregulatorinds+((preassociations+ind2)->ninteractions-1)*2+0)=-1;
		      free(item);
		      item=getitemval3(s,6,&length,0x09);
		      if (strncmp(item,"-",1)!=0)
			{
			  val=atof2(item,length);			  
			  *((preassociations+ind2)->intregulatorinds+((preassociations+ind2)->ninteractions-1)*2+1)=(int)val-1;
			}
		      else
			*((preassociations+ind2)->intregulatorinds+((preassociations+ind2)->ninteractions-1)*2+1)=-1;
		      free(item);
		      (preassociations+ind2)->intdirections=realloc((preassociations+ind2)->intdirections,sizeof(int)*(preassociations+ind2)->ninteractions);
		      item=getitemval3(s,8,&length,0x09);
		      val=atof2(item,length); free(item);
		      *((preassociations+ind2)->intdirections+(preassociations+ind2)->ninteractions-1)=(int)val;
		      
		      (preassociations+ind2)->intlogratios=realloc((preassociations+ind2)->intlogratios,sizeof(double)*(preassociations+ind2)->ninteractions);
		      item=getitemval3(s,9,&length,0x09);
		      val=atof2(item,length); free(item);
		      *((preassociations+ind2)->intlogratios+(preassociations+ind2)->ninteractions-1)=val;
		      (preassociations+ind2)->intpvals=realloc((preassociations+ind2)->intpvals,sizeof(double)*(preassociations+ind2)->ninteractions);
		      item=getitemval3(s,6,&length,0x09);
		      val=atof2(item,length); free(item);
		      *((preassociations+ind2)->intpvals+(preassociations+ind2)->ninteractions-1)=val;

		      (preassociations+ind2)->representatives=realloc((preassociations+ind2)->representatives,sizeof(int)*(preassociations+ind2)->ndrivers);
		      *((preassociations+ind2)->representatives+(preassociations+ind2)->ndrivers-1)=1;
		    }
		}
	    }
	}
      fclose(fp);
    }


  // Load the integrated association list established in the previous step.
  // If the outputs are mRNAs, then this list is for mirs.
  // If the outputs are mirs, then this list is for mRNAs.
  // Only consider the passenger genes within the range.
  // The association list gene/aberration indices are shifted by 1.  
  if (mode==0)
    kk=nmirs;
  else
    kk=ngenes;
  preassociations2=malloc(sizeof(struct intassociation)*kk);
  for (n=0; n<=(kk-1); n++)
    {
      (preassociations2+n)->ndrivers=0;
      (preassociations2+n)->drivertypes=malloc(sizeof(char));
      (preassociations2+n)->driverinds=malloc(sizeof(int));
      (preassociations2+n)->directions=malloc(sizeof(int));
      (preassociations2+n)->nregulators=0;
      (preassociations2+n)->regulatorinds=malloc(sizeof(int));
      (preassociations2+n)->ninteractions=0;
      (preassociations2+n)->intdrivertypes=malloc(sizeof(char));
      (preassociations2+n)->intdriverinds=malloc(sizeof(int));
      (preassociations2+n)->intdirections=malloc(sizeof(int));
      (preassociations2+n)->intregulatorinds=malloc(sizeof(int));  
      (preassociations2+n)->intlogratios=malloc(sizeof(double));
      (preassociations2+n)->intpvals=malloc(sizeof(double));      
      
      (preassociations2+n)->npassengers=1;
      (preassociations2+n)->passengertypes=malloc(sizeof(char)*LONGNAMELENGTH);
      if (mode==1)
	{
	  strncpy((preassociations2+n)->passengertypes,"mRNA",4);
	  *((preassociations2+n)->passengertypes+4)='\0';
	}
      else
	{
	  strncpy((preassociations2+n)->passengertypes,"mir",3);
	  *((preassociations2+n)->passengertypes+3)='\0';
	}      
      (preassociations2+n)->passengerinds=malloc(sizeof(int));
      *((preassociations2+n)->passengerinds)=n;
      (preassociations2+n)->logratios=malloc(sizeof(double));
      (preassociations2+n)->pvals=malloc(sizeof(double));
      (preassociations2+n)->representatives=malloc(sizeof(int));
    }

  // trans-acting SNPs also need intermediate regulators.
  if (level>1)
    {
      if ((fp=fopen(*(argv+29),"r"))==NULL)
	{
	  printf("Cannot open %s.\n",*(argv+29));
	  return -1;
	}
      s=malloc(sizeof(char)*MAXLINEWIDTH); 
      while ((fgets(s,MAXLINEWIDTH,fp))!=NULL)
	{
	  int ind1, ind2, interaction=0;

	  if (strncmp(s,"interaction",11)!=0)
	    {	  
	      item=getitemval3(s,3,&length,0x09);
	      val=atof2(item,length);
	      free(item);
	      ind2=(int)val-1;
	    }
	  else
	    {
	      interaction=1;
	      item=getitemval3(s,7,&length,0x09);
	      val=atof2(item,length);
	      free(item);
	      ind2=(int)val-1;
	    }

	  //if ((ind2>=startindex)&&(ind2<=endindex))
	    {
	      int ff=1;

	      if (strncmp(s,"interaction",11)!=0)
		{
		  // When loading the prior integrated models, only load one inter segment CNV model with the same segment but different intermediate regulators.
		  item=getitemval3(s,0,&length,0x09);
		  if (strncmp(item,"inter segment CNV",17)!=0)
		    ff=0;
		  free(item);
		  if (ff==1)
		    {
		      item=getitemval3(s,1,&length,0x09);
		      val=atof2(item,length); free(item); k=(int)val-1;
		      i=0; j=-1;
		      while ((i<=((preassociations2+ind2)->ndrivers-1))&&(j<0))
			{
			  if ((strncmp((preassociations2+ind2)->drivertypes+i*LONGNAMELENGTH,"inter segment CNV",17)==0)&&(*((preassociations2+ind2)->driverinds+i)==k))
			    j=i;
			  i++;
			}
		      if (j<0)
			ff=0;
		    }
		}
	      else
		ff=0;

	      if (ff==0)
		{
		  // Load additive terms.
		  if (interaction==0)
		    {
		      (preassociations2+ind2)->ndrivers+=1;
		      (preassociations2+ind2)->drivertypes=realloc((preassociations2+ind2)->drivertypes,sizeof(char)*(preassociations2+ind2)->ndrivers*LONGNAMELENGTH);
		      item=getitemval3(s,0,&length,0x09);
		      strncpy((preassociations2+ind2)->drivertypes+((preassociations2+ind2)->ndrivers-1)*LONGNAMELENGTH,item,length);
		      *((preassociations2+ind2)->drivertypes+((preassociations2+ind2)->ndrivers-1)*LONGNAMELENGTH+length)='\0';
		      free(item);
		      (preassociations2+ind2)->driverinds=realloc((preassociations2+ind2)->driverinds,sizeof(int)*(preassociations2+ind2)->ndrivers);
		      item=getitemval3(s,1,&length,0x09);
		      val=atof2(item,length); free(item);
		      *((preassociations2+ind2)->driverinds+(preassociations2+ind2)->ndrivers-1)=(int)val-1;
		      (preassociations2+ind2)->directions=realloc((preassociations2+ind2)->directions,sizeof(int)*(preassociations2+ind2)->ndrivers);
		      item=getitemval3(s,4,&length,0x09);
		      val=atof2(item,length); free(item);
		      *((preassociations2+ind2)->directions+(preassociations2+ind2)->ndrivers-1)=(int)val;
		      item=getitemval3(s,2,&length,0x09);
		      if (strncmp(item,"NA",2)!=0)
			{
			  val=atof2(item,length);
			  (preassociations2+ind2)->nregulators+=1;
			  (preassociations2+ind2)->regulatorinds=realloc((preassociations2+ind2)->regulatorinds,sizeof(int)*(preassociations2+ind2)->nregulators);
			  *((preassociations2+ind2)->regulatorinds+(preassociations2+ind2)->nregulators-1)=(int)val-1;
			}
		      free(item);	  
		      (preassociations2+ind2)->logratios=realloc((preassociations2+ind2)->logratios,sizeof(double)*(preassociations2+ind2)->ndrivers);
		      item=getitemval3(s,5,&length,0x09);
		      val=atof2(item,length); free(item);
		      *((preassociations2+ind2)->logratios+(preassociations2+ind2)->ndrivers-1)=val;
		      (preassociations2+ind2)->pvals=realloc((preassociations2+ind2)->pvals,sizeof(double)*(preassociations2+ind2)->ndrivers);
		      item=getitemval3(s,6,&length,0x09);
		      val=atof2(item,length); free(item);
		      *((preassociations2+ind2)->pvals+(preassociations2+ind2)->ndrivers-1)=val;
		      (preassociations2+ind2)->representatives=realloc((preassociations2+ind2)->representatives,sizeof(int)*(preassociations2+ind2)->ndrivers);
		      item=getitemval3(s,7,&length,0x09);
		      val=atof2(item,length); free(item);
		      *((preassociations2+ind2)->representatives+(preassociations2+ind2)->ndrivers-1)=(int)val;
		    }

		  // Load interaction terms.
		  else
		    {
		      (preassociations2+ind2)->ninteractions+=1;
		      (preassociations2+ind2)->intdrivertypes=realloc((preassociations2+ind2)->intdrivertypes,sizeof(char)*(preassociations2+ind2)->ninteractions*2*LONGNAMELENGTH);
		      item=getitemval3(s,1,&length,0x09);
		      strncpy((preassociations2+ind2)->intdrivertypes+(((preassociations2+ind2)->ninteractions-1)*2+0)*LONGNAMELENGTH,item,length);
		      *((preassociations2+ind2)->intdrivertypes+(((preassociations2+ind2)->ninteractions-1)*2+0)*LONGNAMELENGTH+length)='\0';
		      free(item);
		      item=getitemval3(s,4,&length,0x09);
		      strncpy((preassociations2+ind2)->intdrivertypes+(((preassociations2+ind2)->ninteractions-1)*2+1)*LONGNAMELENGTH,item,length);
		      *((preassociations2+ind2)->intdrivertypes+(((preassociations2+ind2)->ninteractions-1)*2+1)*LONGNAMELENGTH+length)='\0';
		      free(item);
		      (preassociations2+ind2)->intdriverinds=realloc((preassociations2+ind2)->intdriverinds,sizeof(int)*(preassociations2+ind2)->ninteractions*2);
		      item=getitemval3(s,2,&length,0x09);
		      val=atof2(item,length); free(item);
		      *((preassociations2+ind2)->intdriverinds+((preassociations2+ind2)->ninteractions-1)*2+0)=(int)val-1;
		      item=getitemval3(s,5,&length,0x09);
		      val=atof2(item,length); free(item);
		      *((preassociations2+ind2)->intdriverinds+((preassociations2+ind2)->ninteractions-1)*2+1)=(int)val-1;
		      (preassociations2+ind2)->intregulatorinds=realloc((preassociations2+ind2)->intregulatorinds,sizeof(int)*(preassociations2+ind2)->ninteractions*2);
		      item=getitemval3(s,3,&length,0x09);
		      if (strncmp(item,"-",1)!=0)
			{
			  val=atof2(item,length);			  
			  *((preassociations2+ind2)->intregulatorinds+((preassociations2+ind2)->ninteractions-1)*2+0)=(int)val-1;
			}
		      else
			*((preassociations2+ind2)->intregulatorinds+((preassociations2+ind2)->ninteractions-1)*2+0)=-1;
		      free(item);
		      item=getitemval3(s,6,&length,0x09);
		      if (strncmp(item,"-",1)!=0)
			{
			  val=atof2(item,length);			  
			  *((preassociations2+ind2)->intregulatorinds+((preassociations2+ind2)->ninteractions-1)*2+1)=(int)val-1;
			}
		      else
			*((preassociations2+ind2)->intregulatorinds+((preassociations2+ind2)->ninteractions-1)*2+1)=-1;
		      free(item);
		      (preassociations2+ind2)->intdirections=realloc((preassociations2+ind2)->intdirections,sizeof(int)*(preassociations2+ind2)->ninteractions);
		      item=getitemval3(s,8,&length,0x09);
		      val=atof2(item,length); free(item);
		      *((preassociations2+ind2)->intdirections+(preassociations2+ind2)->ninteractions-1)=(int)val;
		      
		      (preassociations2+ind2)->intlogratios=realloc((preassociations2+ind2)->intlogratios,sizeof(double)*(preassociations2+ind2)->ninteractions);
		      item=getitemval3(s,9,&length,0x09);
		      val=atof2(item,length); free(item);
		      *((preassociations2+ind2)->intlogratios+(preassociations2+ind2)->ninteractions-1)=val;
		      (preassociations2+ind2)->intpvals=realloc((preassociations2+ind2)->intpvals,sizeof(double)*(preassociations2+ind2)->ninteractions);
		      item=getitemval3(s,6,&length,0x09);
		      val=atof2(item,length); free(item);
		      *((preassociations2+ind2)->intpvals+(preassociations2+ind2)->ninteractions-1)=val;
		      (preassociations2+ind2)->representatives=realloc((preassociations2+ind2)->representatives,sizeof(int)*(preassociations2+ind2)->ndrivers);
		      *((preassociations2+ind2)->representatives+(preassociations2+ind2)->ndrivers-1)=1;
		    }
		}
	    }
	}
      fclose(fp);
    }

  // Load candidate pairwise associations.
  associations=malloc(sizeof(struct intassociation)*noutputs);
  for (geneind=0; geneind<=(noutputs-1); geneind++)
    {
      (associations+geneind)->ndrivers=0;
      (associations+geneind)->drivertypes=malloc(sizeof(char));
      (associations+geneind)->driverinds=malloc(sizeof(int));
      (associations+geneind)->directions=malloc(sizeof(int));
      (associations+geneind)->nregulators=0;
      (associations+geneind)->regulatorinds=malloc(sizeof(int));
      (associations+geneind)->ninteractions=0;
      (associations+geneind)->intdrivertypes=malloc(sizeof(char));
      (associations+geneind)->intdriverinds=malloc(sizeof(int));
      (associations+geneind)->intdirections=malloc(sizeof(int));
      (associations+geneind)->intregulatorinds=malloc(sizeof(int));    
      (associations+geneind)->intlogratios=malloc(sizeof(double));
      (associations+geneind)->intpvals=malloc(sizeof(double));            
      (associations+geneind)->npassengers=1;
      (associations+geneind)->passengertypes=malloc(sizeof(char)*LONGNAMELENGTH);
      if (mode==0)
	{
	  strncpy((associations+geneind)->passengertypes,"mrna",4);
	  *((associations+geneind)->passengertypes+4)='\0';
	}
      else
	{
	  strncpy((associations+geneind)->passengertypes,"mir",3);
	  *((associations+geneind)->passengertypes+3)='\0';
	}
      (associations+geneind)->passengerinds=malloc(sizeof(int));
      *((associations+geneind)->passengerinds+0)=geneind;
      (associations+geneind)->logratios=malloc(sizeof(double));
      (associations+geneind)->pvals=malloc(sizeof(double));
    }

  // Level 1: intra segment CNVs, local mutations and methylations, local SNPs.
  // Local SNPs refer to the SNPs that pass the logratio, p-value and correlation thresholds and are within a distance as the targets.  
  if (level==1)
    {      
      int *labeled;

      // intra segment CNV list data.
      labeled=malloc(sizeof(int)*noutputs);
      for (n=0; n<=(noutputs-1); n++)
	*(labeled+n)=0;
      for (n=0; n<=(nintraseglist-1); n++)
	{
	  double val1, val2, val3;
	  i=(int)(*(intraseglist+n*ncols+0)-1);
	  j=(int)(*(intraseglist+n*ncols+1)-1);
	  val1=*(intraseglist+n*ncols+3);
	  val2=*(intraseglist+n*ncols+4);
	  val3=*(intraseglist+n*ncols+5);
	  
	  if ((j>=startindex)&&(j<=endindex)&&(val1>=*(llthres+0))&&(val2<=*(pvalthres+0))&&(val3>=*(corrthres+0)))
	    {
	      *(labeled+j)=1;
	      (associations+j)->ndrivers+=1;
	      (associations+j)->drivertypes=realloc((associations+j)->drivertypes,sizeof(char)*(associations+j)->ndrivers*LONGNAMELENGTH);
	      strncpy((associations+j)->drivertypes+((associations+j)->ndrivers-1)*LONGNAMELENGTH,"intra segment CNV",17);
	      *((associations+j)->drivertypes+((associations+j)->ndrivers-1)*LONGNAMELENGTH+17)='\0';
	      (associations+j)->driverinds=realloc((associations+j)->driverinds,sizeof(int)*(associations+j)->ndrivers);
	      *((associations+j)->driverinds+(associations+j)->ndrivers-1)=i;
	      (associations+j)->directions=realloc((associations+j)->directions,sizeof(int)*(associations+j)->ndrivers);
	      if (val3>=0)
		*((associations+j)->directions+(associations+j)->ndrivers-1)=1;
	      else
		*((associations+j)->directions+(associations+j)->ndrivers-1)=-1;		  
	      (associations+j)->logratios=realloc((associations+j)->logratios,sizeof(double)*(associations+j)->ndrivers);
	      *((associations+j)->logratios+(associations+j)->ndrivers-1)=val1;
	      (associations+j)->pvals=realloc((associations+j)->pvals,sizeof(double)*(associations+j)->ndrivers);
	      *((associations+j)->pvals+(associations+j)->ndrivers-1)=val2;		
	    }
	}

      // inter segment list data.
      // If an external segment is associated with a gene and is on the same chromosome and the gene does not have intra segment CNV associations, then label it as intra segment CNV.
      // If there are multiple inter segment CNVs on the same chromosome, then assign all of them as candidates.
      // The one with the highest score will be picked up at the integration step.
      for (n=0; n<=(ninterseglist-1); n++)
	{
	  double val1, val2, val3;
	  i=(int)(*(interseglist+n*ncols+0)-1);
	  j=(int)(*(interseglist+n*ncols+1)-1);
	  val1=*(interseglist+n*ncols+3);
	  val2=*(interseglist+n*ncols+4);
	  val3=*(interseglist+n*ncols+5);

	  if ((j>=startindex)&&(j<=endindex)&&(val1>=*(llthres+0))&&(val2<=*(pvalthres+0))&&(val3>=*(corrthres+0)))
	    {
	      int flag2=0, flag3=0;
	      
	      
	      // Check whether segment i and gene j are on the same chromosome and have positive associations.
	      // If so check whether intra segment CNV association exists.
	      // If the intra segment CNV association does not exist, then label it as intra segment CNV. 
	      flag3=0;
	      if ((mode==0)&&((segments+i)->chr==(genes+j)->chr)&&(val3>0))
		flag3=1;
	      else if ((mode==1)&&((segments+i)->chr==(mirs+j)->chr)&&(val3>0))
		flag3=1;
	      
	      if (flag3==1)
		{
		  k=0; l=-1;
		  while ((k<=(nintraseglist-1))&&(l<0))
		    {
		      m=(int)(*(intraseglist+k*ncols+1)-1);
		      if (m==j)
			l=k;
		      k++;
		    }
		  if (l<0)
		    flag2=1;
		}	      

	      
	      // Check whether segment i and gene j are on the same chromosome and have positive associations.
	      // If so check whether intra segment CNV association exists.
	      // If the intra segment CNV association does not exist, then label it as intra segment CNV.
	      flag3=0;
	      if ((mode==0)&&((segments+i)->chr==(genes+j)->chr)&&(val3>0))
		flag3=1;
	      else if ((mode==1)&&((segments+i)->chr==(mirs+j)->chr)&&(val3>0))
		flag3=1;
	      
	      if (flag3==1)
		{		  
		  if (*(labeled+j)==0)
		    flag2=1;
		}	   

	      // Use an inter segment CNV as an intra segment CNV.
	      if (flag2==1)	  
		{
		  (associations+j)->ndrivers+=1;
		  (associations+j)->drivertypes=realloc((associations+j)->drivertypes,sizeof(char)*(associations+j)->ndrivers*LONGNAMELENGTH);
		  strncpy((associations+j)->drivertypes+((associations+j)->ndrivers-1)*LONGNAMELENGTH,"intra segment CNV",17);
		  *((associations+j)->drivertypes+((associations+j)->ndrivers-1)*LONGNAMELENGTH+17)='\0';
		  (associations+j)->driverinds=realloc((associations+j)->driverinds,sizeof(int)*(associations+j)->ndrivers);
		  *((associations+j)->driverinds+(associations+j)->ndrivers-1)=i;
		  (associations+j)->directions=realloc((associations+j)->directions,sizeof(int)*(associations+j)->ndrivers);
		  if (val3>=0)
		    *((associations+j)->directions+(associations+j)->ndrivers-1)=1;
		  else
		    *((associations+j)->directions+(associations+j)->ndrivers-1)=-1;
		  (associations+j)->logratios=realloc((associations+j)->logratios,sizeof(double)*(associations+j)->ndrivers);
		  *((associations+j)->logratios+(associations+j)->ndrivers-1)=val1;
		  (associations+j)->pvals=realloc((associations+j)->pvals,sizeof(double)*(associations+j)->ndrivers);
		  *((associations+j)->pvals+(associations+j)->ndrivers-1)=val2;
		}
	    }
	}
    
      free(labeled);

      // Only consider mRNA associations for local mutations and methylations.
      // Consider mRNA associations for local SNPs.
      if (mode==0)
	{
	  // mutation list.
	  // Only consider the local associations.
	  for (n=0; n<=(nmutlist-1); n++)
	    {
	      double val1, val2, val3;
	      i=(int)(*(mutlist+n*ncols+0)-1);
	      j=(int)(*(mutlist+n*ncols+1)-1);
	      l=*(mut2gene+i);
	      val1=*(mutlist+n*ncols+3);
	      val2=*(mutlist+n*ncols+4);
	      val3=*(mutlist+n*ncols+5);
	      
	      if ((j>=startindex)&&(j<=endindex)&&(l==j))
		{
		  (associations+j)->ndrivers+=1;
		  (associations+j)->drivertypes=realloc((associations+j)->drivertypes,sizeof(char)*(associations+j)->ndrivers*LONGNAMELENGTH);
		  strncpy((associations+j)->drivertypes+((associations+j)->ndrivers-1)*LONGNAMELENGTH,"mutation",8);
		  *((associations+j)->drivertypes+((associations+j)->ndrivers-1)*LONGNAMELENGTH+8)='\0';
		  (associations+j)->driverinds=realloc((associations+j)->driverinds,sizeof(int)*(associations+j)->ndrivers);
		  *((associations+j)->driverinds+(associations+j)->ndrivers-1)=i;
		  (associations+j)->directions=realloc((associations+j)->directions,sizeof(int)*(associations+j)->ndrivers);
		  if (val3>=0)
		    *((associations+j)->directions+(associations+j)->ndrivers-1)=1;
		  else
		    *((associations+j)->directions+(associations+j)->ndrivers-1)=-1;
		  (associations+j)->logratios=realloc((associations+j)->logratios,sizeof(double)*(associations+j)->ndrivers);
		  *((associations+j)->logratios+(associations+j)->ndrivers-1)=val1;
		  (associations+j)->pvals=realloc((associations+j)->pvals,sizeof(double)*(associations+j)->ndrivers);
		  *((associations+j)->pvals+(associations+j)->ndrivers-1)=val2;
		}
	    }
	  
	  // methylation list.
	  // Only consider the local negative associations.
	  for (n=0; n<=(nmethllist-1); n++)
	    {
	      double val1, val2, val3;
	      i=(int)(*(methllist+n*ncols+0)-1);
	      j=(int)(*(methllist+n*ncols+1)-1);
	      l=*(methl2gene+i);
	      val1=*(methllist+n*ncols+3);
	      val2=*(methllist+n*ncols+4);
	      val3=*(methllist+n*ncols+5);
	      
	      if ((j>=startindex)&&(j<=endindex)&&(l==j)&&(val3<0))
		{	      
		  (associations+j)->ndrivers+=1;
		  (associations+j)->drivertypes=realloc((associations+j)->drivertypes,sizeof(char)*(associations+j)->ndrivers*LONGNAMELENGTH);
		  strncpy((associations+j)->drivertypes+((associations+j)->ndrivers-1)*LONGNAMELENGTH,"methylation",11);
		  *((associations+j)->drivertypes+((associations+j)->ndrivers-1)*LONGNAMELENGTH+11)='\0';
		  (associations+j)->driverinds=realloc((associations+j)->driverinds,sizeof(int)*(associations+j)->ndrivers);
		  *((associations+j)->driverinds+(associations+j)->ndrivers-1)=i;
		  (associations+j)->directions=realloc((associations+j)->directions,sizeof(int)*(associations+j)->ndrivers);
		  if (val3>=0)
		    *((associations+j)->directions+(associations+j)->ndrivers-1)=1;
		  else
		    *((associations+j)->directions+(associations+j)->ndrivers-1)=-1;
		  (associations+j)->logratios=realloc((associations+j)->logratios,sizeof(double)*(associations+j)->ndrivers);
		  *((associations+j)->logratios+(associations+j)->ndrivers-1)=val1;
		  (associations+j)->pvals=realloc((associations+j)->pvals,sizeof(double)*(associations+j)->ndrivers);
		  *((associations+j)->pvals+(associations+j)->ndrivers-1)=val2;
		}
	    }
	
	  // SNP list.
	  // Only consider the local SNP associations.
	  // Here the SNP of a local association is within certain distance from the target mRNA.  It may not be on the same gene.
	  for (n=0; n<=(nsnplist-1); n++)
	    {
	      double val1, val2, val3, val4;
	      i=(int)(*(snplist+n*ncols+0)-1);
	      j=(int)(*(snplist+n*ncols+1)-1);	      
	      val1=*(snplist+n*ncols+3);
	      val2=*(snplist+n*ncols+4);
	      val3=*(snplist+n*ncols+5);	      	  	      	     

	      // Check whether SNP probe i and gene j are on the same chromosome and within 1Mb.
	      if ((j>=startindex)&&(j<=endindex)&&(val1>=*(llthres+1))&&(val2<=*(pvalthres+1))&&(fabs(val3)>=*(corrthres+1))&&((snps+i)->chr==(genes+j)->chr)&&(fabs(((snps+i)->startcoord+(snps+i)->endcoord-(genes+j)->startcoord-(genes+j)->endcoord)/2)<=1e6))
		{			  		      
		  (associations+j)->ndrivers+=1;
		  (associations+j)->drivertypes=realloc((associations+j)->drivertypes,sizeof(char)*(associations+j)->ndrivers*LONGNAMELENGTH);
		  strncpy((associations+j)->drivertypes+((associations+j)->ndrivers-1)*LONGNAMELENGTH,"intra snp",9);
		  *((associations+j)->drivertypes+((associations+j)->ndrivers-1)*LONGNAMELENGTH+9)='\0';
		  (associations+j)->driverinds=realloc((associations+j)->driverinds,sizeof(int)*(associations+j)->ndrivers);
		  *((associations+j)->driverinds+(associations+j)->ndrivers-1)=*(snplistinds+n);
		  (associations+j)->directions=realloc((associations+j)->directions,sizeof(int)*(associations+j)->ndrivers);
		  if (val3>=0)
		    *((associations+j)->directions+(associations+j)->ndrivers-1)=1;
		  else
		    *((associations+j)->directions+(associations+j)->ndrivers-1)=-1;
		  (associations+j)->logratios=realloc((associations+j)->logratios,sizeof(double)*(associations+j)->ndrivers);
		  *((associations+j)->logratios+(associations+j)->ndrivers-1)=val1;
		  (associations+j)->pvals=realloc((associations+j)->pvals,sizeof(double)*(associations+j)->ndrivers);
		  *((associations+j)->pvals+(associations+j)->ndrivers-1)=val2;
		}
	    }
	}
    
      // Consider mir associations for local SNPs.
      else if (mode==1)
	{	  
	  // SNP list.
	  // Only consider the local SNP associations.
	  // Here the SNP of a local association is within certain distance from the target mRNA.  It may not be on the same gene.
	  for (n=0; n<=(nmirsnplist-1); n++)
	    {
	      double val1, val2, val3;
	      i=(int)(*(mirsnplist+n*ncols+0)-1);
	      j=(int)(*(mirsnplist+n*ncols+1)-1);	      
	      val1=*(mirsnplist+n*ncols+3);
	      val2=*(mirsnplist+n*ncols+4);
	      val3=*(mirsnplist+n*ncols+5);
	  
	      // Check whether SNP probe i and gene j are on the same chromosome and within 1Mb.
	      if ((j>=startindex)&&(j<=endindex)&&(val1>=*(llthres+1))&&(val2<=*(pvalthres+1))&&(fabs(val3)>=*(corrthres+1))&&((snps+i)->chr==(mirs+j)->chr)&&(fabs(((snps+i)->startcoord+(snps+i)->endcoord-(mirs+j)->startcoord-(mirs+j)->endcoord)/2)<=1e6))
		{	      
		  (associations+j)->ndrivers+=1;
		  (associations+j)->drivertypes=realloc((associations+j)->drivertypes,sizeof(char)*(associations+j)->ndrivers*LONGNAMELENGTH);
		  strncpy((associations+j)->drivertypes+((associations+j)->ndrivers-1)*LONGNAMELENGTH,"intra snp",9);
		  *((associations+j)->drivertypes+((associations+j)->ndrivers-1)*LONGNAMELENGTH+9)='\0';
		  (associations+j)->driverinds=realloc((associations+j)->driverinds,sizeof(int)*(associations+j)->ndrivers);
		  *((associations+j)->driverinds+(associations+j)->ndrivers-1)=*(mirsnplistinds+n);
		  (associations+j)->directions=realloc((associations+j)->directions,sizeof(int)*(associations+j)->ndrivers);
		  if (val3>=0)
		    *((associations+j)->directions+(associations+j)->ndrivers-1)=1;
		  else
		    *((associations+j)->directions+(associations+j)->ndrivers-1)=-1;
		  (associations+j)->logratios=realloc((associations+j)->logratios,sizeof(double)*(associations+j)->ndrivers);
		  *((associations+j)->logratios+(associations+j)->ndrivers-1)=val1;
		  (associations+j)->pvals=realloc((associations+j)->pvals,sizeof(double)*(associations+j)->ndrivers);
		  *((associations+j)->pvals+(associations+j)->ndrivers-1)=val2;
		}
	    }
	}    
    }

  // Level 2: inter segment CNV, non-local mutation, non-local methylation, non-local SNPs.
  // Only consider positive associations with segments on selected chromosomes.
  // Only consider associations on other chromosomes.
  else if (level==2)
    {            
      // inter segment CNV list.
      // If an external segment is associated with a gene and is on the same chromosome and the gene has intra segment CNV associations,  then label it as intra segment CNV.
      // If an external segment is on a different chromosome and associated with the gene, and there is at least one intermediate regulator, then label it as inter segment CNV.
      // Only consider the segments on selected chromosomes.
      for (n=0; n<=(ninterseglist-1); n++)
	{
	  int chr1, chr2;
	  double val1, val2, val3;
	  i=(int)(*(interseglist+n*ncols+0)-1);
	  j=(int)(*(interseglist+n*ncols+1)-1);
	  val1=*(interseglist+n*ncols+3);
	  val2=*(interseglist+n*ncols+4);
	  val3=*(interseglist+n*ncols+5);
	  chr1=(segments+i)->chr; 
	  if (mode==0)
	    chr2=(genes+j)->chr;
	  else
	    chr2=(mirs+j)->chr;

	  if ((j>=startindex)&&(j<=endindex)&&(val1>=*(llthres+2))&&(val2<=*(pvalthres+2))&&(val3>=*(corrthres+2))&&(chr1!=chr2)&&(inlist(nselchrs,selchrs,chr1)==1))
	    {
	      int flag2=0;
	      	      
	      // The association is not a candidate intra segment CNV.
	      if (flag2==0)
		{
		  int nregs, *regs;

		  // Check whether the segment accommodates an intermediate regulator.
		  // If mode=1, then look at preassociations2 instead of preassociations.
		  nregs=0; regs=malloc(sizeof(int)*nTFs);
		  for (k=0; k<=((geneTFs+j)->nmembers-1); k++)
		    {
		      l=*((geneTFs+j)->members1+k); l=*(TF2gene+l);
		      m=0; p=-1;

		      if (mode==0)
			{
			  while ((m<=((preassociations+l)->ndrivers-1))&&(p<0))
			    {
			      if ((strncmp((preassociations+l)->drivertypes+m*LONGNAMELENGTH,"intra segment CNV",17)==0)&&(*((preassociations+l)->driverinds+m)==i))
				p=m;
			      m++;
			    }
			}
		      else
			{
			  while ((m<=((preassociations2+l)->ndrivers-1))&&(p<0))
			    {
			      if ((strncmp((preassociations2+l)->drivertypes+m*LONGNAMELENGTH,"intra segment CNV",17)==0)&&(*((preassociations2+l)->driverinds+m)==i))
				p=m;
			      m++;
			    }
			}

		      if (p>=0)
			{
			  nregs++; *(regs+nregs-1)=*((geneTFs+j)->members1+k);
			}
		    }
		 
		  // Report associations with all intermediate TFs.
		  if (nregs>0)
		    {
		      int pp;
		      for (pp=0; pp<=(nregs-1); pp++)
			{
			  (associations+j)->ndrivers+=1;
			  (associations+j)->drivertypes=realloc((associations+j)->drivertypes,sizeof(char)*(associations+j)->ndrivers*LONGNAMELENGTH);
			  strncpy((associations+j)->drivertypes+((associations+j)->ndrivers-1)*LONGNAMELENGTH,"inter segment CNV",17);
			  *((associations+j)->drivertypes+((associations+j)->ndrivers-1)*LONGNAMELENGTH+17)='\0';
			  (associations+j)->driverinds=realloc((associations+j)->driverinds,sizeof(int)*(associations+j)->ndrivers);
			  *((associations+j)->driverinds+(associations+j)->ndrivers-1)=i;
			  (associations+j)->directions=realloc((associations+j)->directions,sizeof(int)*(associations+j)->ndrivers);
			  if (val3>=0)
			    *((associations+j)->directions+(associations+j)->ndrivers-1)=1;
			  else
			    *((associations+j)->directions+(associations+j)->ndrivers-1)=-1;
			  (associations+j)->nregulators+=1;
			  (associations+j)->regulatorinds=realloc((associations+j)->regulatorinds,sizeof(int)*(associations+j)->nregulators);
			  *((associations+j)->regulatorinds+(associations+j)->nregulators-1)=*(regs+pp);
			  (associations+j)->logratios=realloc((associations+j)->logratios,sizeof(double)*(associations+j)->ndrivers);
			  *((associations+j)->logratios+(associations+j)->ndrivers-1)=val1;
			  (associations+j)->pvals=realloc((associations+j)->pvals,sizeof(double)*(associations+j)->ndrivers);
			  *((associations+j)->pvals+(associations+j)->ndrivers-1)=val2;			  
			}
		    }
		  free(regs);
		}
	    }
	}		  

      // non-local mutation.
      for (n=0; n<=(nmutlist-1); n++)
	{
	  double val1, val2, val3;
	  i=(int)(*(mutlist+n*ncols+0)-1);
	  j=(int)(*(mutlist+n*ncols+1)-1);
	  l=*(mut2gene+i);
	  val1=*(mutlist+n*ncols+3);
	  val2=*(mutlist+n*ncols+4);
	  val3=*(mutlist+n*ncols+5);

	  if ((j>=startindex)&&(j<=endindex)&&(val1>=*(llthres+3))&&(val2<=*(pvalthres+3))&&(fabs(val3)>=*(corrthres+3))&&(l!=j))
	    {
	      (associations+j)->ndrivers+=1;
	      (associations+j)->drivertypes=realloc((associations+j)->drivertypes,sizeof(char)*(associations+j)->ndrivers*LONGNAMELENGTH);
	      strncpy((associations+j)->drivertypes+((associations+j)->ndrivers-1)*LONGNAMELENGTH,"mutation",8);
	      *((associations+j)->drivertypes+((associations+j)->ndrivers-1)*LONGNAMELENGTH+8)='\0';
	      (associations+j)->driverinds=realloc((associations+j)->driverinds,sizeof(int)*(associations+j)->ndrivers);
	      *((associations+j)->driverinds+(associations+j)->ndrivers-1)=i;
	      (associations+j)->directions=realloc((associations+j)->directions,sizeof(int)*(associations+j)->ndrivers);
	      if (val3>=0)
		*((associations+j)->directions+(associations+j)->ndrivers-1)=1;
	      else
		*((associations+j)->directions+(associations+j)->ndrivers-1)=-1;
	      (associations+j)->logratios=realloc((associations+j)->logratios,sizeof(double)*(associations+j)->ndrivers);
	      *((associations+j)->logratios+(associations+j)->ndrivers-1)=val1;
	      (associations+j)->pvals=realloc((associations+j)->pvals,sizeof(double)*(associations+j)->ndrivers);
	      *((associations+j)->pvals+(associations+j)->ndrivers-1)=val2;
	    }
	}
  
      // non-local methylation.
      // Only consider the methylations of regulators.
      // Only consider negative associations.
      // Exclude the associations of gender-specific methylation profiles.
      // Only consider the methylated genes with negative associations with their mRNA levels.
      // Only consider the candidate methylated genes with sufficient number of associations.
      for (n=0; n<=(nmethllist-1); n++)
	{
	  double val1, val2, val3, val4;
	  int chr1, chr2;
	  i=(int)(*(methllist+n*ncols+0)-1);
	  j=(int)(*(methllist+n*ncols+1)-1);
	  l=*(methl2gene+i);
	  val1=*(methllist+n*ncols+3);
	  val2=*(methllist+n*ncols+4);
	  val3=*(methllist+n*ncols+5);
	  chr1=(methlgenes+i)->chr;
	  if (mode==0)
	    chr2=(genes+j)->chr;
	  else
	    chr2=(mirs+j)->chr;	  
	  val4=new_missing_corrcoef(nsamples,methldata+i*nsamples,mrnadata+l*nsamples);
	 
	  if ((j>=startindex)&&(j<=endindex)&&(val1>=*(llthres+4))&&(val2<=*(pvalthres+4))&&(val3<=(*(corrthres+4)*(-1)))&&(val4<=(*(corrthres+0)*(-1)))&&(chr1!=chr2)&&(~(((chr1==23)||(chr1==24))&&((chr2==23)||(chr2==24))))&&(inlist(nselmethlgenes,selmethlgenes,i)==1)&&(*(gene2TF+l)>=0))
	    {	      
	      (associations+j)->ndrivers+=1;
	      (associations+j)->drivertypes=realloc((associations+j)->drivertypes,sizeof(char)*(associations+j)->ndrivers*LONGNAMELENGTH);
	      strncpy((associations+j)->drivertypes+((associations+j)->ndrivers-1)*LONGNAMELENGTH,"methylation",11);
	      *((associations+j)->drivertypes+((associations+j)->ndrivers-1)*LONGNAMELENGTH+11)='\0';
	      (associations+j)->driverinds=realloc((associations+j)->driverinds,sizeof(int)*(associations+j)->ndrivers);
	      *((associations+j)->driverinds+(associations+j)->ndrivers-1)=i;
	      (associations+j)->directions=realloc((associations+j)->directions,sizeof(int)*(associations+j)->ndrivers);
	      if (val3>=0)
		*((associations+j)->directions+(associations+j)->ndrivers-1)=1;
	      else
		*((associations+j)->directions+(associations+j)->ndrivers-1)=-1;
	      (associations+j)->logratios=realloc((associations+j)->logratios,sizeof(double)*(associations+j)->ndrivers);
	      *((associations+j)->logratios+(associations+j)->ndrivers-1)=val1;
	      (associations+j)->pvals=realloc((associations+j)->pvals,sizeof(double)*(associations+j)->ndrivers);
	      *((associations+j)->pvals+(associations+j)->ndrivers-1)=val2;
	    }
	}

      // non-local SNPs.
      // Consider the SNPs with strong cis-acting associations.
      // SNPs on the selected list (excluding local associations) are candidates for non-local associations.
      // Require the association with the intermediate genes of local SNP associations.
      // For instance, (snp1,gene1) is a candidate non-local association.  Then there exists another gene2 such that (snp1,gene2) is a local association and (gene1,gene2) are associated.
      
      // mRNA associations.
      if (mode==0)
	{
	  for (n=0; n<=(nsnplist-1); n++)
	    {
	      double val1, val2, val3;
	      i=(int)(*(snplist+n*ncols+0)-1);
	      j=(int)(*(snplist+n*ncols+1)-1);	      
	      val1=*(snplist+n*ncols+3);
	      val2=*(snplist+n*ncols+4);
	      val3=*(snplist+n*ncols+5);
	  
	      // Check whether SNP probe i and gene j are on distinct chromosomes or distant.
	      if ((j>=startindex)&&(j<=endindex)&&(val1>=*(llthres+5))&&(val2<=*(pvalthres+5))&&(fabs(val3)>=*(corrthres+5))&&(((snps+i)->chr!=(genes+j)->chr)||(fabs(((snps+i)->startcoord+(snps+i)->endcoord-(genes+j)->startcoord-(genes+j)->endcoord)/2)>1e6)))
		{
		  // Find the gene k with local association with SNP probe i.
		  // From the association results one SNP probe is associated with at most one local gene.
		  l=0; k=-1;
		  while ((l<=(nsnplist-1))&&(k<0))
		    {
		      if (((int)(*(snplist+l*ncols+0)-1)==i)&&(*(snplist+l*ncols+4)<=*(pvalthres+1))&&(*(snplist+l*ncols+3)>=*(llthres+1))&&(*(snplist+l*ncols+5)>=*(corrthres+1)))
			{
			  int j2=(int)(*(snplist+l*ncols+1)-1);
			  if (((snps+i)->chr==(genes+j2)->chr)&&(fabs(((snps+i)->startcoord+(snps+i)->endcoord-(genes+j2)->startcoord-(genes+j2)->endcoord)/2)<=1e6))
			    {
			      k=j2;			      
			    }

			}
		      l++;
		    }		

		  if (k>=0)
		    {
		      double *retvec;
		      l=0;
		      val=new_missing_corrcoef(nsamples,mrnadata+k*nsamples,mrnadata+j*nsamples);
		      if (val>=0)
			l=1;
		      else
			l=2;
		      retvec=ml_pairwise_model_corrected(nsamples,mrnadata+k*nsamples,mrnadata+j*nsamples,l,3);
		      
		      // Here the interemdiate regulators are genes instead of TFs.
		      if ((*(retvec+0)>=*(llthres+0))&&(*(retvec+1)<=*(pvalthres+0)))
			{
			  (associations+j)->ndrivers+=1;
			  (associations+j)->drivertypes=realloc((associations+j)->drivertypes,sizeof(char)*(associations+j)->ndrivers*LONGNAMELENGTH);
			  strncpy((associations+j)->drivertypes+((associations+j)->ndrivers-1)*LONGNAMELENGTH,"inter snp",9);
			  *((associations+j)->drivertypes+((associations+j)->ndrivers-1)*LONGNAMELENGTH+9)='\0';
			  (associations+j)->driverinds=realloc((associations+j)->driverinds,sizeof(int)*(associations+j)->ndrivers);
			  *((associations+j)->driverinds+(associations+j)->ndrivers-1)=*(snplistinds+n);
			  (associations+j)->directions=realloc((associations+j)->directions,sizeof(int)*(associations+j)->ndrivers);
			  if (val3>=0)
			    *((associations+j)->directions+(associations+j)->ndrivers-1)=1;
			  else
			    *((associations+j)->directions+(associations+j)->ndrivers-1)=-1;
			  (associations+j)->nregulators+=1;
			  (associations+j)->regulatorinds=realloc((associations+j)->regulatorinds,sizeof(int)*(associations+j)->nregulators);
			  *((associations+j)->regulatorinds+(associations+j)->nregulators-1)=k;	
			  
			  (associations+j)->logratios=realloc((associations+j)->logratios,sizeof(double)*(associations+j)->ndrivers);
			  *((associations+j)->logratios+(associations+j)->ndrivers-1)=val1;
			  (associations+j)->pvals=realloc((associations+j)->pvals,sizeof(double)*(associations+j)->ndrivers);
			  *((associations+j)->pvals+(associations+j)->ndrivers-1)=val2;
			 
			}
		      free(retvec);
		    }
		}
	    }	
	}

      // mir associations.
      // Intermediate genes are mRNA-coding genes, since SNPs are on mRNAs.
      else if (mode==1)
	{
	  for (n=0; n<=(nmirsnplist-1); n++)
	    {
	      double val1, val2, val3;
	      i=(int)(*(mirsnplist+n*ncols+0)-1);
	      j=(int)(*(mirsnplist+n*ncols+1)-1);	      
	      val1=*(mirsnplist+n*ncols+3);
	      val2=*(mirsnplist+n*ncols+4);
	      val3=*(mirsnplist+n*ncols+5);
	  
	      // Check whether SNP probe i and mir j are on distinct chromosomes or distant.
	      // Only consider negative associations.
	      if ((j>=startindex)&&(j<=endindex)&&(val1>=*(llthres+5))&&(val2<=*(pvalthres+5))&&(val3<=(*(corrthres+5)*(-1)))&&(((snps+i)->chr!=(mirs+j)->chr)||(fabs(((snps+i)->startcoord+(snps+i)->endcoord-(mirs+j)->startcoord-(mirs+j)->endcoord)/2)>1e6)))
		{
		  // Find the gene k with local association with SNP probe i.
		  // From the association results one SNP probe is associated with at most one local gene.
		  l=0; k=-1;
		  while ((l<=(nsnplist-1))&&(k<0))
		    {
		      if (((int)(*(snplist+l*ncols+0)-1)==i)&&(*(snplist+l*ncols+4)<=*(pvalthres+1))&&(*(snplist+l*ncols+3)>=*(llthres+1))&&(*(snplist+l*ncols+5)>=*(corrthres+1)))
			{
			  int j2=(int)(*(snplist+l*ncols+1)-1);
			  if (((snps+i)->chr==(genes+j2)->chr)&&(fabs(((snps+i)->startcoord+(snps+i)->endcoord-(genes+j2)->startcoord-(genes+j2)->endcoord)/2)<=1e6))
			    k=j2;
			}
		      l++;
		    }
		  if (k>=0)
		    {
		      double *retvec;
		      l=0;
		      val=new_missing_corrcoef(nsamples,mrnadata+k*nsamples,mirdata+j*nsamples);
		      if (val>=0)
			l=1;
		      else
			l=2;
		      retvec=ml_pairwise_model_corrected(nsamples,mrnadata+k*nsamples,mirdata+j*nsamples,l,3);

		      // Here the interemdiate regulator is in genes instead of TFs.
		      if ((*(retvec+0)>=*(llthres+0))&&(*(retvec+1)<=*(pvalthres+0)))		      
			{
			  (associations+j)->ndrivers+=1;
			  (associations+j)->drivertypes=realloc((associations+j)->drivertypes,sizeof(char)*(associations+j)->ndrivers*LONGNAMELENGTH);
			  strncpy((associations+j)->drivertypes+((associations+j)->ndrivers-1)*LONGNAMELENGTH,"inter snp",9);
			  *((associations+j)->drivertypes+((associations+j)->ndrivers-1)*LONGNAMELENGTH+9)='\0';
			  (associations+j)->driverinds=realloc((associations+j)->driverinds,sizeof(int)*(associations+j)->ndrivers);
			  *((associations+j)->driverinds+(associations+j)->ndrivers-1)=*(mirsnplistinds+n);
			  (associations+j)->directions=realloc((associations+j)->directions,sizeof(int)*(associations+j)->ndrivers);
			  if (val3>=0)
			    *((associations+j)->directions+(associations+j)->ndrivers-1)=1;
			  else
			    *((associations+j)->directions+(associations+j)->ndrivers-1)=-1;
			  (associations+j)->nregulators+=1;
			  (associations+j)->regulatorinds=realloc((associations+j)->regulatorinds,sizeof(int)*(associations+j)->nregulators);
			  *((associations+j)->regulatorinds+(associations+j)->nregulators-1)=k;			  
			  (associations+j)->logratios=realloc((associations+j)->logratios,sizeof(double)*(associations+j)->ndrivers);
			  *((associations+j)->logratios+(associations+j)->ndrivers-1)=val1;
			  (associations+j)->pvals=realloc((associations+j)->pvals,sizeof(double)*(associations+j)->ndrivers);
			  *((associations+j)->pvals+(associations+j)->ndrivers-1)=val2;
			}
		      free(retvec);
		    }
		}
	    }
	}
    }

  
  // Level 3: negative inter segment CNV.
  // Only consider negative associations with segments on selected chromosomes.
  // Only consider associations on other chromosomes.
  else if (level==3)
    {      
      // Debug
      //printf("interseglist\n");
      
      // inter segment CNV list.
      // If an external segment is associated with a gene and is on the same chromosome and the gene has intra segment CNV associations,  then label it as intra segment CNV.
      // If an external segment is on a different chromosome and associated with the gene, and there is at least one intermediate regulator, then label it as inter segment CNV.
      // Only consider the segments on selected chromosomes.
      for (n=0; n<=(ninterseglist-1); n++)
	{
	  int chr1, chr2;
	  double val1, val2, val3;
	  i=(int)(*(interseglist+n*ncols+0)-1);
	  j=(int)(*(interseglist+n*ncols+1)-1);
	  val1=*(interseglist+n*ncols+3);
	  val2=*(interseglist+n*ncols+4);
	  val3=*(interseglist+n*ncols+5);
	  chr1=(segments+i)->chr; 
	  if (mode==0)
	    chr2=(genes+j)->chr;
	  else
	    chr2=(mirs+j)->chr;
	  	    

	  //if ((j>=startindex)&&(j<=endindex)&&(val2<=pvalsegthre2))
	  if ((j>=startindex)&&(j<=endindex)&&(val1>=*(llthres+2))&&(val2<=*(pvalthres+2))&&(val3<=(*(corrthres+2)*(-1)))&&(chr1!=chr2)&&(inlist(nselchrs,selchrs,chr1)==1))
	    {
	      int flag2=0;
	      	      
	      // The association is not a candidate intra segment CNV.
	      if (flag2==0)
		{
		  int nregs, *regs;

		  // Check whether the segment accommodates an intermediate regulator.
		  // If mode=1, then look at preassociations2 instead of preassociations.
		  nregs=0; regs=malloc(sizeof(int)*nTFs);
		  for (k=0; k<=((geneTFs+j)->nmembers-1); k++)
		    {
		      l=*((geneTFs+j)->members1+k); l=*(TF2gene+l);
		      m=0; p=-1;

		      if (mode==0)
			{
			  while ((m<=((preassociations+l)->ndrivers-1))&&(p<0))
			    {
			      if ((strncmp((preassociations+l)->drivertypes+m*LONGNAMELENGTH,"intra segment CNV",17)==0)&&(*((preassociations+l)->driverinds+m)==i))
				p=m;
			      m++;
			    }
			}
		      else
			{
			  while ((m<=((preassociations2+l)->ndrivers-1))&&(p<0))
			    {
			      if ((strncmp((preassociations2+l)->drivertypes+m*LONGNAMELENGTH,"intra segment CNV",17)==0)&&(*((preassociations2+l)->driverinds+m)==i))
				p=m;
			      m++;
			    }
			}

		      if (p>=0)
			{
			  nregs++; *(regs+nregs-1)=*((geneTFs+j)->members1+k);
			}
		    }
		 
		  // Report associations with all intermediate TFs.
		  if (nregs>0)
		    {
		      int pp;
		      for (pp=0; pp<=(nregs-1); pp++)
			{
			  (associations+j)->ndrivers+=1;
			  (associations+j)->drivertypes=realloc((associations+j)->drivertypes,sizeof(char)*(associations+j)->ndrivers*LONGNAMELENGTH);
			  strncpy((associations+j)->drivertypes+((associations+j)->ndrivers-1)*LONGNAMELENGTH,"inter segment CNV",17);
			  *((associations+j)->drivertypes+((associations+j)->ndrivers-1)*LONGNAMELENGTH+17)='\0';
			  (associations+j)->driverinds=realloc((associations+j)->driverinds,sizeof(int)*(associations+j)->ndrivers);
			  *((associations+j)->driverinds+(associations+j)->ndrivers-1)=i;
			  (associations+j)->directions=realloc((associations+j)->directions,sizeof(int)*(associations+j)->ndrivers);
			  if (val3>=0)
			    *((associations+j)->directions+(associations+j)->ndrivers-1)=1;
			  else
			    *((associations+j)->directions+(associations+j)->ndrivers-1)=-1;
			  (associations+j)->nregulators+=1;
			  (associations+j)->regulatorinds=realloc((associations+j)->regulatorinds,sizeof(int)*(associations+j)->nregulators);
			  *((associations+j)->regulatorinds+(associations+j)->nregulators-1)=*(regs+pp);
			  (associations+j)->logratios=realloc((associations+j)->logratios,sizeof(double)*(associations+j)->ndrivers);
			  *((associations+j)->logratios+(associations+j)->ndrivers-1)=val1;
			  (associations+j)->pvals=realloc((associations+j)->pvals,sizeof(double)*(associations+j)->ndrivers);
			  *((associations+j)->pvals+(associations+j)->ndrivers-1)=val2;			  
			}
		    }
		  free(regs);
		}
	    }
	}		               	          
    }

  // Level 4: negative mirs that are not explained by lower level models.  Used only for mRNA associations.
  else if (level==4)
    {      
      // mRNA associations.
      if (mode==0)
	{
	  for (n=0; n<=(nmirlist-1); n++)
	    {
	      double val1, val2, val3;
	      i=(int)(*(mirlist+n*ncols+0)-1);
	      j=(int)(*(mirlist+n*ncols+1)-1);
	      val1=*(mirlist+n*ncols+3);
	      val2=*(mirlist+n*ncols+4);
	      val3=*(mirlist+n*ncols+5);

	      // Only consider the candidate mirs with sufficient number of associations.
	      if ((j>=startindex)&&(j<=endindex)&&(val1>=*(llthres+6))&&(val2<=*(pvalthres+6))&&(val3<=(*(corrthres+6)*(-1)))&&((preassociations2+i)->ndrivers<=0)&&(inlist(nselmirs,selmirs,i)==1))
		{
		  (associations+j)->ndrivers+=1;
		  (associations+j)->drivertypes=realloc((associations+j)->drivertypes,sizeof(char)*(associations+j)->ndrivers*LONGNAMELENGTH);
		  strncpy((associations+j)->drivertypes+((associations+j)->ndrivers-1)*LONGNAMELENGTH,"mir",3);
		  *((associations+j)->drivertypes+((associations+j)->ndrivers-1)*LONGNAMELENGTH+3)='\0';
		  (associations+j)->driverinds=realloc((associations+j)->driverinds,sizeof(int)*(associations+j)->ndrivers);
		  *((associations+j)->driverinds+(associations+j)->ndrivers-1)=i;
		  (associations+j)->directions=realloc((associations+j)->directions,sizeof(int)*(associations+j)->ndrivers);
		  if (val3>=0)
		    *((associations+j)->directions+(associations+j)->ndrivers-1)=1;
		  else
		    *((associations+j)->directions+(associations+j)->ndrivers-1)=-1;
		  (associations+j)->logratios=realloc((associations+j)->logratios,sizeof(double)*(associations+j)->ndrivers);
		  *((associations+j)->logratios+(associations+j)->ndrivers-1)=val1;
		  (associations+j)->pvals=realloc((associations+j)->pvals,sizeof(double)*(associations+j)->ndrivers);
		  *((associations+j)->pvals+(associations+j)->ndrivers-1)=val2;		    
		}
	    }
	}
    }
    
  // Level 5: inter segment CNV, non-local mutation for interactions.
  // Consider positive and negative associations with segments on all chromosomes.
  // Only consider associations on other chromosomes.
  // Choose the drivers with small marginal effects.
  else if (level==5)
    {  
      int *isselected, ind;

      // For each gene in the selected range, find the segment CNVs with marginal effects less than the thresholds.  Choose one representative from each chromosome.
      isselected=malloc(sizeof(int)*nsegments);      
      for (ind=startindex; ind<=endindex; ind++)
	{
	  int chr, nsubset, *subset, maxind;
	  double *scores, maxval;

	  for (i=0; i<=(nsegments-1); i++)
	    *(isselected+i)=1;
	  
	  for (n=0; n<=(nsegments-1); n++)
	    if ((genes+ind)->chr==(segments+n)->chr)
	      *(isselected+n)=0;

	  // Filter out the segments on the same chromosome of the passenger or pass the thresholds.
	  for (n=0; n<=(ninterseglist-1); n++)
	    {
	      int chr1, chr2;
	      double val1, val2, val3;
	      i=(int)(*(interseglist+n*ncols+0)-1);
	      j=(int)(*(interseglist+n*ncols+1)-1);
	      val1=*(interseglist+n*ncols+3);
	      val2=*(interseglist+n*ncols+4);
	      val3=*(interseglist+n*ncols+5);
	      chr1=(segments+i)->chr; 
	      if (mode==0)
		chr2=(genes+j)->chr;
	      else
		chr2=(mirs+j)->chr;
	      if ((j==ind)&&(val1>=*(llthres+2))&&(val2<=*(pvalthres+2))&&(fabs(val3)>=*(corrthres+2))&&(chr1!=chr2))
		*(isselected+i)=0;
	      else if (j==ind)
		*(isselected+i)=0;

	      // Filter out the segments with high correlations.
	      if ((j==ind)&&(*(isselected+i)==1)&&(fabs(val3)>=*(corrthres+9)))
		*(isselected+i)=0;	      
	    }

	  // For each chromosome, pick up a representative with the highest ll score.
	  subset=malloc(sizeof(int)*nsegments);
	  scores=malloc(sizeof(double)*nsegments);
	  for (n=0; n<=(nsegments-1); n++)
	    {
	      double *retvec;

	      *(scores+n)=-1;
	      if (*(isselected+n)==1)
		{
		  val=new_missing_corrcoef(nsamples,segcnvdata+n*nsamples,outdata+ind*nsamples);		  
		  *(scores+n)=val;

		  // Filter out the segments with high correlations.
		  if (fabs(*(scores+n))>=*(corrthres+9))
		    {
		      *(scores+n)=-1; *(isselected+n)=0;
		    }		  
		}
	    }
	  for (chr=1; chr<=24; chr++)
	    {
	      nsubset=0;
	      for (n=0; n<=(nsegments-1); n++)
		{
		  if (((segments+n)->chr==chr)&&(*(isselected+n)==1)&&(*(scores+n)>0))
		    {
		      nsubset++; *(subset+nsubset-1)=n;
		    }
		}
	      if (nsubset>0)
		{
		  maxind=-1; maxval=-1;
		  for (n=0; n<=(nsubset-1); n++)
		    {
		      if (*(scores+*(subset+n))>maxval)
			{
			  maxind=*(subset+n);
			  maxval=*(scores+*(subset+n));
			}
		    }

		  for (n=0; n<=(nsegments-1); n++)
		    if (((segments+n)->chr==chr)&&(*(isselected+n)==1)&&(n!=maxind))
		      *(isselected+n)=0;
		}	      
	      else
		{
		  for (n=0; n<=(nsegments-1); n++)
		    if (((segments+n)->chr==chr)&&(*(isselected+n)==1))
		      *(isselected+n)=0;
		}	      
	    }
	  free(subset); free(scores);
	 

	  // Add selected segment CNVs to the candidate association list.
	  for (n=0; n<=(nsegments-1); n++)
	    {
	      if (*(isselected+n)==1)
		{
		  double *retvec;

		  (associations+ind)->ndrivers+=1;
		  (associations+ind)->drivertypes=realloc((associations+ind)->drivertypes,sizeof(char)*(associations+ind)->ndrivers*LONGNAMELENGTH);
		  strncpy((associations+ind)->drivertypes+((associations+ind)->ndrivers-1)*LONGNAMELENGTH,"inter segment CNV",17);
		  *((associations+ind)->drivertypes+((associations+ind)->ndrivers-1)*LONGNAMELENGTH+17)='\0';
		  (associations+ind)->driverinds=realloc((associations+ind)->driverinds,sizeof(int)*(associations+ind)->ndrivers);
		  *((associations+ind)->driverinds+(associations+ind)->ndrivers-1)=n;
		  (associations+ind)->directions=realloc((associations+ind)->directions,sizeof(int)*(associations+ind)->ndrivers);
		  *((associations+ind)->directions+(associations+ind)->ndrivers-1)=1;
		  
		  retvec=malloc(sizeof(double)*2); *(retvec+0)=0; *(retvec+1)=1;

		  (associations+ind)->logratios=realloc((associations+ind)->logratios,sizeof(double)*(associations+ind)->ndrivers);   
		  *((associations+ind)->logratios+(associations+ind)->ndrivers-1)=*(retvec+0);
		  (associations+ind)->pvals=realloc((associations+ind)->pvals,sizeof(double)*(associations+ind)->ndrivers);
		  *((associations+ind)->pvals+(associations+ind)->ndrivers-1)=*(retvec+1);
		  free(retvec);
		}
	    }	  
	}
      
      // Include positive and negative mutation associations not on the list.
      // Only consider the genes with >= 10 mutations on the data.
      for (ind=startindex; ind<=endindex; ind++)
	{	 
	  for (i=0; i<=(nsegments-1); i++)
	    *(isselected+i)=0;

	  for (i=0; i<=(nmuts-1); i++)
	    *(isselected+i)=1;
	  
	  for (i=0; i<=(nmuts-1); i++)
	    if (*(mutvalid+i)==0)
	      *(isselected+i)=0;

	  // Filter out the mutations on the list.	  
	  for (n=0; n<=(nmutlist-1); n++)
	    {
	      double val1, val2, val3;
	      i=(int)(*(mutlist+n*ncols+0)-1);
	      j=(int)(*(mutlist+n*ncols+1)-1);
	      l=*(mut2gene+i);
	      val1=*(mutlist+n*ncols+3);
	      val2=*(mutlist+n*ncols+4);
	      val3=*(mutlist+n*ncols+5);	      
	      if ((j==ind)&&(val1>=*(llthres+3))&&(val2<=*(pvalthres+3))&&(fabs(val3)>=*(corrthres+3)))
		*(isselected+i)=0;
	      else if (l==ind)
		*(isselected+i)=0;
	    }

	  // Add selected mutated genes to the candidate association list.
	  for (n=0; n<=(nmuts-1); n++)
	    {
	      if (*(isselected+n)==1)
		{
		  double *retvec;

		  (associations+ind)->ndrivers+=1;
		  (associations+ind)->drivertypes=realloc((associations+ind)->drivertypes,sizeof(char)*(associations+ind)->ndrivers*LONGNAMELENGTH);
		  strncpy((associations+ind)->drivertypes+((associations+ind)->ndrivers-1)*LONGNAMELENGTH,"mutation",8);
		  *((associations+ind)->drivertypes+((associations+ind)->ndrivers-1)*LONGNAMELENGTH+8)='\0';
		  (associations+ind)->driverinds=realloc((associations+ind)->driverinds,sizeof(int)*(associations+ind)->ndrivers);
		  *((associations+ind)->driverinds+(associations+ind)->ndrivers-1)=n;
		  val=new_missing_corrcoef(nsamples,muttable+i*nsamples,outdata+ind*nsamples);
		  (associations+ind)->directions=realloc((associations+ind)->directions,sizeof(int)*(associations+ind)->ndrivers);
		  if (val>0)
		    *((associations+ind)->directions+(associations+ind)->ndrivers-1)=1;
		  else
		    *((associations+ind)->directions+(associations+ind)->ndrivers-1)=-1;
		  
		  retvec=malloc(sizeof(double)*2); *(retvec+0)=0; *(retvec+1)=1;

		  (associations+ind)->logratios=realloc((associations+ind)->logratios,sizeof(double)*(associations+ind)->ndrivers);		  		  
		  *((associations+ind)->logratios+(associations+ind)->ndrivers-1)=*(retvec+0);
		  (associations+ind)->pvals=realloc((associations+ind)->pvals,sizeof(double)*(associations+ind)->ndrivers);
		  *((associations+ind)->pvals+(associations+ind)->ndrivers-1)=*(retvec+1);
		  free(retvec);
		}
	    }	  
	}
      
      free(isselected);
    }


  // Integrate pairwise associations to additive models.
  // Check whether each aberration provides an additional explanatory power compared to the models established in the prior step.
  // For intra segment CNVs, identify the segment that gives the highest score.
  // For local SNPs, identify the SNP block that gives the highest score.
  // For inter segment CNVs, identify the representative segment on a chromosome.
  // For external methylation, identify the representative from a cluster.
  // For external mir, identify the representative from a cluster.
  // For external TF, identify the representative from a cluster.
  
  newassociations=malloc(sizeof(struct intassociation)*noutputs);
  for (geneind=0; geneind<=(noutputs-1); geneind++)
    {
      (newassociations+geneind)->ndrivers=0;
      (newassociations+geneind)->drivertypes=malloc(sizeof(char));
      (newassociations+geneind)->driverinds=malloc(sizeof(int));
      (newassociations+geneind)->directions=malloc(sizeof(int));
      (newassociations+geneind)->nregulators=0;
      (newassociations+geneind)->regulatorinds=malloc(sizeof(int));
      (newassociations+geneind)->ninteractions=0;
      (newassociations+geneind)->intdrivertypes=malloc(sizeof(char));
      (newassociations+geneind)->intdriverinds=malloc(sizeof(int));
      (newassociations+geneind)->intdirections=malloc(sizeof(int));
      (newassociations+geneind)->intregulatorinds=malloc(sizeof(int));   
      (newassociations+geneind)->intlogratios=malloc(sizeof(double));
      (newassociations+geneind)->intpvals=malloc(sizeof(double));            
      (newassociations+geneind)->npassengers=1;
      (newassociations+geneind)->passengertypes=malloc(sizeof(char)*LONGNAMELENGTH);
      if (mode==0)
	{
	  strncpy((newassociations+geneind)->passengertypes,"mrna",4);
	  *((newassociations+geneind)->passengertypes+4)='\0';
	}
      else
	{
	  strncpy((newassociations+geneind)->passengertypes,"mir",3);
	  *((newassociations+geneind)->passengertypes+3)='\0';
	}
      (newassociations+geneind)->passengerinds=malloc(sizeof(int));
      *((newassociations+geneind)->passengerinds+0)=geneind;
      (newassociations+geneind)->logratios=malloc(sizeof(double));
      (newassociations+geneind)->pvals=malloc(sizeof(double));
      (newassociations+geneind)->representatives=malloc(sizeof(int));
    }

  // Incorporate new associations to additive models.

  for (geneind=startindex; geneind<=endindex; geneind++)  
    {
      double *addlls, *addpvals;

      addlls=malloc(sizeof(double)*(associations+geneind)->ndrivers);
      addpvals=malloc(sizeof(double)*(associations+geneind)->ndrivers);

      // The gene contains pairwise associations at the current level.      
      if ((associations+geneind)->ndrivers>0)
	{
	  int *isselected, regind, tmpflag=1;	  

	  isselected=malloc(sizeof(int)*(associations+geneind)->ndrivers);
	  for (n=0; n<=((associations+geneind)->ndrivers-1); n++)
	    *(isselected+n)=1;
	  	  
	  
	  // Level 1: If a gene is associated with multiple intra segment CNVs, then choose the one with the strongest score.  Keep other associations that pass the filter.
	  // If a gene is associated with multiple cis-acting SNPs, then include all of them in level 1 models.  But when evaluating likelihood scores, only use the one with the strongest score.  All SNPs locally associated with a gene should be within the same LD block.
	  // Check whether an intra SNP provides additional explanatory power to other local aberrations.  If not then do not report the intra SNP associations.	  

	  if (level==1)
	    {
	      int ncandidates=0, *candidates, maxind;
	      double maxval;	      
	      
	      // Choose the intra segment CNV.
	      candidates=malloc(sizeof(int)*(associations+geneind)->ndrivers);
	      for (n=0; n<=((associations+geneind)->ndrivers-1); n++)
		{		  
		  if (strncmp((associations+geneind)->drivertypes+n*LONGNAMELENGTH,"intra segment CNV",17)==0)
		    {
		      ncandidates++; *(candidates+ncandidates-1)=n; *(isselected+n)=0;
		    }
		}
 
	      if (ncandidates>0)
		{
		  maxind=-1; maxval=-1e6;
		  for (n=0; n<=(ncandidates-1); n++)
		    {
		      if ((val=*((associations+geneind)->logratios+*(candidates+n)))>maxval)
			{
			  maxind=*(candidates+n); maxval=val;
			}
		    }
		  *(isselected+maxind)=1;
		}
	      

	      // Choose one representative for intra SNPs within the same LD block.
	      ncandidates=0;
	      for (n=0; n<=((associations+geneind)->ndrivers-1); n++)
		{
		  if (strncmp((associations+geneind)->drivertypes+n*LONGNAMELENGTH,"intra snp",9)==0)
		    {
		      ncandidates++; *(candidates+ncandidates-1)=n; *(isselected+n)=0;
		    }
		}

	      // Choose the one with the strongest score.
	      if (ncandidates>0)
		{
		  int maxind=-1;
		  double maxval=-1;
		  for (m=0; m<=(ncandidates-1); m++)
		    {
		      n=*(candidates+m); 
		      val=*((associations+geneind)->logratios+n);
		      if (val>maxval)
			{
			  maxind=n; maxval=val;
			}
		    }
		  if (maxind>=0)
		    *(isselected+maxind)=1;
		}
		      
	      // Build the models using local associations.
	      // Report all pairwise associations passing the thresholds but only use the representatives in model building.
	      (newassociations+geneind)->ndrivers=0;
	      for (n=0; n<=((associations+geneind)->ndrivers-1); n++)
		{
		  if (*(isselected+n)==1)
		    {
		      (newassociations+geneind)->ndrivers+=1;
		      (newassociations+geneind)->drivertypes=realloc((newassociations+geneind)->drivertypes,sizeof(char)*(newassociations+geneind)->ndrivers*LONGNAMELENGTH);
		      strncpy((newassociations+geneind)->drivertypes+((newassociations+geneind)->ndrivers-1)*LONGNAMELENGTH,(associations+geneind)->drivertypes+n*LONGNAMELENGTH,strlen((associations+geneind)->drivertypes+n*LONGNAMELENGTH));
		      *((newassociations+geneind)->drivertypes+((newassociations+geneind)->ndrivers-1)*LONGNAMELENGTH+strlen((associations+geneind)->drivertypes+n*LONGNAMELENGTH))='\0';
		      (newassociations+geneind)->driverinds=realloc((newassociations+geneind)->driverinds,sizeof(int)*(newassociations+geneind)->ndrivers);
		      *((newassociations+geneind)->driverinds+(newassociations+geneind)->ndrivers-1)=*((associations+geneind)->driverinds+n);
		      (newassociations+geneind)->directions=realloc((newassociations+geneind)->directions,sizeof(int)*(newassociations+geneind)->ndrivers);
		      *((newassociations+geneind)->directions+(newassociations+geneind)->ndrivers-1)=*((associations+geneind)->directions+n);
		      (newassociations+geneind)->logratios=realloc((newassociations+geneind)->logratios,sizeof(double)*(newassociations+geneind)->ndrivers);
		      *((newassociations+geneind)->logratios+(newassociations+geneind)->ndrivers-1)=*((associations+geneind)->logratios+n);
		      (newassociations+geneind)->pvals=realloc((newassociations+geneind)->pvals,sizeof(double)*(newassociations+geneind)->ndrivers);
		      *((newassociations+geneind)->pvals+(newassociations+geneind)->ndrivers-1)=*((associations+geneind)->pvals+n);
		      (newassociations+geneind)->representatives=realloc((newassociations+geneind)->representatives,sizeof(int)*(newassociations+geneind)->ndrivers);		      
		      if (*(isselected+n)==1)
			*((newassociations+geneind)->representatives+(newassociations+geneind)->ndrivers-1)=1;
		    }
		}	      	           

	      free(candidates); 
	      	     
	    }
	
	  // Levels 2-4: Perform various filters to selected associations.	
	  // Level 2 considers the trans-acting associations with representative SNPs from the candidate SNPs within the same LD blocks.	  
	  // Level 3 considers the trans-acting associations of opposite directions.
	  // Level 4 considers the mirs not explained by lower level associations.	  	  
	  else if ((level>=2)&&(level<=4))	  
	    {	
	      int noldvars, *oldnstates, *oldsigns, newnstates, newsign, *G, ncomps, ind;
	      int *pairtests, *overwhelm, *representatives, ind1, ind2, i1, i2;
	      double *olddata, *newvardata, *retvec, *tmpdata1, *tmpdata2;
	      char *tmpstr, *tmpstr2;
	      struct component *comps;
	      int tmpflag=0, intraind=-1;
	      int *preserved, flag;
	      	    
	      for (n=0; n<=((associations+geneind)->ndrivers-1); n++)
		{
		  *(isselected+n)=1; 
		}	      

	      // Filter 1: Choose representatives from each cluster or chromosome.
	      // For each external chromosome, choose the representative segment with the strongest score.
	      // Allow segments from multiple chromosomes co-appear in the model.
	      for (ind=1; ind<=24; ind++)
		{
		  int maxind;
		  double maxval;
		  if (ind!=(genes+geneind)->chr)
		    {
		      maxind=-1; maxval=-1;
		      for (n=0; n<=((associations+geneind)->ndrivers-1); n++)
			{
			  if (strncmp((associations+geneind)->drivertypes+n*LONGNAMELENGTH,"inter segment CNV",17)==0)
			    {
			      k=*((associations+geneind)->driverinds+n);
			      if ((segments+k)->chr==ind)
				{
				  *(isselected+n)=0;
				  val=*((associations+geneind)->logratios+n);
				  if (val>maxval)
				    {
				      maxind=n; maxval=val;
				    }
				}
			    }
			}
		      if (maxind>=0)
			*(isselected+maxind)=1;
		    }
		}


	      // Choose a representative for each cluster of methylated genes.
	      for (ind=0; ind<=(nmethlclus-1); ind++)
		{
		  int maxind=-1;
		  double maxval=-1;
		  for (n=0; n<=((associations+geneind)->ndrivers-1); n++)
		    {
		      if (strncmp((associations+geneind)->drivertypes+n*LONGNAMELENGTH,"methylation",11)==0)
			{
			  k=*((associations+geneind)->driverinds+n);
			  if (inlist((methlclus+ind)->nmembers,(methlclus+ind)->members,k)==1)
			    {
			      *(isselected+n)=0;
			       val=*((associations+geneind)->logratios+n);
			       if (val>maxval)
				 {
				   maxind=n; maxval=val;
				 }
			    }
			}
		    }
		  if (maxind>=0)
		    *(isselected+maxind)=1;
		}

	      // Choose a representative for each cluster of mirs.
	      for (ind=0; ind<=(nmirclus-1); ind++)
		{
		  int maxind=-1;
		  double maxval=-1;
		  for (n=0; n<=((associations+geneind)->ndrivers-1); n++)
		    {
		      if (strncmp((associations+geneind)->drivertypes+n*LONGNAMELENGTH,"mir",3)==0)
			{
			  k=*((associations+geneind)->driverinds+n);
			  if (inlist((mirclus+ind)->nmembers,(mirclus+ind)->members,k)==1)
			    {
			      *(isselected+n)=0;
			       val=*((associations+geneind)->logratios+n);
			       if (val>maxval)
				 {
				   maxind=n; maxval=val;
				 }
			    }
			}
		    }
		  if (maxind>=0)
		    *(isselected+maxind)=1;
		}
		  
	      // Choose a representative for each cluster of TFs.
	      for (ind=0; ind<=(nTFclus-1); ind++)
		{
		  int maxind=-1;
		  double maxval=-1;
		  for (n=0; n<=((associations+geneind)->ndrivers-1); n++)
		    {
		      if (strncmp((associations+geneind)->drivertypes+n*LONGNAMELENGTH,"TF",2)==0)
			{
			  k=*((associations+geneind)->driverinds+n);
			  if (inlist((TFclus+ind)->nmembers,(TFclus+ind)->members,k)==1)
			    {
			      *(isselected+n)=0;
			       val=*((associations+geneind)->logratios+n);
			       if (val>maxval)
				 {
				   maxind=n; maxval=val;
				 }
			    }
			}
		    }
		  if (maxind>=0)
		    *(isselected+maxind)=1;
		}
	 
	    	    	      
	      // Filter 2: Remove the drivers that do not provide additional explanatory power relative to the current model.
	      // If the gene is not explained by models in the prior level, then skip this filter.
	      if ((preassociations+geneind)->ndrivers>0)
		{
		  noldvars=0; oldnstates=malloc(sizeof(int)*((preassociations+geneind)->ndrivers+1));
		  oldsigns=malloc(sizeof(int)*(preassociations+geneind)->ndrivers);
		  olddata=malloc(sizeof(double)*((preassociations+geneind)->ndrivers+1)*nsamples);
		  for (n=0; n<=((preassociations+geneind)->ndrivers-1); n++)
		    {
		      if (*((preassociations+geneind)->representatives+n)==1)
			{		      
			  noldvars++; 
			  tmpstr=(preassociations+geneind)->drivertypes+n*LONGNAMELENGTH;
			  if (strncmp(tmpstr,"mutation",8)==0)
			    *(oldnstates+noldvars-1)=2;
			  else
			    *(oldnstates+noldvars-1)=3;
			  *(oldsigns+noldvars-1)=*((preassociations+geneind)->directions+n);
			  k=*((preassociations+geneind)->driverinds+n);
			  if ((strncmp(tmpstr,"intra segment CNV",17)==0)||(strncmp(tmpstr,"inter segment CNV",17)==0))
			    {
			      for (i=0; i<=(nsamples-1); i++)
				*(olddata+(noldvars-1)*nsamples+i)=*(segcnvdata+k*nsamples+i);
			    }
			  else if (strncmp(tmpstr,"mutation",8)==0)
			    {
			      for (i=0; i<=(nsamples-1); i++)
				*(olddata+(noldvars-1)*nsamples+i)=*(muttable+k*nsamples+i);
			    }
			  else if (strncmp(tmpstr,"methylation",11)==0)
			    {
			      for (i=0; i<=(nsamples-1); i++)
				*(olddata+(noldvars-1)*nsamples+i)=*(methldata+k*nsamples+i);
			    }
			  else if (strncmp(tmpstr,"mir",3)==0)
			    {
			      for (i=0; i<=(nsamples-1); i++)
				*(olddata+(noldvars-1)*nsamples+i)=*(mirdata+k*nsamples+i);
			    }
			  else if (strncmp(tmpstr,"TF",2)==0)
			    {
			      for (i=0; i<=(nsamples-1); i++)
				*(olddata+(noldvars-1)*nsamples+i)=*(TFdata+k*nsamples+i);
			    }
			  else if ((strncmp(tmpstr,"intra snp",9)==0)||(strncmp(tmpstr,"inter snp",9)==0))
			    {
			      // Convert the global driver index into the index of selected SNPs.
			      k=inlistind(nsnps,snpinds,k);

			      for (i=0; i<=(nsamples-1); i++)
				*(olddata+(noldvars-1)*nsamples+i)=*(snptable+k*nsamples+i);
			    }
			}
		    }
		  noldvars++; *(oldnstates+noldvars-1)=3;
		  for (i=0; i<=(nsamples-1); i++)
		    *(olddata+(noldvars-1)*nsamples+i)=*(outdata+geneind*nsamples+i);
						  
		  newvardata=malloc(sizeof(double)*nsamples);		  
		  for (n=0; n<=((associations+geneind)->ndrivers-1); n++)
		    {
		      if (*(isselected+n)==1)
			{
			  *(addlls+n)=0; *(addpvals+n)=1;
			  k=*((associations+geneind)->driverinds+n);
			  newsign=*((associations+geneind)->directions+n);
			  tmpstr=(associations+geneind)->drivertypes+n*LONGNAMELENGTH;
			  if (strncmp(tmpstr,"mutation",8)==0)
			    newnstates=2;
			  else
			    newnstates=3;
			  if ((strncmp(tmpstr,"intra segment CNV",17)==0)||(strncmp(tmpstr,"inter segment CNV",17)==0))
			    {
			      for (i=0; i<=(nsamples-1); i++)
				*(newvardata+i)=*(segcnvdata+k*nsamples+i);
			    }
			  else if (strncmp(tmpstr,"mutation",8)==0)
			    {
			      for (i=0; i<=(nsamples-1); i++)
				*(newvardata+i)=*(muttable+k*nsamples+i);
			    }
			  else if (strncmp(tmpstr,"methylation",11)==0)
			    {
			      for (i=0; i<=(nsamples-1); i++)
				*(newvardata+i)=*(methldata+k*nsamples+i);
			    }
			  else if (strncmp(tmpstr,"mir",3)==0)
			    {
			      for (i=0; i<=(nsamples-1); i++)
				*(newvardata+i)=*(mirdata+k*nsamples+i);
			    }
			  else if (strncmp(tmpstr,"TF",2)==0)
			    {
			      for (i=0; i<=(nsamples-1); i++)
				*(newvardata+i)=*(TFdata+k*nsamples+i);
			    }			  
			  else if ((strncmp(tmpstr,"intra snp",9)==0)||(strncmp(tmpstr,"inter snp",9)==0))
			    {
			      // Convert the global driver index into the index of selected SNPs.
			      k=inlistind(nsnps,snpinds,k);

			      for (i=0; i<=(nsamples-1); i++)
				*(newvardata+i)=*(snptable+k*nsamples+i);
			    }

			  retvec=additional_power_corrected(nsamples,noldvars,olddata,oldnstates,oldsigns,newvardata,newnstates,newsign);
			  
			  *(addlls+n)=*(retvec+0); *(addpvals+n)=*(retvec+1);
			  				  			  
		  
			  // Test whether the new feature has an additional power to fit the data.
			  if ((*(retvec+0)<*(llthres+8))||(*(retvec+1)>*(pvalthres+8)))
			    *(isselected+n)=0;
			  free(retvec);
			}
		    }	    
		  free(oldnstates); free(oldsigns); free(olddata); free(newvardata);
		}
	      else
		{
		  for (n=0; n<=((associations+geneind)->ndrivers-1); n++)
		    {
		      *(addlls+n)=*((associations+geneind)->logratios+n);
		      *(addpvals+n)=*((associations+geneind)->pvals+n);
		    }
		}
	    

	      // Filter 3: Remove the drivers overwhelmed by other drivers.	      
	      // For each driver pair (n,m), test the joint model n+m against n and m.
	      // pairtests[n][m]: test model (n+m) against model m.
	      pairtests=malloc(sizeof(int)*((associations+geneind)->ndrivers*(associations+geneind)->ndrivers));
	      noldvars=2; oldnstates=malloc(sizeof(int)*2); oldsigns=malloc(sizeof(int));
	      olddata=malloc(sizeof(double)*2*nsamples);
	      for (i=0; i<=(nsamples-1); i++)
		*(olddata+nsamples+i)=*(outdata+geneind*nsamples+i);
	      newvardata=malloc(sizeof(double)*nsamples);
	      *(oldnstates+1)=3;
	      for (n=0; n<=((associations+geneind)->ndrivers*(associations+geneind)->ndrivers-1); n++)
		*(pairtests+n)=0;
	      for (n=0; n<=((associations+geneind)->ndrivers-1); n++)
		{
		  for (m=0; m<=((associations+geneind)->ndrivers-1); m++)
		    {
		      if ((n!=m)&&(*(isselected+n)==1)&&(*(isselected+m)==1))
			{
			  int i1, i2;
			  char *tmpstr, *tmpstr2;
			  i1=*((associations+geneind)->driverinds+n);
			  i2=*((associations+geneind)->driverinds+m);
			  tmpstr=(associations+geneind)->drivertypes+n*LONGNAMELENGTH;
			  tmpstr2=(associations+geneind)->drivertypes+m*LONGNAMELENGTH;
			  *(oldsigns+0)=*((associations+geneind)->directions+n);
			  newsign=*((associations+geneind)->directions+m);
			  if (strncmp(tmpstr,"mutation",8)==0)
			    *(oldnstates+0)=2;
			  else
			    *(oldnstates+0)=3;
			  if (strncmp(tmpstr2,"mutation",8)==0)
			    newnstates=2;
			  else
			    newnstates=3;
			  if ((strncmp(tmpstr,"intra segment CNV",17)==0)||(strncmp(tmpstr,"inter segment CNV",17)==0))
			    {				      
			      for (i=0; i<=(nsamples-1); i++)
				{
				  *(olddata+i)=*(segcnvdata+i1*nsamples+i);
				}		      
			    }
			  else if (strncmp(tmpstr,"mutation",8)==0)
			    {
			      for (i=0; i<=(nsamples-1); i++)
				{
				  *(olddata+i)=*(muttable+i1*nsamples+i);
				}		      
			    }
			  else if (strncmp(tmpstr,"methylation",11)==0)
			    {
			      for (i=0; i<=(nsamples-1); i++)
				{
				  *(olddata+i)=*(methldata+i1*nsamples+i);
				}		      
			    }
			  else if (strncmp(tmpstr,"mir",3)==0)
			    {
			      for (i=0; i<=(nsamples-1); i++)
				{
				  *(olddata+i)=*(mirdata+i1*nsamples+i);
				}		      
			    }
			  else if (strncmp(tmpstr,"TF",2)==0)
			    {
			      for (i=0; i<=(nsamples-1); i++)
				{
				  *(olddata+i)=*(TFdata+i1*nsamples+i);
				}		      
			    }
			  else if ((strncmp(tmpstr,"intra snp",9)==0)||(strncmp(tmpstr,"inter snp",9)==0))
			    {
			      // Convert the global driver index into the index of selected SNPs.
			      i1=inlistind(nsnps,snpinds,i1);
			      
			      for (i=0; i<=(nsamples-1); i++)
				{
				  *(olddata+i)=*(snptable+i1*nsamples+i);
				}		      
			    }
			  
			  if ((strncmp(tmpstr2,"intra segment CNV",17)==0)||(strncmp(tmpstr2,"inter segment CNV",17)==0))
			    {				      
			      for (i=0; i<=(nsamples-1); i++)
				{
				  *(newvardata+i)=*(segcnvdata+i2*nsamples+i);
				}		      
			    }
			  else if (strncmp(tmpstr2,"mutation",8)==0)
			    {
			      for (i=0; i<=(nsamples-1); i++)
				{
				  *(newvardata+i)=*(muttable+i2*nsamples+i);
				}		      
			    }
			  else if (strncmp(tmpstr2,"methylation",11)==0)
			    {
			      for (i=0; i<=(nsamples-1); i++)
				{
				  *(newvardata+i)=*(methldata+i2*nsamples+i);
				}		      
			    }
			  else if (strncmp(tmpstr2,"mir",3)==0)
			    {
			      for (i=0; i<=(nsamples-1); i++)
				{
				  *(newvardata+i)=*(mirdata+i2*nsamples+i);
				}		      
			    }
			  else if (strncmp(tmpstr2,"TF",2)==0)
			    {
			      for (i=0; i<=(nsamples-1); i++)
				{
				  *(newvardata+i)=*(TFdata+i2*nsamples+i);
				}		      
			    }
			  else if ((strncmp(tmpstr2,"intra snp",9)==0)||(strncmp(tmpstr2,"inter snp",9)==0))
			    {
			      // Convert the global driver index into the index of selected SNPs.
			      i2=inlistind(nsnps,snpinds,i2);
			      
			      for (i=0; i<=(nsamples-1); i++)
				{
				  *(newvardata+i)=*(snptable+i2*nsamples+i);
				}		      
			    }
			  retvec=additional_power_corrected(nsamples,noldvars,olddata,oldnstates,oldsigns,newvardata,newnstates,newsign);
			  if ((*(retvec+0)>=*(llthres+8))&&(*(retvec+1)<=*(pvalthres+8)))
			    *(pairtests+m*(associations+geneind)->ndrivers+n)=1; 
			  
			}
		    }
		}
	      free(oldnstates); free(oldsigns); free(olddata); free(newvardata);
	    	      
	    	    
	      // Build the overwhelm graph from triplet test results.
	      // overwhelm[n][m]=1: n+m>m, n+m<=n.  n overwhelms m.
	      // overwhelm[m][n]=1: n+m>n, n+m<=m.  m overwhelms n.
	      // overwhel[n][m]=1, overwhelm[m][n]=1: n+m<=m, n+m<=n.
	      overwhelm=malloc(sizeof(int)*((associations+geneind)->ndrivers*(associations+geneind)->ndrivers));
	      for (n=0; n<=((associations+geneind)->ndrivers*(associations+geneind)->ndrivers-1); n++)
		*(overwhelm+n)=0;
	      for (n=0; n<=((associations+geneind)->ndrivers-1); n++)
		{
		  for (m=0; m<=((associations+geneind)->ndrivers-1); m++)
		    {
		      if ((n!=m)&&(*(isselected+n)==1)&&(*(isselected+m)==1))
			{
			  if ((*(pairtests+n*(associations+geneind)->ndrivers+m)==1)&&(*(pairtests+m*(associations+geneind)->ndrivers+n)==0))
			    *(overwhelm+n*(associations+geneind)->ndrivers+m)=1;
			  else if ((*(pairtests+n*(associations+geneind)->ndrivers+m)==0)&&(*(pairtests+m*(associations+geneind)->ndrivers+n)==1))
			    *(overwhelm+m*(associations+geneind)->ndrivers+n)=1;
			  else if ((*(pairtests+n*(associations+geneind)->ndrivers+m)==1)&&(*(pairtests+m*(associations+geneind)->ndrivers+n)==1))
			    {
			      *(overwhelm+n*(associations+geneind)->ndrivers+m)=1;
			      *(overwhelm+m*(associations+geneind)->ndrivers+n)=1;
			    }			 
			}
		    }
		}

	      
	      // Build a directed graph with overwhelm relations.
	      // Find the drivers that are not overwhelmed by any other drivers.
	      // Keep these upstream drivers.
	      // Iteratively augment the preserved drivers by adding the drivers that have bidirectional links or undidirectional liks to the preserved drivers.
	      preserved=malloc(sizeof(int)*(associations+geneind)->ndrivers);
	      for (n=0; n<=((associations+geneind)->ndrivers-1); n++)
		{
		  *(preserved+n)=0;
		  if (*(isselected+n)==1)
		    {		     
		      flag=1; m=0;
		      while ((m<=((associations+geneind)->ndrivers-1))&&(flag==1))
			{
			  if ((n!=m)&&(*(isselected+m)==1)&&(*(overwhelm+m*(associations+geneind)->ndrivers+n)==1)&&(*(overwhelm+n*(associations+geneind)->ndrivers+m)==0))
			    flag=0;
			  m++;
			}
		      if (flag==1)
			*(preserved+n)=1;
		    }
		}
	      
	      // Augment the bidirectional and unidirectional parents of drivers in the current set.
	      flag=1;
	      while (flag==1)
		{
		  flag=0;
		  for (n=0; n<=((associations+geneind)->ndrivers-1); n++)
		    {
		      if ((*(isselected+n)==1)&&(*(preserved+n)==0))
			{
			  int flag2=1;
			  m=0;
			  while ((m<=((associations+geneind)->ndrivers-1))&&(flag2==1))
			    {
			      if ((m!=n)&&(*(isselected+m)==1)&&(*(preserved+m)==1)&&(*(overwhelm+n*(associations+geneind)->ndrivers+m)==1))
				flag2=0;
			      m++;
			    }
			  if (flag2==0)
			    {
			      *(preserved+n)=1;
			      flag=1;
			    }
			}
		    }
		}

	      for (n=0; n<=((associations+geneind)->ndrivers-1); n++)
		if ((*(isselected+n)==1)&&(*(preserved+n)==0))
		  *(isselected+n)=0;

	      free(preserved);


	      free(pairtests); free(overwhelm);
	      
	      // Include the associations from the prior levels.
	      (newassociations+geneind)->ndrivers=0;	      
	      for (n=0; n<=((preassociations+geneind)->ndrivers-1); n++)
		{
		  (newassociations+geneind)->ndrivers+=1;
		  (newassociations+geneind)->drivertypes=realloc((newassociations+geneind)->drivertypes,sizeof(char)*(newassociations+geneind)->ndrivers*LONGNAMELENGTH);
		  strncpy((newassociations+geneind)->drivertypes+((newassociations+geneind)->ndrivers-1)*LONGNAMELENGTH,(preassociations+geneind)->drivertypes+n*LONGNAMELENGTH,strlen((preassociations+geneind)->drivertypes+n*LONGNAMELENGTH));
		  *((newassociations+geneind)->drivertypes+((newassociations+geneind)->ndrivers-1)*LONGNAMELENGTH+strlen((preassociations+geneind)->drivertypes+n*LONGNAMELENGTH))='\0';
		  (newassociations+geneind)->driverinds=realloc((newassociations+geneind)->driverinds,sizeof(int)*(newassociations+geneind)->ndrivers);
		  *((newassociations+geneind)->driverinds+(newassociations+geneind)->ndrivers-1)=*((preassociations+geneind)->driverinds+n);
		  (newassociations+geneind)->directions=realloc((newassociations+geneind)->directions,sizeof(int)*(newassociations+geneind)->ndrivers);
		  *((newassociations+geneind)->directions+(newassociations+geneind)->ndrivers-1)=*((preassociations+geneind)->directions+n);
		  (newassociations+geneind)->logratios=realloc((newassociations+geneind)->logratios,sizeof(double)*(newassociations+geneind)->ndrivers);
		  *((newassociations+geneind)->logratios+(newassociations+geneind)->ndrivers-1)=*((preassociations+geneind)->logratios+n);
		  (newassociations+geneind)->pvals=realloc((newassociations+geneind)->pvals,sizeof(double)*(newassociations+geneind)->ndrivers);
		  *((newassociations+geneind)->pvals+(newassociations+geneind)->ndrivers-1)=*((preassociations+geneind)->pvals+n);
		  (newassociations+geneind)->representatives=realloc((newassociations+geneind)->representatives,sizeof(int)*(newassociations+geneind)->ndrivers);		      
		  *((newassociations+geneind)->representatives+(newassociations+geneind)->ndrivers-1)=*((preassociations+geneind)->representatives+n);		  
		}
	      
	      (newassociations+geneind)->nregulators=0;
	      for (n=0; n<=((preassociations+geneind)->nregulators-1); n++)
		{		 
		  (newassociations+geneind)->nregulators+=1;
		  (newassociations+geneind)->regulatorinds=realloc((newassociations+geneind)->regulatorinds,sizeof(int)*(newassociations+geneind)->nregulators);
		  *((newassociations+geneind)->regulatorinds+(newassociations+geneind)->nregulators-1)=*((preassociations+geneind)->regulatorinds+n);
		}
	      
	      
	      // Include the selected associations in the new model.
	      // The llratios and pvalues are the scores of the additional power test. 
	      // Also have to include the regulator indices.
	      regind=-1;
	      for (n=0; n<=((associations+geneind)->ndrivers-1); n++)
		{
		  if (strncmp((associations+geneind)->drivertypes+n*LONGNAMELENGTH,"inter segment CNV",17)==0)
		    regind++;
		  
		  if (strncmp((associations+geneind)->drivertypes+n*LONGNAMELENGTH,"inter snp",9)==0)
		    regind++;
		  
		  if (*(isselected+n)==1)
		    {		  
		      (newassociations+geneind)->ndrivers+=1;
		      (newassociations+geneind)->drivertypes=realloc((newassociations+geneind)->drivertypes,sizeof(char)*(newassociations+geneind)->ndrivers*LONGNAMELENGTH);
		      strncpy((newassociations+geneind)->drivertypes+((newassociations+geneind)->ndrivers-1)*LONGNAMELENGTH,(associations+geneind)->drivertypes+n*LONGNAMELENGTH,strlen((associations+geneind)->drivertypes+n*LONGNAMELENGTH));
		      *((newassociations+geneind)->drivertypes+((newassociations+geneind)->ndrivers-1)*LONGNAMELENGTH+strlen((associations+geneind)->drivertypes+n*LONGNAMELENGTH))='\0';
		      
		      (newassociations+geneind)->driverinds=realloc((newassociations+geneind)->driverinds,sizeof(int)*(newassociations+geneind)->ndrivers);
		      *((newassociations+geneind)->driverinds+(newassociations+geneind)->ndrivers-1)=*((associations+geneind)->driverinds+n);
		      
		      (newassociations+geneind)->directions=realloc((newassociations+geneind)->directions,sizeof(int)*(newassociations+geneind)->ndrivers);
		      *((newassociations+geneind)->directions+(newassociations+geneind)->ndrivers-1)=*((associations+geneind)->directions+n);
		      
		      (newassociations+geneind)->logratios=realloc((newassociations+geneind)->logratios,sizeof(double)*(newassociations+geneind)->ndrivers);
		      *((newassociations+geneind)->logratios+(newassociations+geneind)->ndrivers-1)=*(addlls+n);
		      
		      (newassociations+geneind)->pvals=realloc((newassociations+geneind)->pvals,sizeof(double)*(newassociations+geneind)->ndrivers);
		      *((newassociations+geneind)->pvals+(newassociations+geneind)->ndrivers-1)=*(addpvals+n);
		      
		      (newassociations+geneind)->representatives=realloc((newassociations+geneind)->representatives,sizeof(int)*(newassociations+geneind)->ndrivers);		      
		      *((newassociations+geneind)->representatives+(newassociations+geneind)->ndrivers-1)=1;		      
		      
		      if (strncmp((associations+geneind)->drivertypes+n*LONGNAMELENGTH,"inter segment CNV",17)==0)
			{
			  (newassociations+geneind)->nregulators+=1;
			  (newassociations+geneind)->regulatorinds=realloc((newassociations+geneind)->regulatorinds,sizeof(int)*(newassociations+geneind)->nregulators);
			  *((newassociations+geneind)->regulatorinds+(newassociations+geneind)->nregulators-1)=*((associations+geneind)->regulatorinds+regind);			  
			}
		      
		      if (strncmp((associations+geneind)->drivertypes+n*LONGNAMELENGTH,"inter snp",9)==0)
			{
			  (newassociations+geneind)->nregulators+=1;
			  (newassociations+geneind)->regulatorinds=realloc((newassociations+geneind)->regulatorinds,sizeof(int)*(newassociations+geneind)->nregulators);
			  *((newassociations+geneind)->regulatorinds+(newassociations+geneind)->nregulators-1)=*((associations+geneind)->regulatorinds+regind);			  
			}		      
		    }
		}
	      	    
	      free(isselected);
	    }
	}	  

      // The gene does not contain pairwise associations at the current level.
      // Copy the associations at the prior level to the new associations.
      else 
	{
	  if (level>1)
	    {
	      (newassociations+geneind)->ndrivers=0;	      
	      for (n=0; n<=((preassociations+geneind)->ndrivers-1); n++)
		{
		  (newassociations+geneind)->ndrivers+=1;
		  (newassociations+geneind)->drivertypes=realloc((newassociations+geneind)->drivertypes,sizeof(char)*(newassociations+geneind)->ndrivers*LONGNAMELENGTH);
		  strncpy((newassociations+geneind)->drivertypes+((newassociations+geneind)->ndrivers-1)*LONGNAMELENGTH,(preassociations+geneind)->drivertypes+n*LONGNAMELENGTH,strlen((preassociations+geneind)->drivertypes+n*LONGNAMELENGTH));
		  *((newassociations+geneind)->drivertypes+((newassociations+geneind)->ndrivers-1)*LONGNAMELENGTH+strlen((preassociations+geneind)->drivertypes+n*LONGNAMELENGTH))='\0';
		  (newassociations+geneind)->driverinds=realloc((newassociations+geneind)->driverinds,sizeof(int)*(newassociations+geneind)->ndrivers);
		  *((newassociations+geneind)->driverinds+(newassociations+geneind)->ndrivers-1)=*((preassociations+geneind)->driverinds+n);
		  (newassociations+geneind)->directions=realloc((newassociations+geneind)->directions,sizeof(int)*(newassociations+geneind)->ndrivers);
		  *((newassociations+geneind)->directions+(newassociations+geneind)->ndrivers-1)=*((preassociations+geneind)->directions+n);
		  (newassociations+geneind)->logratios=realloc((newassociations+geneind)->logratios,sizeof(double)*(newassociations+geneind)->ndrivers);
		  *((newassociations+geneind)->logratios+(newassociations+geneind)->ndrivers-1)=*((preassociations+geneind)->logratios+n);
		  (newassociations+geneind)->pvals=realloc((newassociations+geneind)->pvals,sizeof(double)*(newassociations+geneind)->ndrivers);
		  *((newassociations+geneind)->pvals+(newassociations+geneind)->ndrivers-1)=*((preassociations+geneind)->pvals+n);
		  (newassociations+geneind)->representatives=realloc((newassociations+geneind)->representatives,sizeof(int)*(newassociations+geneind)->ndrivers);		      
		  *((newassociations+geneind)->representatives+(newassociations+geneind)->ndrivers-1)=*((preassociations+geneind)->representatives+n);		  
		}
	      
	      (newassociations+geneind)->nregulators=0;
	      for (n=0; n<=((preassociations+geneind)->nregulators-1); n++)
		{		 
		  (newassociations+geneind)->nregulators+=1;
		  (newassociations+geneind)->regulatorinds=realloc((newassociations+geneind)->regulatorinds,sizeof(int)*(newassociations+geneind)->nregulators);
		  *((newassociations+geneind)->regulatorinds+(newassociations+geneind)->nregulators-1)=*((preassociations+geneind)->regulatorinds+n);
		}
	    }
	}

      free(addlls); free(addpvals);
    }


  // Level 5: Combinatorial interactions.
  // Consider the combinatorial interactions of external segment CNVs and mutations.
  // Consider the drivers that have small marginal effects but jointly achieve better explanatory power.
  if (level==5)
    {
      for (geneind=startindex; geneind<=endindex; geneind++)  
	{	  

	  // The gene contains pairwise associations at the current level.      
	  if ((associations+geneind)->ndrivers>0)
	    {
	      // For each pair of drivers, evaluate their interaction score.
	      for (n=0; n<=((associations+geneind)->ndrivers-2); n++)
		{
		  for (m=(n+1); m<=((associations+geneind)->ndrivers-1); m++)
		    {
		      if ((strncmp((associations+geneind)->drivertypes+n*LONGNAMELENGTH,"inter segment CNV",17)==0)||(strncmp((associations+geneind)->drivertypes+m*LONGNAMELENGTH,"inter segment CNV",17)==0))
			{		      
			  double *vec1, *vec2, *vec3, *retvec;
			  int n1, n2;
			  
			  vec3=outdata+geneind*nsamples;
			  if (strncmp((associations+geneind)->drivertypes+n*LONGNAMELENGTH,"inter segment CNV",17)==0)
			    {
			      k=*((associations+geneind)->driverinds+n);
			      vec1=segcnvdata+k*nsamples; n1=3;
			    }
			  else
			    {
			      k=*((associations+geneind)->driverinds+n);
			      vec1=muttable+k*nsamples; n1=2;
			    }
			  if (strncmp((associations+geneind)->drivertypes+m*LONGNAMELENGTH,"inter segment CNV",17)==0)
			    {
			      k=*((associations+geneind)->driverinds+m);
			      vec2=segcnvdata+k*nsamples; n2=3;
			    }
			  else
			    {
			      k=*((associations+geneind)->driverinds+m);
			      vec2=muttable+k*nsamples; n2=2;
			    }
			  retvec=ml_pairwise_interaction_model_corrected(nsamples,vec1,vec2,vec3,n1,n2,&funcsign);
			  			  
			  // Add the interaction term if the score passes the threshold.
			  if ((*(retvec+0)>=*(llthres+9))&&(*(retvec+1)<=*(pvalthres+9)))
			    {			      			      
			      (newassociations+geneind)->ninteractions+=1;
			      (newassociations+geneind)->intdrivertypes=realloc((newassociations+geneind)->intdrivertypes,sizeof(char)*(newassociations+geneind)->ninteractions*2*LONGNAMELENGTH);
			      strncpy((newassociations+geneind)->intdrivertypes+(((newassociations+geneind)->ninteractions-1)*2+0)*LONGNAMELENGTH,(associations+geneind)->drivertypes+n*LONGNAMELENGTH,strlen((associations+geneind)->drivertypes+n*LONGNAMELENGTH));
			      *((newassociations+geneind)->intdrivertypes+(((newassociations+geneind)->ninteractions-1)*2+0)*LONGNAMELENGTH+strlen((associations+geneind)->drivertypes+n*LONGNAMELENGTH))='\0';
			      strncpy((newassociations+geneind)->intdrivertypes+(((newassociations+geneind)->ninteractions-1)*2+1)*LONGNAMELENGTH,(associations+geneind)->drivertypes+m*LONGNAMELENGTH,strlen((associations+geneind)->drivertypes+m*LONGNAMELENGTH));
			      *((newassociations+geneind)->intdrivertypes+(((newassociations+geneind)->ninteractions-1)*2+1)*LONGNAMELENGTH+strlen((associations+geneind)->drivertypes+m*LONGNAMELENGTH))='\0';
			      (newassociations+geneind)->intdriverinds=realloc((newassociations+geneind)->intdriverinds,sizeof(int)*(newassociations+geneind)->ninteractions*2);
			      *((newassociations+geneind)->intdriverinds+((newassociations+geneind)->ninteractions-1)*2+0)=*((associations+geneind)->driverinds+n);
			      *((newassociations+geneind)->intdriverinds+((newassociations+geneind)->ninteractions-1)*2+1)=*((associations+geneind)->driverinds+m);
			      (newassociations+geneind)->intdirections=realloc((newassociations+geneind)->intdirections,sizeof(int)*(newassociations+geneind)->ninteractions);
			      *((newassociations+geneind)->intdirections+(newassociations+geneind)->ninteractions-1)=funcsign;			  
			      (newassociations+geneind)->intlogratios=realloc((newassociations+geneind)->intlogratios,sizeof(double)*(newassociations+geneind)->ninteractions);
			      *((newassociations+geneind)->intlogratios+(newassociations+geneind)->ninteractions-1)=*(retvec+0);
			      (newassociations+geneind)->intpvals=realloc((newassociations+geneind)->intpvals,sizeof(double)*(newassociations+geneind)->ninteractions);
			      *((newassociations+geneind)->intpvals+(newassociations+geneind)->ninteractions-1)=*(retvec+1);
			    }
			  free(retvec);
			}
		    }
		}
	    }
	}
    }

  
  // Generate the file name of the output file.
  // The file name is prefix + number.
  // To facilitate file concatenation, force all numbers to have the same number of digits.
  // Fill 0s between the start and the index number.
  prefix=malloc(sizeof(char)*MAXLINEWIDTH);
  strncpy(prefix,*(argv+30),strlen(*(argv+30))); *(prefix+strlen(*(argv+30)))='\0';
  val=(double)curind;
  if (curind==0)
    k=1;
  else
    {
      val=log10(val); val=floor(val)+1; k=(int)val;
    }
  l=ndigits-k;
  filename=malloc(sizeof(char)*MAXLINEWIDTH);
  str=malloc(sizeof(char)*MAXLINEWIDTH);
  for (i=0; i<=(l-1); i++)
    *(str+i)='0';
  *(str+l)='\0';
  str2=malloc(sizeof(char)*MAXLINEWIDTH);
  sprintf(str2,"%d\0",curind); m=0;
  strncpy(filename,prefix,strlen(prefix)); *(filename+m+strlen(prefix))='\0'; m+=strlen(prefix);
  strncat(filename,str,strlen(str)); *(filename+m+strlen(str))='\0'; m+=strlen(str);
  strncat(filename,str2,strlen(str2)); *(filename+m+strlen(str2))='\0'; m+=strlen(str2);
        
	      
  // Report the models to the output file.  
  if ((fout=fopen(filename,"w"))==NULL)
    {
      printf("Cannot open %s.\n",filename);
      return -1;
    }

  for (geneind=startindex; geneind<=endindex; geneind++)
    {            
      // Report additive models.
      l=0;
      for (n=0; n<=((newassociations+geneind)->ndrivers-1); n++)
	{	 
	  wrote=1;
	  fprintf(fout,"%s%c%d%c",(newassociations+geneind)->drivertypes+n*LONGNAMELENGTH,0x09,*((newassociations+geneind)->driverinds+n)+1,0x09);
	  if ((strncmp((newassociations+geneind)->drivertypes+n*LONGNAMELENGTH,"inter segment CNV",17)==0)||(strncmp((newassociations+geneind)->drivertypes+n*LONGNAMELENGTH,"inter snp",9)==0))
	    {	      
	      fprintf(fout,"%d%c",*((newassociations+geneind)->regulatorinds+l)+1,0x09); l++;
	    }
	  else
	    fprintf(fout,"-%c",0x09);
	  fprintf(fout,"%d%c%d%c%f%c%.4e%c%d\n",geneind+1,0x09,*((newassociations+geneind)->directions+n),0x09,*((newassociations+geneind)->logratios+n),0x09,*((newassociations+geneind)->pvals+n),0x09,*((newassociations+geneind)->representatives+n));

	}

      // Report interaction models.
      for (n=0; n<=((newassociations+geneind)->ninteractions-1); n++)
	{
	  wrote=1;
	  fprintf(fout,"interaction%c%s%c%d%c-%c",0x09,(newassociations+geneind)->intdrivertypes+(n*2)*LONGNAMELENGTH,0x09,*((newassociations+geneind)->intdriverinds+n*2)+1,0x09,0x09);	  
	  fprintf(fout,"%s%c%d%c-%c",(newassociations+geneind)->intdrivertypes+(n*2+1)*LONGNAMELENGTH,0x09,*((newassociations+geneind)->intdriverinds+n*2+1)+1,0x09,0x09);
	  fprintf(fout,"%d%c-%c",geneind+1,0x09,0x09);
	  fprintf(fout,"-%c%f%c%.4e%c%d\n",0x09,*((newassociations+geneind)->intlogratios+n),0x09,*((newassociations+geneind)->intpvals+n),0x09,1);	  
	}	  	  
    }
  fclose(fout);

  // If nothing is written on fout, then delete the file.
  if (wrote==0)
    {
      char *commandstr;
      commandstr=malloc(sizeof(char)*MAXLINEWIDTH);
      sprintf(commandstr,"rm %s\0",filename);
      system(commandstr);
      free(commandstr);
    }
    
  return 0;
}



