/* Compute the pairwise scores between a type of aberration data and mRNA expressions.
   Assume the tables of the all types of data have the same dimensions.  Genes and samples are also in the same order.
   Cluster version.  Specify the ranges of input data and mRNA expression data to evaluate associations.
   Encode the ranges as one number.
   2011/06/11: Add the associations with SNP data.
   The SNP association part does not work for a large number of SNPs since the total number of associations is ngenes*nsps.  Too many to feed into clusters.
   Difference from compute_pairwise_scores_cluster.c: correct the log likelihood evaluation when the null model is independent.  This will reduce the ostensibly inflated log likelihood values of poorly associated pairs.
*/

# include <stdio.h>
# include <stdlib.h>
# include <string.h>
# include <math.h>
# include <time.h>
# include "motif.h"
# define MAXLINEWIDTH 1000000
# define NAMELENGTH 30
# define LONGNAMELENGTH 100
# define max(i,j) (i>j ? i : j)
# define min(i,j) (i<j ? i : j)
# define argmax(i,j) (j>i ? 2 : 1)

extern int pair_cmp(struct pair *p1, struct pair *p2);
extern char * getitemval2(char *s, int ind, int *length);
extern char *i2a(long num, int *length);
extern double atof2(char *s, int length);
extern char * getitemval3(char *s, int ind, int *length, char sepch);
extern double * ml_pairwise_model(int nsamples, double *x, double *y, int mode);
extern double * ml_pairwise_model2(int nsamples, double *x, double *y, int mode);
extern double new_missing_corrcoef(int nsamples, double *x, double *y);
extern double * ml_pairwise_model_corrected(int nsamples, double *x, double *y, int mode, int nstates);

struct geneseginfo {
  char *name;
  char *snpname;
  int  chr;
  char *dir;
  long startcoord;
  long endcoord;
};

int nintervals=100;
int nmaxiterations=1000, nperm=10000;
double NAval=-1e6, epsilon=0.001, infval=1e20;
double high_thre=0.6, low_thre=0.4;
int ndigits=10;

int main(int argc, char *argv[])
{
  // *(argv+1) -- mode.
  // *(argv+2) -- input data.
  // *(argv+3) -- mRNA/mir expression data.
  // *(argv+4) -- correlation coefficient threshold.
  // *(argv+5) -- logratio threshold.
  // *(argv+6) -- pvalue threshold.    
  // *(argv+7) -- increment size of the range number.
  // *(argv+8) -- the number specifying the starting input index and expression index.
  // *(argv+9) -- prefix of output association scores file.
 
  int i, j, k, l, m, n, mode, ngenes, nsamples, length, ninputs, npairs, wrote=0;
  double val, val2, logratio, pval, *mrnadata, *inputdata;
  char *s, *item, ch, *samples, *prefix, *filename, *str, *str2;
  struct geneseginfo *inputs, *genes;  
  double corrthre, logratiothre, pvalthre, *scores, *retvec;
  long incsize, startind, startindex, inputindex, expindex, endindex, index;
  FILE *fp, *fout;
  time_t *tp;
  
  // Import the mode of the analysis.
  // 0: intra-segment CNVs. 
  // 1: inter-segment CNVs.
  // 2: mutations.
  // 3: DNA methylations.
  // 4: mir expressions.
  // 5: TF mRNA expressions.
  // Others: do not perform analysis.
  mode=atoi(*(argv+1));


  // Load input data.

  // Debug
  printf("Load input data\n");

  if ((fp=fopen(*(argv+2),"r"))==NULL)
    {
      printf("Cannot open %s.\n",*(argv+2));
      return -1;
    }
  s=malloc(sizeof(char)*MAXLINEWIDTH); 
  fgets(s,MAXLINEWIDTH,fp);
  nsamples=0; i=0; 
  while ((i<strlen(s))&&((ch=*(s+i))!=0x0a))
    {
      if (ch==0x09)
	nsamples++;
      i++;
    } 
  samples=malloc(sizeof(char)*nsamples*LONGNAMELENGTH);
  for (n=0; n<=(nsamples-1); n++)
    {
      item=getitemval3(s,n+1,&length,0x09);
      strncpy(samples+n*LONGNAMELENGTH,item,length);
      *(samples+n*LONGNAMELENGTH+length)='\0';
      free(item);
    }
  
  ninputs=0; inputs=malloc(sizeof(struct geneseginfo));
  inputdata=malloc(sizeof(double)*nsamples);
  while ((fgets(s,MAXLINEWIDTH,fp))!=NULL)
    {
      ninputs++; inputs=realloc(inputs,sizeof(struct geneseginfo)*ninputs);

      // Get the gene/segment info.

      // Mutations, methylations, mirs, TFs.
      if (mode>1)
	{
	  item=getitemval3(s,1,&length,'|');
	  val=atof2(item,length);
	  (inputs+ninputs-1)->chr=(int)val;
	  free(item);

	  item=getitemval3(s,2,&length,'|');
	  val=atof2(item,length);
	  (inputs+ninputs-1)->startcoord=(long)val;
	  free(item);
	  
	  item=getitemval3(s,3,&length,'|');
	  val=atof2(item,length);
	  (inputs+ninputs-1)->endcoord=(long)val;
	  free(item);
	}

      // Segment CNVs.
      else
	{	  
	  item=getitemval3(s,1,&length,'|');
	  val=atof2(item,length);
	  (inputs+ninputs-1)->chr=(int)val;
	  free(item);

	  item=getitemval3(s,2,&length,'|');
	  val=atof2(item,length);
	  (inputs+ninputs-1)->startcoord=(long)val;
	  free(item);
	  
	  item=getitemval3(s,3,&length,'|');
	  val=atof2(item,length);
	  (inputs+ninputs-1)->endcoord=(long)val;
	  free(item);
	}
      
      // Get the data.
      inputdata=realloc(inputdata,sizeof(double)*ninputs*nsamples);
      for (i=0; i<=(nsamples-1); i++)
	{
	  item=getitemval3(s,i+1,&length,0x09);	  
	  if (strncmp(item,"NaN",3)==0)
	    val=NAval;
	  else
	    val=atof2(item,length);	  
	  *(inputdata+(ninputs-1)*nsamples+i)=val;	  
	  free(item);
	}           
    }
  fclose(fp);

  
  // Convert the mutation data into binary.
  if (mode==2)
    {
      for (m=0; m<=(ninputs-1); m++)
	{
	  for (n=0; n<=(nsamples-1); n++)
	    {
	      val=*(inputdata+m*nsamples+n);	      
	      if (val>NAval)
		{
		  if (val>0)
		    *(inputdata+m*nsamples+n)=1;
		  else
		    *(inputdata+m*nsamples+n)=0;
		}
	      
	    }
	}
    }
		    

  // Load mRNA expression data.
  // Assume the genes and samples have the same order as input.

  // Debug
  printf("Load mRNA data\n");

  if ((fp=fopen(*(argv+3),"r"))==NULL)
    {
      printf("Cannot open %s.\n",*(argv+3));
      return -1;
    }
  s=malloc(sizeof(char)*MAXLINEWIDTH); 
  fgets(s,MAXLINEWIDTH,fp);  
  ngenes=0; genes=malloc(sizeof(struct geneseginfo));
  mrnadata=malloc(sizeof(double)*nsamples);
  while ((fgets(s,MAXLINEWIDTH,fp))!=NULL)
    {
      ngenes++; 
      
      genes=realloc(genes,sizeof(struct geneseginfo)*ngenes);
            
      // Get the gene info.
      
      item=getitemval3(s,1,&length,'|');
      val=atof2(item,length);
      (genes+ngenes-1)->chr=(int)val;
      free(item);
      
      item=getitemval3(s,2,&length,'|');
      val=atof2(item,length);
      (genes+ngenes-1)->startcoord=(long)val;
      free(item);
      
      item=getitemval3(s,3,&length,'|');
      val=atof2(item,length);
      (genes+ngenes-1)->endcoord=(long)val;
     

      free(item);	
      
    
      // Get the data.
      mrnadata=realloc(mrnadata,sizeof(double)*ngenes*nsamples);
      for (i=0; i<=(nsamples-1); i++)
	{
	  item=getitemval3(s,i+1,&length,0x09);	  
	  if (strncmp(item,"NaN",3)==0)
	    val=NAval;
	  else
	    val=atof2(item,length);	  
	  *(mrnadata+(ngenes-1)*nsamples+i)=val;	  
	  free(item);
	}           
    }
  fclose(fp);


  // Load threshold values of logratio and pvalue.
  corrthre=atof(*(argv+4)); logratiothre=atof(*(argv+5)); pvalthre=atof(*(argv+6));

  // Debug
  printf("ngenes=%d, ninputs=%d\n",ngenes,ninputs);
  
  // Load the increment size and start index.
  incsize=atoi(*(argv+7)); startind=atoi(*(argv+8));
  startindex=startind*incsize;

  endindex=startindex+incsize-1;

  
  // Generate the file name of the output file.
  // The file name is prefix + number.
  // To facilitate file concatenation, force all numbers to have the same number of digits.
  // Fill 0s between the start and the index number.
  prefix=malloc(sizeof(char)*MAXLINEWIDTH);
  strncpy(prefix,*(argv+9),strlen(*(argv+9))); *(prefix+strlen(*(argv+9)))='\0';
  val=(double)startind;
  if (startind==0)
    k=1;
  else
    {
      val=log10(val); val=floor(val)+1; k=(int)val;
    }
  l=ndigits-k;
  filename=malloc(sizeof(char)*MAXLINEWIDTH);
  str=malloc(sizeof(char)*MAXLINEWIDTH);
  for (i=0; i<=(l-1); i++)
    *(str+i)='0';
  *(str+l)='\0';
  str2=malloc(sizeof(char)*MAXLINEWIDTH);
  sprintf(str2,"%d\0",startind); m=0;
  strncpy(filename,prefix,strlen(prefix)); *(filename+m+strlen(prefix))='\0'; m+=strlen(prefix);
  strncat(filename,str,strlen(str)); *(filename+m+strlen(str))='\0'; m+=strlen(str);
  strncat(filename,str2,strlen(str2)); *(filename+m+strlen(str2))='\0'; m+=strlen(str2);
        

  // Report the association lists passing the thresholds.
  if ((fout=fopen(filename,"w"))==NULL)  
    {
      printf("Cannot open %s.\n",filename);
      return -1;
    }
  
  
  // Calculate the pairwise association scores.
  // Get the pairs specified in the range [startindex,endindex].

  tp=malloc(sizeof(time_t)); srand((int)(time(tp)));
  npairs=0; scores=malloc(sizeof(double)*6);
 

  for (index=startindex; index<=endindex; index++)
    {      
      
      // Get the indices of input and expression data.
      n=index%ngenes; m=(index-n)/ngenes;      

      if ((n<ngenes)&&(m<ninputs))      
	{
      
	  val=new_missing_corrcoef(nsamples,inputdata+m*nsamples,mrnadata+n*nsamples);
	  val2=fabs(val);
	  
	  if (val2>=corrthre)
	    {
	      int flag=0, funcmode=0, funcsign=0;	      	      
	      
	      // mode=0: only consider the genes on the same segments.
	      if ((mode==0)&&((inputs+m)->chr==(genes+n)->chr)&&((genes+n)->startcoord>=(inputs+m)->startcoord)&&((genes+n)->startcoord<=(inputs+m)->endcoord)&&(val>=corrthre))
		{
		  flag=1; funcmode=1; funcsign=1;
		}

	      else if ((mode==0)&&((inputs+m)->chr==(genes+n)->chr)&&((genes+n)->endcoord>=(inputs+m)->startcoord)&&((genes+n)->endcoord<=(inputs+m)->endcoord)&&(val>=corrthre))
		{
		  flag=1; funcmode=1; funcsign=1;
		}
	      
	      // mode=1: only consider the genes on different chromosomes.
	      else if ((mode==1)&&((inputs+m)->chr!=(genes+n)->chr))
		{
		  flag=1;
		  if (val>0)
		    {
		      funcmode=1; funcsign=1;
		    }
		  else
		    {
		      funcmode=2; funcsign=-1;
		    }
		}
	      
	      // Added after rescaling the cnv data, since the new segments give a better resolution.
	      // mode=1: consider the genes on different segments.
	      else if ((mode==1)&&((inputs+m)->chr==(genes+n)->chr)&&((((genes+n)->startcoord<(inputs+m)->startcoord)&&((genes+n)->endcoord<(inputs+m)->startcoord))||(((genes+n)->startcoord>(inputs+m)->endcoord)&&((genes+n)->endcoord>(inputs+m)->endcoord))))
		{
		  flag=1;
		  if (val>0)
		    {
		      funcmode=1; funcsign=1;
		    }
		  else
		    {
		      funcmode=2; funcsign=-1;
		    }
		}
				
	      // mutations: apply ml_pairwise_model2.
	      else if (mode==2)
		{
		  flag=2;
		  if (val>0)
		    {
		      funcmode=3; funcsign=1;
		    }
		  else
		    {
		      funcmode=4; funcsign=-1;
		    }
		}
	      
	      // other modes: get the proper signs.
	      else if (mode>=3)
		{
		  flag=1;
		  if (val>0)
		    {
		      funcmode=1; funcsign=1;
		    }
		  else
		    {
		      funcmode=2; funcsign=-1;
		    }
		}
	     
	      if (flag==1)	      
		{  		  
		  retvec=ml_pairwise_model_corrected(nsamples,inputdata+m*nsamples,mrnadata+n*nsamples,funcmode,3);

		  
		  if ((*(retvec+0)>=logratiothre)&&(*(retvec+1)<=pvalthre))
		    {		      	      		      		      
		      fprintf(fout,"%d%c%d%c%d%c%f%c%.4e%c%f\n",m+1,0x09,n+1,0x09,funcmode,0x09,*(retvec+0),0x09,*(retvec+1),0x09,val);
		      wrote=1;

		    }
		  free(retvec);
		}
	      
	      else if (flag==2)
		{  			  		  
		  retvec=ml_pairwise_model_corrected(nsamples,inputdata+m*nsamples,mrnadata+n*nsamples,funcmode,2);
		  
		  if ((*(retvec+0)>=logratiothre)&&(*(retvec+1)<=pvalthre))
		    {		      
		      fprintf(fout,"%d%c%d%c%d%c%f%c%.4e%c%f\n",m+1,0x09,n+1,0x09,funcmode,0x09,*(retvec+0),0x09,*(retvec+1),0x09,val);
		      wrote=1;
		      		      
		    }
		  free(retvec);
		}
	    }
	}	
    }

  
  fclose(fout);

  
  // If nothing is written on fout, then delete the file.
  if (wrote==0)
    {
      char *commandstr;
      commandstr=malloc(sizeof(char)*MAXLINEWIDTH);
      sprintf(commandstr,"rm %s\0",filename);
      system(commandstr);
      free(commandstr);
    }
  

  return 0;
}


