// Utility functions of matrix operations.
// All the functions are extracted from numerical recipes.


# include <stdio.h>
# include <stdlib.h>
# include <string.h>
# include <math.h>
# include <time.h>
# include "motif.h"
# include "models.h"
# define MAXLINEWIDTH 100000
# define NAMELENGTH 100
# define LONGNAMELENGTH 100
# define max(i,j) (i>j ? i : j)
# define min(i,j) (i<j ? i : j)
# define argmax(i,j) (j>i ? 2 : 1)
#define ITMAX 100 // Maximum allowed number of iterations.
#define EPS 3.0e-7 // Relative accuracy.
#define FPMIN 1.0e-30 // Number near the smallest representable

double * kron(int nrows1, int ncols1, double *mat1, int nrows2, int ncols2, double *mat2);
void padm(int ndim, double *A, double t, int p, double *E);
void mat_multiply(int ndim1, int ndim2, int ndim3, double *A1, double *A2, double *A3);
void mat_inv(int ndim, double *A1, double *A2);
void ludcmp(int ndim, double *A1, int *indx, int *d, double *A2);
void lubksb(int ndim, double *A, int *indx, double *b);
double mat_det(int ndim, double *A);
double gammln(double xx);
void gcf(double *gammcf, double a, double x, double *gln);
void gser(double *gamser, double a, double x, double *gln);
double gammp(double a, double x);
double chi2cdf(double val, int deg);



// Calculate the Kronecker tensor product of two matrices.
double * kron(int nrows1, int ncols1, double *mat1, int nrows2, int ncols2, double *mat2)
{
  int i, j, k, l, m, n, nrows, ncols;
  double *mat;

  nrows=nrows1*nrows2; ncols=ncols1*ncols2;
  mat=malloc(sizeof(double)*nrows*ncols);
  for (i=0; i<=(nrows1-1); i++)
    {
      for (j=0; j<=(ncols1-1); j++)
	{
	  for (k=0; k<=(nrows2-1); k++)
	    {
	      for (l=0; l<=(ncols2-1); l++)
		{
		  m=i*nrows2+k; n=j*ncols2+l;
		  *(mat+m*ncols+n)=*(mat1+i*ncols1+j)*(*(mat2+k*ncols2+l));
		}
	    }
	}
    }
 
  return mat;
}


// Pade approximation of matrix exponential.
// Inputs: a square matrix, scaling constant t, approximation degree.
// Output: a square matrix.
void padm(int ndim, double *A, double t, int p, double *E)
{
  int i, j, k, l, m, n, odd;
  double *temp, *temp2, *tempA, *c, s, val, val2, *A2, *P, *Q;

  // If t=0, then return the identifty matrix.
  if (fabs(t)<1e-50)
    {
      for (i=0; i<=(ndim*ndim-1); i++)
	*(E+i)=0;
      for (i=1; i<=ndim; i++)
	*(E+(i-1)*ndim+(i-1))=1.0;
      return;
    }

  // Pade coefficients.
  c=malloc(sizeof(double)*(p+2));
  *c=0; *(c+1)=1;
  for (k=1; k<=p; k++)
    *(c+k+1)=*(c+k)*((double)(p+1-k)/(double)(k*(2*p+1-k)));
  
  // Scaling.
  tempA=malloc(sizeof(double)*ndim*ndim);
  
  s=0;
  for (i=1; i<=ndim; i++)
    {
      val=0;
      for (j=1; j<=ndim; j++)
	{
	  val2=*(A+(i-1)*ndim+(j-1)); val2=fabs(val2);
	  val+=val2;
	}
      if (val>s)
	s=val;
    }
  
  if (s>0.5)
    {
      val=log(s)/log(2);
      if (val>=0)
	k=(int)val;
      else
	k=(-1)*((int)((-1)*val));
      k=max(0,k+2);
      val=(-1)*k*log(2);
      val=exp(val);
      for (i=0; i<=(ndim*ndim-1); i++)
	*(tempA+i)=*(A+i)*val*t;
      s=k;
    }
  else
    for (i=0; i<=(ndim*ndim-1); i++)
      *(tempA+i)=*(A+i)*t;

  // Horner evaluation of the irreducible fraction (see ref. above).  
  A2=malloc(sizeof(double)*ndim*ndim);
  mat_multiply(ndim,ndim,ndim,tempA,tempA,A2);
  Q=malloc(sizeof(double)*ndim*ndim);
  P=malloc(sizeof(double)*ndim*ndim);
  for (i=0; i<=(ndim*ndim-1); i++)
    {
      *(Q+i)=0; *(P+i)=0;
    }
  for (i=1; i<=ndim; i++)
    {
      *(Q+(i-1)*ndim+(i-1))=*(c+p+1);
      *(P+(i-1)*ndim+(i-1))=*(c+p);
    }
  odd=1; temp=malloc(sizeof(double)*ndim*ndim);
  temp2=malloc(sizeof(double)*ndim*ndim); 
  for (k=(p-1); k>=1; k--)
    {
      if (odd==1)
	{
	  mat_multiply(ndim,ndim,ndim,Q,A2,temp);
	  for (i=1; i<=ndim; i++)
	    *(temp+(i-1)*ndim+(i-1))+=*(c+k);
	  for (i=0; i<=(ndim*ndim-1); i++)
	    *(Q+i)=*(temp+i);
	}
      else
	{
	  mat_multiply(ndim,ndim,ndim,P,A2,temp);
	  for (i=1; i<=ndim; i++)
	    *(temp+(i-1)*ndim+(i-1))+=*(c+k);
	  for (i=0; i<=(ndim*ndim-1); i++)
	    *(P+i)=*(temp+i);
	}
      odd=1-odd;
    }

  if (odd==1)
    {
      mat_multiply(ndim,ndim,ndim,Q,tempA,temp);
      for (i=0; i<=(ndim*ndim-1); i++)
	*(Q+i)=*(temp+i)-*(P+i);
      mat_inv(ndim,Q,temp2);
      mat_multiply(ndim,ndim,ndim,temp2,P,temp);
      for (i=0; i<=(ndim*ndim-1); i++)
	*(E+i)=*(temp+i)*(-2);
      for (i=1; i<=ndim; i++)
	*(E+(i-1)*ndim+(i-1))-=1;
    }
  else
    {
      mat_multiply(ndim,ndim,ndim,P,tempA,temp);
      for (i=0; i<=(ndim*ndim-1); i++)
	{
	  *(P+i)=*(temp+i);
	  *(Q+i)=*(Q+i)-*(P+i);
	}
      mat_inv(ndim,Q,temp2);
      mat_multiply(ndim,ndim,ndim,temp2,P,temp);
      for (i=0; i<=(ndim*ndim-1); i++)
	*(E+i)=*(temp+i)*2;
      for (i=1; i<=ndim; i++)
	*(E+(i-1)*ndim+(i-1))+=1;
    }

  // Squaring.
  i=(int)s;
  for (k=1; k<=i; k++)
    {
      mat_multiply(ndim,ndim,ndim,E,E,temp);
      for (j=0; j<=(ndim*ndim-1); j++)
	*(E+j)=*(temp+j);
    }

  free(temp); free(temp2); free(tempA); free(c); free(A2); 
  free(P); free(Q);
}


// Matrix multiplication.  A3=A1*A2.  
// A1: ndim1 x ndim2, A2: ndim2 x ndim3, A3: ndim1 x ndim3.
void mat_multiply(int ndim1, int ndim2, int ndim3, double *A1, double *A2, double *A3)
{
  int i, j, k;
  double val;
  for (i=1; i<=ndim1; i++)
    {
      for (j=1; j<=ndim3; j++)
	{
	  val=0;
	  for (k=1; k<=ndim2; k++)
	    val+=*(A1+(i-1)*ndim2+(k-1))*(*(A2+(k-1)*ndim3+(j-1)));
	  *(A3+(i-1)*ndim3+(j-1))=val;
	}
    }
}

// Square matrix inversion.  Use the C codes in numerical recipe.
// Inputs: matrix dimension ndim, matrix A1.
// Output: inverted matrix A2.
void mat_inv(int ndim, double *A1, double *A2)
{
  int i, j, k, *indx, d;
  double *temp, *cols;
  
  indx=malloc(sizeof(int)*ndim);
  temp=malloc(sizeof(double)*ndim*ndim);
  cols=malloc(sizeof(double)*ndim);

  // Perform LU decomposition.
  ludcmp(ndim,A1,indx,&d,temp);

  // Solve the linear system of equations.
  for (j=1; j<=ndim; j++)
    {
      for (i=1; i<=ndim; i++)
	*(cols+i-1)=0;
      *(cols+j-1)=1.0;
      lubksb(ndim,temp,indx,cols);
      for (i=1; i<=ndim; i++)
	*(A2+(i-1)*ndim+(j-1))=*(cols+i-1);
    }
  free(indx); free(temp); free(cols);
}


// LU decomposition of a square matrix.  Use the C codes in numerical recipe.
// Inputs: matrix dimension ndim, matrix A1, row permutation affected by partial pivoting indx, eveness/oddnedd of row interchange.
void ludcmp(int ndim, double *A1, int *indx, int *d, double *A2)
{
  int i, j, k, imax=0;
  double big=0.0, dum, sum, temp, *vv;

  for (i=0; i<=(ndim*ndim-1); i++)
    *(A2+i)=*(A1+i);

  vv=malloc(sizeof(double)*ndim); *d=1;
  for (i=1; i<=ndim; i++)
    *(vv+i-1)=i;
  
  for (i=1; i<=ndim; i++)
    {
      big=0.0;
      for (j=1; j<=ndim; j++)
	if ((temp=fabs(*(A2+(i-1)*ndim+(j-1))))>big)
	  big=temp;
      *(vv+i-1)=1.0/big;
    }
  
  for (j=1; j<=ndim; j++)
    {      
      for (i=1; i<j; i++)
	{
	  sum=*(A2+(i-1)*ndim+(j-1));
	  for (k=1; k<i; k++)
	    sum-=*(A2+(i-1)*ndim+(k-1))*(*(A2+(k-1)*ndim+(j-1)));
	  *(A2+(i-1)*ndim+(j-1))=sum;
	}

      big=0.0;
      for (i=j; i<=ndim; i++)
	{
	  sum=*(A2+(i-1)*ndim+(j-1));
	  for (k=1; k<j; k++)
	    sum-=*(A2+(i-1)*ndim+(k-1))*(*(A2+(k-1)*ndim+(j-1)));
	  *(A2+(i-1)*ndim+(j-1))=sum;
	  if ((dum=*(vv+i-1)*fabs(sum))>=big)
	    {
	      big=dum;
	      imax=i;
	    }
	}
     
      if (j!=imax)
	{
	  for (k=1; k<=ndim; k++)
	    {
	      dum=*(A2+(imax-1)*ndim+(k-1));
	      *(A2+(imax-1)*ndim+(k-1))=*(A2+(j-1)*ndim+(k-1));
	      *(A2+(j-1)*ndim+(k-1))=dum;
	    }
	  *d=-(*d);

	  *(vv+imax-1)=*(vv+j-1);	  

	}

      *(indx+j-1)=imax;
      if (*(A2+(j-1)*ndim+(j-1))==0.0)
	*(A2+(j-1)*ndim+(j-1))=1e-15;

      if (j!=ndim)
	{	  
	  dum=*(A2+(j-1)*ndim+(j-1)); dum=1.0/dum;
	  for (i=(j+1); i<=ndim; i++)
	    *(A2+(i-1)*ndim+(j-1))*=dum;
	}
    }

  free(vv);
}


// Perform forward and backward substitutions to solve the equation Ax=b.
// Inputs: A, b, ndim, indx.
// Output: x replaces b.
void lubksb(int ndim, double *A, int *indx, double *b)
{
  int i, ii=0, ip, j;
  double sum, val;

  for (i=1; i<=ndim; i++)
    {
      ip=*(indx+i-1);
      sum=*(b+ip-1);
      *(b+ip-1)=*(b+i-1);
      if (ii)
	for (j=ii; j<=(i-1); j++)
	  sum-=*(A+(i-1)*ndim+(j-1))*(*(b+j-1));
      else if (sum)
	ii=i;
      *(b+i-1)=sum;
    }

  for (i=ndim; i>=1; i--)
    {
      sum=*(b+i-1);
      for (j=(i+1); j<=ndim; j++)
	sum-=*(A+(i-1)*ndim+(j-1))*(*(b+j-1));
      val=*(A+(i-1)*ndim+(i-1));
      *(b+i-1)=sum/val;
    }
}

  
// Evaluate the determinant of a squate matrix.
// Given the LU decomposition of a matrix, the determinant is the product of diagonal elements times the sign from the LU decomposition.
double mat_det(int ndim, double *A)
{
  int i, j, k, *indx, d;
  double *temp, val;
  
  if (ndim<1)
    return 0;
  else if (ndim==1)
    return *A;

  k=0;
  for (i=0; i<=(ndim-1); i++)
    for (j=0; j<=(ndim-1); j++)
      if (fabs(*(A+i*ndim+j))>=EPS)
	k=1;
  if (k==0)
    return 0;

  indx=malloc(sizeof(int)*ndim);
  temp=malloc(sizeof(double)*ndim*ndim);
  
  // Perform LU decomposition of A.
  ludcmp(ndim,A,indx,&d,temp);

  // Calculate the determinent.
  val=d;
  for (i=0; i<=(ndim-1); i++)
    val*=*(temp+i*ndim+i);
  
  free(indx); free(temp);
  return val;
}


// log Gamma function of a real number.
// Extracted numerical recipes in C.
double gammln(double xx)
{
  double x, y, tmp, ser, retval;
  static double cof[6]={76.18009172947146,-86.50532032941677,24.01409824083091,-1.231739572450155,0.1208650973866179e-2,-0.5395239384953e-5};
  int j;
  y=x=xx;
  tmp=x+5.5;
  tmp -= (x+0.5)*log(tmp);
  ser=1.000000000190015;
  for (j=0;j<=5;j++) ser += cof[j]/++y;
  retval=-tmp+log(2.5066282746310005*ser/x);
  return retval;
}


// Returns the incomplete gamma function Q(a, x) evaluated by its continued fraction representation as gammcf. Also returns lnÎ(a) as gln.
void gcf(double *gammcf, double a, double x, double *gln)
{
  int i;
  double an,b,c,d,del,h;
  
  *gln=gammln(a);
  b=x+1.0-a; // Set up for evaluating continued fraction by modified Lentzâs method (Â§5.2) with b0 = 0.
  c=1.0/FPMIN;
  d=1.0/b;
  h=d;
  for (i=1;i<=ITMAX;i++) 
    { 
      // Iterate to convergence.
      an = -i*(i-a);
      b += 2.0;
      d=an*d+b;
      if (fabs(d) < FPMIN) d=FPMIN;
      c=b+an/c;
      if (fabs(c) < FPMIN) c=FPMIN;
      d=1.0/d;
      del=d*c;
      h *= del;
      if (fabs(del-1.0) < EPS) break;
    }
  if (i > ITMAX) 
    {
      printf("a too large, ITMAX too small in gcf");
      return;
    }
  
  *gammcf=exp(-x+a*log(x)-(*gln))*h; //Put factors in front.
  

  return;
}


// Returns the incomplete gamma function P(a, x) evaluated by its series representation as gamser.
// Also returns ln Î(a) as gln.
void gser(double *gamser, double a, double x, double *gln)
{
  int n;
  double sum,del,ap;
  *gln=gammln(a);
  if (x <= 0.0)
    {
      if (x < 0.0) 
	{
	  printf("x less than 0 in routine gser");
	  return;
	}
      *gamser=0.0;
      return;
    } 
  else 
    {
      ap=a;
      del=sum=1.0/a;
      for (n=1;n<=ITMAX;n++) 
	{
	  ++ap;
	  del *= x/ap;
	  sum += del;
	  if (fabs(del) < fabs(sum)*EPS) 
	    {
	      *gamser=sum*exp(-x+a*log(x)-(*gln));
	      return;
	    }
	}
      printf("a too large, ITMAX too small in routine gser");
      return;
    }
}


// Incomplete gamma function.
// Returns the incomplete gamma function P(a, x).
double gammp(double a, double x)
{
  double gamser, gammcf, gln;
  if (x < 0.0 || a <= 0.0) 
    {
      printf("Invalid arguments in routine gammp");
      return -1;
    }
  if (x < (a+1.0)) 
    { 
      // Use the series representation.
      gser(&gamser,a,x,&gln);
      return gamser;
    } 
  else 
    { 
      // Use the continued fraction representation, and take its complement.
      gcf(&gammcf,a,x,&gln);      
      return 1.0-gammcf;       
    }
}


// chi-square CDF function.
double chi2cdf(double val, int deg)
{
  double v1, v2, v3;
  v1=(double)deg*0.5;
  v2=val/2;
  v3=gammp(v1,v2);
  return v3;
}

