% Process and merge the probe data.

% Inputs are level 2 data of mRNA, CNV, methylation, mutation and mir expression.
% Each type of data typically consists of multiple datasets.
% Outputs are the processed and merged data of each type.
% Each type contains one data file.  All the data files contain the same set of samples.
% The probe values of microarrays are normalized into cumulative distribution function values in [0,1].
% The values from multiple datasets and multiple probes are combined.
% Partition each chromosome into multiple segments and report the segment CNV values.



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Load and process mRNA expression data.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Affy arrays.
filename='mrna_affy_bigTable.cleaned';
[nmrnaprobes1,nmrnasamples1,mrnaprobes1,mrnasamples1,mrnadata1]=load_highRes_data(filename);

% Sort the probes and data by chromosomal coordinates.
ngenes=nmrnaprobes1; genes=mrnaprobes1; data=mrnadata1;
coords=zeros(ngenes,2);
for i=1:ngenes
coords(i,:)=[genes{i}.chr genes{i}.start];
end
sinds=zeros(1,ngenes);
[Y1,I1]=sort(coords(:,1));
for i=0:24
ss=find(coords(:,1)==i);
ss2=coords(ss,2);
[Y2,I2]=sort(ss2);
cc=find(Y1==i);
sinds(cc)=ss(I2);
end
clear sgenes tmpdata;
for i=1:ngenes
j=sinds(i);
sgenes{i}=genes{j};
end
tmpdata(1:ngenes,:)=data(sinds,:);
genes=sgenes; clear sgenes;
mrnaprobes1=genes; mrnadata1=tmpdata; clear ngenes genes data tmpdata;

% Agilent arrays.
filename='mrna_agl_bigTable.cleaned';
[nmrnaprobes2,nmrnasamples2,mrnaprobes2,mrnasamples2,mrnadata2]=load_highRes_data(filename);

% Sort the probes and data by chromosomal coordinates.
ngenes=nmrnaprobes2; genes=mrnaprobes2; data=mrnadata2;
coords=zeros(ngenes,2);
for i=1:ngenes
coords(i,:)=[genes{i}.chr genes{i}.start];
end
sinds=zeros(1,ngenes);
[Y1,I1]=sort(coords(:,1));
for i=0:24
ss=find(coords(:,1)==i);
ss2=coords(ss,2);
[Y2,I2]=sort(ss2);
cc=find(Y1==i);
sinds(cc)=ss(I2);
end
clear sgenes tmpdata;
for i=1:ngenes
j=sinds(i);
sgenes{i}=genes{j};
end
tmpdata(1:ngenes,:)=data(sinds,:);
genes=sgenes; clear sgenes;
mrnaprobes2=genes; mrnadata2=tmpdata; clear ngenes genes data tmpdata;

% Convert the raw data into cdf values, and merge the probe data into gene data.
mrnacdfs1=get_cdf_values2_version2(mrnadata1);
[nmrnagenes1,mrnagenes1,mergedmrnadata1]=merge_probes_singlegenes2(nmrnaprobes1,mrnaprobes1,nmrnasamples1,mrnasamples1,mrnacdfs1);

mrnacdfs2=get_cdf_values2_version2(mrnadata2);
[nmrnagenes2,mrnagenes2,mergedmrnadata2]=merge_probes_singlegenes2(nmrnaprobes2,mrnaprobes2,nmrnasamples2,mrnasamples2,mrnacdfs2);

% Merge the Affy and Agilent datasets.
clear datasets; ndatasets=2;
datasets{1}.ngenes=nmrnagenes1;
datasets{1}.genes=mrnagenes1;
datasets{1}.nsamples=nmrnasamples1;
datasets{1}.samples=mrnasamples1;
datasets{1}.data=mergedmrnadata1;
datasets{2}.ngenes=nmrnagenes2;
datasets{2}.genes=mrnagenes2;
datasets{2}.nsamples=nmrnasamples2;
datasets{2}.samples=mrnasamples2;
datasets{2}.data=mergedmrnadata2;
[nmrnagenes,mrnagenes,mrnadata]=merge_datasets(ndatasets,datasets);
nmrnasamples=nmrnasamples1; mrnasamples=mrnasamples1;



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Rescale CNV data such that the global cdf values approach the distributions of the CGH data.
% Do not need to re-process the CNV data during analysis.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Load and process CNV data.  Merge the probe data into gene data.

% HMS arrays.
filename='cnv_hms_bigTable.cleaned2';
[ncnvprobes1,ncnvsamples1,cnvprobes1,cnvsamples1,cnvdata1]=load_highRes_data(filename);

% Sort the probes and data by chromosomal coordinates.
ngenes=ncnvprobes1; genes=cnvprobes1; data=cnvdata1;
coords=zeros(ngenes,2);
for i=1:ngenes
coords(i,:)=[genes{i}.chr genes{i}.start];
end
sinds=zeros(1,ngenes);
[Y1,I1]=sort(coords(:,1));
for i=0:24
ss=find(coords(:,1)==i);
ss2=coords(ss,2);
[Y2,I2]=sort(ss2);
cc=find(Y1==i);
sinds(cc)=ss(I2);
end
clear sgenes tmpdata;
for i=1:ngenes
j=sinds(i);
sgenes{i}=genes{j};
end
tmpdata(1:ngenes,:)=data(sinds,:);
genes=sgenes; clear sgenes;
cnvprobes1=genes; cnvdata1=tmpdata; clear ngenes genes data tmpdata;

% Calculate the ga coefficient of the quantization function.
% Count the fraction of amplified and deleted entries among all valid entries.
% \int_{0}^{1} (x^ga)/(1-log(x^ga)) dx ~ 0.5 P(entries are amplified or deleted).
% \int_{0}^{1} ((1-x)^ga)/(1-log((1-x)^ga)) dx ~ 0.5 P(entries are amplified or deleted).
% Thresholds; log2ratio>=log(1.5) or log2ratio<=-1.
% Threshold log2 ratio is log2(3/2)=0.5850 and -log(2/2)=-1.
tmp=find((cnvdata1>=0)|(cnvdata1<=0));
tmp=cnvdata1(tmp); tmp=tmp(:);
ratio1=sum((tmp>=log2(3/2))|(tmp<=-1))/length(tmp);
clear tmp;
vals=zeros(1,100);
x=0:0.01:1; 
for i=1:100
y=(x.^i)./(1-log(x.^i));
vals(i)=sum(y)*0.01;
end
ga1=find(abs(vals-ratio1*0.5)<=min(abs(vals-ratio1*0.5)));

% Convert the raw data into cdf values, and merge the probe data into gene data.
% Both CNV datasets are two-channel microarrays.  Thus convert to global cdfs using get_cdf_values_version2.
cnvcdfs1=get_cdf_values_version2(cnvdata1);
[ncnvgenes1,cnvgenes1,mergedcnvdata1]=merge_probes_singlegenes3(ncnvprobes1,cnvprobes1,ncnvsamples1,cnvsamples1,cnvcdfs1,ga1);
cnvchrprobes1=zeros(1,ncnvprobes1);
for n=1:ncnvprobes1
cnvchrprobes1(n)=cnvprobes1{n}.chr;
end


% MSKCC arrays.
filename='cnv_mskc_bigTable.cleaned2';
[ncnvprobes2,ncnvsamples2,cnvprobes2,cnvsamples2,cnvdata2]=load_highRes_data(filename);

% Sort the probes and data by chromosomal coordinates.
ngenes=ncnvprobes2; genes=cnvprobes2; data=cnvdata2;
coords=zeros(ngenes,2);
for i=1:ngenes
coords(i,:)=[genes{i}.chr genes{i}.start];
end
sinds=zeros(1,ngenes);
[Y1,I1]=sort(coords(:,1));
for i=0:24
ss=find(coords(:,1)==i);
ss2=coords(ss,2);
[Y2,I2]=sort(ss2);
cc=find(Y1==i);
sinds(cc)=ss(I2);
end
clear sgenes tmpdata;
for i=1:ngenes
j=sinds(i);
sgenes{i}=genes{j};
end
tmpdata(1:ngenes,:)=data(sinds,:);
genes=sgenes; clear sgenes;
cnvprobes2=genes; cnvdata2=tmpdata; clear ngenes genes data tmpdata;

% Calculate the ga coefficient of the quantization function.
% Count the fraction of amplified and deleted entries among all valid entries.
% \int_{0}^{1} (x^ga)/(1-log(x^ga)) dx ~ 0.5 P(entries are amplified or deleted).
% \int_{0}^{1} ((1-x)^ga)/(1-log((1-x)^ga)) dx ~ 0.5 P(entries are amplified or deleted).
% Thresholds; log2ratio>=log(1.5) or log2ratio<=-1.
% Threshold log2 ratio is log2(3/2)=0.5850 and -log(2/2)=-1.
tmp=find((cnvdata2>=0)|(cnvdata2<=0));
tmp=cnvdata2(tmp); tmp=tmp(:);
ratio2=sum((tmp>=log2(3/2))|(tmp<=-1))/length(tmp);
clear tmp;
vals=zeros(1,100);
x=0:0.01:1; 
for i=1:100
y=(x.^i)./(1-log(x.^i));
vals(i)=sum(y)*0.01;
end
ga2=find(abs(vals-ratio2*0.5)<=min(abs(vals-ratio2*0.5)));

% Convert the raw data into cdf values, and merge the probe data into gene data.
% Both CNV datasets are two-channel microarrays.  Thus convert to global cdfs using get_cdf_values_version2.
cnvcdfs2=get_cdf_values_version2(cnvdata2);
[ncnvgenes2,cnvgenes2,mergedcnvdata2]=merge_probes_singlegenes3(ncnvprobes2,cnvprobes2,ncnvsamples2,cnvsamples2,cnvcdfs2,ga2);
cnvchrprobes2=zeros(1,ncnvprobes2);
for n=1:ncnvprobes2
cnvchrprobes2(n)=cnvprobes2{n}.chr;
end


% Merge the two CNV datasets.
% Since the two merged CNV data are adjusted by rescaling, do not need to do it again.
clear datasets; ndatasets=2;
datasets{1}.ngenes=ncnvgenes1;
datasets{1}.genes=cnvgenes1;
datasets{1}.nsamples=ncnvsamples1;
datasets{1}.samples=cnvsamples1;
datasets{1}.data=mergedcnvdata1;
datasets{2}.ngenes=ncnvgenes2;
datasets{2}.genes=cnvgenes2;
datasets{2}.nsamples=ncnvsamples2;
datasets{2}.samples=cnvsamples2;
datasets{2}.data=mergedcnvdata2;
gas=[1 1];
[ncnvgenes,cnvgenes,cnvdata]=merge_datasets2(ndatasets,datasets,gas);
ncnvsamples=ncnvsamples1; cnvsamples=cnvsamples1;


% Partition CNV data into segments.

% Generate segments from two CNV datasets.
lowressize=1000; numthre=10; valthre=5; fracthre=0.005;
signalthre=0.5; recurthre=1e-10; distthre=10;
[segmentinfo1,smallfrags1]=partition_CNV3(ncnvprobes1,ncnvsamples1,cnvprobes1,cnvsamples1,cnvcdfs1,lowressize,numthre,valthre,ga1,signalthre,recurthre,distthre);
[segmentinfo2,smallfrags2]=partition_CNV3(ncnvprobes2,ncnvsamples2,cnvprobes2,cnvsamples2,cnvcdfs2,lowressize,numthre,valthre,ga2,signalthre,recurthre,distthre);

% Merge the segments generated by two CNV datasets.
nsamples=ncnvsamples;
[nsegments,segments,segmentCNVcdfs]=merge_segmentinfo(nsamples,ncnvprobes1,cnvprobes1,cnvdata1,cnvcdfs1,segmentinfo1,smallfrags1,ga1,ncnvprobes2,cnvprobes2,cnvdata2,cnvcdfs2,segmentinfo2,smallfrags2,ga2,distthre);



% The remaining datasets do not need rescaling.  (1)mRNA and microRNA data are relative, (2)mutations are categorical, (3)methylation data are scaled in [0,1].



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Load and process methylation data.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% oma2.
filename='met_oma2_bigTable.cleaned';
[nmethlprobes1,nmethlsamples1,methlprobes1,methlsamples1,methldata1]=load_highRes_data(filename);

% Sort the probes and data by chromosomal coordinates.
ngenes=nmethlprobes1; genes=methlprobes1; data=methldata1;
coords=zeros(ngenes,2);
for i=1:ngenes
coords(i,:)=[genes{i}.chr genes{i}.start];
end
sinds=zeros(1,ngenes);
[Y1,I1]=sort(coords(:,1));
for i=0:24
ss=find(coords(:,1)==i);
ss2=coords(ss,2);
[Y2,I2]=sort(ss2);
cc=find(Y1==i);
sinds(cc)=ss(I2);
end
clear sgenes tmpdata;
for i=1:ngenes
j=sinds(i);
sgenes{i}=genes{j};
%tmpdata(i,:)=data(j,:);
end
tmpdata(1:ngenes,:)=data(sinds,:);
genes=sgenes; clear sgenes;
methlprobes1=genes; methldata1=tmpdata; clear ngenes genes data tmpdata;

% Convert the raw data into cdf values, and merge the probe data into gene data.
% Methylation data do not need normalization.
methlcdfs1=methldata1;
[nmethlgenes1,methlgenes1,mergedmethldata1]=merge_probes_singlegenes2(nmethlprobes1,methlprobes1,nmethlsamples1,methlsamples1,methlcdfs1);

% oma3.
filename='met_oma3_bigTable.cleaned';
[nmethlprobes2,nmethlsamples2,methlprobes2,methlsamples2,methldata2]=load_highRes_data(filename);

% Sort the probes and data by chromosomal coordinates.
ngenes=nmethlprobes2; genes=methlprobes2; data=methldata2;
coords=zeros(ngenes,2);
for i=1:ngenes
coords(i,:)=[genes{i}.chr genes{i}.start];
end
sinds=zeros(1,ngenes);
[Y1,I1]=sort(coords(:,1));
for i=0:24
ss=find(coords(:,1)==i);
ss2=coords(ss,2);
[Y2,I2]=sort(ss2);
cc=find(Y1==i);
sinds(cc)=ss(I2);
end
clear sgenes tmpdata;
for i=1:ngenes
j=sinds(i);
sgenes{i}=genes{j};
end
tmpdata(1:ngenes,:)=data(sinds,:);
genes=sgenes; clear sgenes;
methlprobes2=genes; methldata2=tmpdata; clear ngenes genes data tmpdata;

% Convert the raw data into cdf values, and merge the probe data into gene data.
% Methylation data do not need normalization.
methlcdfs2=methldata2;
[nmethlgenes2,methlgenes2,mergedmethldata2]=merge_probes_singlegenes2(nmethlprobes2,methlprobes2,nmethlsamples2,methlsamples2,methlcdfs2);

% met27.
filename='met_hum_bigTable.cleaned';
[nmethlprobes3,nmethlsamples3,methlprobes3,methlsamples3,methldata3]=load_highRes_data(filename);

% Sort the probes and data by chromosomal coordinates.
ngenes=nmethlprobes3; genes=methlprobes3; data=methldata3;
coords=zeros(ngenes,2);
for i=1:ngenes
coords(i,:)=[genes{i}.chr genes{i}.start];
end
sinds=zeros(1,ngenes);
[Y1,I1]=sort(coords(:,1));
for i=0:24
ss=find(coords(:,1)==i);
ss2=coords(ss,2);
[Y2,I2]=sort(ss2);
cc=find(Y1==i);
sinds(cc)=ss(I2);
end
clear sgenes tmpdata;
for i=1:ngenes
j=sinds(i);
sgenes{i}=genes{j};
end
tmpdata(1:ngenes,:)=data(sinds,:);
genes=sgenes; clear sgenes;
methlprobes3=genes; methldata3=tmpdata; clear ngenes genes data tmpdata;

% Convert the raw data into cdf values, and merge the probe data into gene data.
% Methylation data do not need normalization.
methlcdfs3=methldata3;
[nmethlgenes3,methlgenes3,mergedmethldata3]=merge_probes_singlegenes2(nmethlprobes3,methlprobes3,nmethlsamples3,methlsamples3,methlcdfs3);

% Merge the three methylation datasets.
clear datasets; ndatasets=3;
datasets{1}.ngenes=nmethlgenes1;
datasets{1}.genes=methlgenes1;
datasets{1}.nsamples=nmethlsamples1;
datasets{1}.samples=methlsamples1;
datasets{1}.data=mergedmethldata1;
datasets{2}.ngenes=nmethlgenes2;
datasets{2}.genes=methlgenes2;
datasets{2}.nsamples=nmethlsamples2;
datasets{2}.samples=methlsamples2;
datasets{2}.data=mergedmethldata2;
datasets{3}.ngenes=nmethlgenes3;
datasets{3}.genes=methlgenes3;
datasets{3}.nsamples=nmethlsamples3;
datasets{3}.samples=methlsamples3;
datasets{3}.data=mergedmethldata3;
[nmethlgenes,methlgenes,methldata]=merge_datasets(ndatasets,datasets);
nmethlsamples=nmethlsamples1; methlsamples=methlsamples1;



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Load and process mutation data.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Load the joint table.
filename='mut_bigTable';
[nmutgenes,nmutsamples,mutgenes,mutsamples,mutdata]=load_highRes_data(filename);

% Change the format from probes to genes.
clear template;
for i=1:nmutgenes
template{i}.name=mutgenes{i}.genename;
template{i}.chr=mutgenes{i}.chr;
template{i}.start=mutgenes{i}.start;
template{i}.end=mutgenes{i}.end;
end
mutgenes=template;


% Sort the genes and data by chromosomal coordinates.
ngenes=nmutgenes; genes=mutgenes; data=mutdata;
coords=zeros(ngenes,2);
for i=1:ngenes
coords(i,:)=[genes{i}.chr genes{i}.start];
end
sinds=zeros(1,ngenes);
[Y1,I1]=sort(coords(:,1));
for i=0:24
ss=find(coords(:,1)==i);
ss2=coords(ss,2);
[Y2,I2]=sort(ss2);
cc=find(Y1==i);
sinds(cc)=ss(I2);
end
clear sgenes tmpdata;
for i=1:ngenes
j=sinds(i);
sgenes{i}=genes{j};
end
tmpdata(1:ngenes,:)=data(sinds,:);
genes=sgenes; clear sgenes;
mutgenes=genes; mutdata=tmpdata; clear ngenes genes data tmpdata;



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Load and process microRNA expression data.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

filename='mir_bigTable.cleaned';
[nmirs,nmirsamples,mirs,mirsamples,mirdata]=load_highRes_data(filename);

% Sort mir probes by chromosomal coordinates.
coords=zeros(nmirs,2); data=mirdata;
for i=1:nmirs
coords(i,:)=[mirs{i}.chr mirs{i}.start];
end
sinds=zeros(1,nmirs);
[Y1,I1]=sort(coords(:,1));
for i=0:24
ss=find(coords(:,1)==i);
ss2=coords(ss,2);
[Y2,I2]=sort(ss2);
cc=find(Y1==i);
sinds(cc)=ss(I2);
end
clear smirs tmpdata;
for i=1:nmirs
j=sinds(i);
smirs{i}=mirs{j};
end
tmpdata(1:nmirs,:)=data(sinds,:);
mirs=smirs; clear smirs; mirdata=tmpdata;

% Convert the raw data into cdf values.
mircdfs=get_cdf_values2_version2(mirdata);



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Find the union of mRNA, CNV, methylation and mutation genes.  Convert the data of each type into tables of the same dimensions.
% Sort the genes by coordinates.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

clear datasets; ndatasets=4;
datasets{1}.ngenes=nmrnagenes;
datasets{1}.genes=mrnagenes;
datasets{1}.nsamples=nmrnasamples;
datasets{1}.samples=mrnasamples;
datasets{1}.data=mrnadata;
datasets{2}.ngenes=ncnvgenes;
datasets{2}.genes=cnvgenes;
datasets{2}.nsamples=ncnvsamples;
datasets{2}.samples=cnvsamples;
datasets{2}.data=cnvdata;
datasets{3}.ngenes=nmethlgenes;
datasets{3}.genes=methlgenes;
datasets{3}.nsamples=nmethlsamples;
datasets{3}.samples=methlsamples;
datasets{3}.data=methldata;
datasets{4}.ngenes=nmutgenes;
datasets{4}.genes=mutgenes;
datasets{4}.nsamples=nmutsamples;
datasets{4}.samples=mutsamples;
datasets{4}.data=mutdata;

maxval=0; maxind=0;
for n=1:ndatasets
if (datasets{n}.ngenes>maxval)
maxind=n; maxval=datasets{n}.ngenes;
end
end


% Get genes.
% Mix the gene info of all data types together.
clear mixgenenames mixcoords; nn=0;
for n=1:ndatasets
for i=1:datasets{n}.ngenes
nn=nn+1; mixgenenames{nn}=datasets{n}.genes{i}.name;
mixcoords(nn,:)=[n i datasets{n}.genes{i}.chr datasets{n}.genes{i}.start datasets{n}.genes{i}.end];
end
end

% Sort the gene info by gene names.
[Y,I]=sort(mixgenenames);

% For the same genes on different datasets, choose representative coordinates.
clear repmixgenenames repmixcoords mixrepmap; mm=0;
mixrepmap=zeros(1,nn);

% The first few items have empty gene names.  Treat them differently.
i=1; k=0;
while ((i<=nn)&(k==0))
if (length(mixgenenames{I(i)})==0)
mm=mm+1; repmixgenenames{mm}=mixgenenames{I(i)};
repmixcoords(mm,:)=mixcoords(I(i),:);
mixrepmap(I(i))=mm;
else
k=1;
end
i=i+1;
end
startind=i-1;

for i=startind:nn
str=mixgenenames{I(i)};
if (strcmp(str,repmixgenenames{mm})==1)
mixrepmap(I(i))=mm;
else
mm=mm+1; repmixgenenames{mm}=str;
repmixcoords(mm,:)=mixcoords(I(i),:);
mixrepmap(I(i))=mm;
end
end

% Sort the genes by coordinates.
sinds=zeros(1,mm);
[Y1,I1]=sort(repmixcoords(:,3));
for i=1:24
ss=find(repmixcoords(:,3)==i);
ss2=repmixcoords(ss,4);
[Y2,I2]=sort(ss2);
cc=find(Y1==i);
sinds(cc)=ss(I2);
end

% Get gene names and coordinates.
ngenes=mm; clear genes; 
for n=1:ngenes
k=sinds(n);
genes{n}.name=repmixgenenames{k};
genes{n}.chr=repmixcoords(k,3);
ii=find(mixrepmap==k);
a1=min(mixcoords(ii,4));
a2=max(mixcoords(ii,5));
genes{n}.start=a1; genes{n}.end=a2;
end
nsamples=datasets{1}.nsamples;
samples=datasets{1}.samples;


% Build the joint data according to the sorted genes and samples.
% The tables all have compatible rows and columns of genes and samples.
% Assume the data of each type already conforms with the same order of samples.

jdata=NaN*ones(ngenes,nsamples,ndatasets);
for n=1:ngenes
k=sinds(n);
ii=find(mixrepmap==k);
% Find each type of data corresponding to the target gene.
for i=1:length(ii)
j=ii(i);
n1=mixcoords(j,1); n2=mixcoords(j,2);
jdata(n,:,n1)=datasets{n1}.data(n2,:);
end
end
jmrnadata=jdata(:,:,1);
jcnvdata=jdata(:,:,2);
jmethldata=jdata(:,:,3);
jmutdata=jdata(:,:,4);
clear jdata mixgenenames mixcoords repmixgenenames repmixcoords mixrepmap;

% The last sample contains end-of-line characters.
samples{nsamples}=samples{nsamples}(1:(length(samples{nsamples})-2));


% Check whether each methylation profile contains too many missing entries.
methlvalid=zeros(1,ngenes);
for n=1:ngenes
%if (sum(~((jmethldata(n,:)>=0)|(jmethldata(n,:)<0)))<=(nsamples/3))
if (sum(((jmethldata(n,:)>=0)|(jmethldata(n,:)<0)))>=50)
if ((sum(jmethldata(n,:)>=0.7)>=10)&(sum(jmethldata(n,:)<=0.3)>=10))
methlvalid(n)=1;
end
end
end

% Check whether each mutation profile contains too many missing entries.
mutvalid=zeros(1,ngenes);
for n=1:ngenes
%if (sum(~((jmutdata(n,:)>=0)|(jmutdata(n,:)<0)))<=(nsamples/3))
if (sum(((jmutdata(n,:)>=0)|(jmutdata(n,:)<0)))>=50)
if (sum(jmutdata(n,:)>0)>=3)
mutvalid(n)=1;
end
end
end



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Report the joint data.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Report mRNA expression.
fout=fopen('jmrna_cleaned.csv','w');
fprintf(fout,'gene%c',9);
for i=1:nsamples
fprintf(fout,'%s',samples{i});
if (i==nsamples)
fprintf(fout,'\n');
else
fprintf(fout,'%c',9);
end
end
for n=1:ngenes
fprintf(fout,'%s|%d|%d|%d|%c',genes{n}.name,genes{n}.chr,genes{n}.start,genes{n}.end,9);
for i=1:nsamples
fprintf(fout,'%f',jmrnadata(n,i));
if (i==nsamples)
fprintf(fout,'\n');
else
fprintf(fout,'%c',9);
end
end
end
fclose(fout);

% Report the rescaled segment CNV data.
fout=fopen('jsegcnv_cleaned_rescaled.csv','w');
fprintf(fout,'segment%c',9);
for n=1:nsamples
fprintf(fout,'%s',samples{n});
if (n<nsamples)
fprintf(fout,'%c',9);
end
end
fprintf(fout,'\n');
for n=1:nsegments
fprintf(fout,'%d|%d|%d|%d|%c',n,segments{n}.chr,segments{n}.startcoord,segments{n}.endcoord,9);
for i=1:nsamples
fprintf(fout,'%f',segmentCNVcdfs(n,i));
if (i<nsamples)
fprintf(fout,'%c',9);
end
end
fprintf(fout,'\n');
end
fclose(fout);

% Report the valid methyldation data.  Use them to generate clusters.
fout=fopen('jmethl_cleaned_valid.csv','w');
fprintf(fout,'gene%c',9);
for i=1:nsamples
fprintf(fout,'%s',samples{i});
if (i==nsamples)
fprintf(fout,'\n');
else
fprintf(fout,'%c',9);
end
end
for n=1:ngenes
if (methlvalid(n)==1)
fprintf(fout,'%s|%d|%d|%d%c',genes{n}.name,genes{n}.chr,genes{n}.start,genes{n}.end,9);
for i=1:nsamples
fprintf(fout,'%f',jmethldata(n,i));
if (i==nsamples)
fprintf(fout,'\n');
else
fprintf(fout,'%c',9);
end
end
end
end
fclose(fout);


% Execute the following C programs to generate clusters of DNA methylation data.
% The cluster data are required for the procedures below.

./generate_cliques_clusters2 jmethl_cleaned_valid.csv 0.4 jmethl_cleaned_cliques.txt 0.1 jmethl_cleaned_clusters.txt

% Load methylation cluster data and report the mean profiles.
[nmethclus,methclus]=load_clusters('jmethl_cleaned_clusters.txt','cluster');
aa=find(methlvalid==1); nmethlvalidgenes=length(aa);
for n=1:nmethclus
methclus{n}=aa(methclus{n});
end

% Modify the cluster files such that singletons are also reported.
fp=fopen('jmethl_cleaned_clusters2.txt','w');
%zz=zeros(1,nmethlvalidgenes);
zz=zeros(1,ngenes);
for n=1:nmethclus
zz(methclus{n})=1;
end
for n=1:nmethclus
fprintf(fp,'cluster %d has %d genes\n',n,length(methclus{n}));
for i=1:length(methclus{n})
l=methclus{n}(i); j=find(aa==l);
fprintf(fp,'%d %s|%d|%d|%d\n',j,genes{l}.name,genes{l}.chr,genes{l}.start,genes{l}.end);
end
fprintf(fp,'\n');
end
subset=find((zz==0)&(methlvalid==1));
for i=1:length(subset)
l=subset(i); j=find(aa==l);
fprintf(fp,'cluster %d has 1 gene\n',nmethclus+i);
fprintf(fp,'%d %s|%d|%d|%d\n\n',j,genes{l}.name,genes{l}.chr,genes{l}.start,genes{l}.end);
end
fclose(fp);

% Load updated methylation clusters.
[nmethclus,methclus]=load_clusters('jmethl_cleaned_clusters2.txt','cluster');
aa=find(methlvalid==1); nmethlvalidgenes=length(aa);
for n=1:nmethclus
methclus{n}=aa(methclus{n});
end

% Report the methyldation cluster data.
fout=fopen('jmethl_cleaned_cludata.csv','w');
fprintf(fout,'gene%c',9);
for i=1:nsamples
fprintf(fout,'%s',samples{i});
if (i==nsamples)
fprintf(fout,'\n');
else
fprintf(fout,'%c',9);
end
end
for m=1:nmethclus
tmp=jmethldata(methclus{m},:);
if (length(tmp(:,1))>1)

tmp2=NaN*ones(1,nsamples);
for l=1:nsamples
vv=tmp(:,l);
val=mean(vv(find((vv>=0)&(vv<=1))));
tmp2(l)=val;
end
tmp=tmp2;

end

% Choose the first element of the cluster as a representative.
n=methclus{m}(1);
fprintf(fout,'%s|%d|%d|%d%c',genes{n}.name,genes{n}.chr,genes{n}.start,genes{n}.end,9);
for i=1:nsamples
fprintf(fout,'%f',tmp(i));
if (i==nsamples)
fprintf(fout,'\n');
else
fprintf(fout,'%c',9);
end
end
end
fclose(fout);


% Report the valid mutation data.
fout=fopen('jmut_cleaned_valid.csv','w');
fprintf(fout,'gene%c',9);
for i=1:nsamples
fprintf(fout,'%s',samples{i});
if (i==nsamples)
fprintf(fout,'\n');
else
fprintf(fout,'%c',9);
end
end
for n=1:ngenes
if (mutvalid(n)==1)
fprintf(fout,'%s|%d|%d|%d%c',genes{n}.name,genes{n}.chr,genes{n}.start,genes{n}.end,9);
for i=1:nsamples
fprintf(fout,'%d',jmutdata(n,i));
if (i==nsamples)
fprintf(fout,'\n');
else
fprintf(fout,'%c',9);
end
end
end
end
fclose(fout);


% Report mir expressions.
fout=fopen('jmir_cleaned.csv','w');
fprintf(fout,'mir%c',9);
for i=1:nsamples
fprintf(fout,'%s',samples{i});
if (i==nsamples)
fprintf(fout,'\n');
else
fprintf(fout,'%c',9);
end
end
for n=1:nmirs
fprintf(fout,'%s|%d|%d|%d|%c',mirs{n}.genename,mirs{n}.chr,mirs{n}.start,mirs{n}.end,9);
for i=1:nsamples
fprintf(fout,'%f',mircdfs(n,i));
if (i==nsamples)
fprintf(fout,'\n');
else
fprintf(fout,'%c',9);
end
end
end
fclose(fout);


% Execute the following C program to cluster microRNA expression data.
% The microRNA cluster data are required for the procedures below.

./generate_cliques_clusters2 jmir_cleaned.csv 0.4 jmir_cleaned_cliques.txt 0.1 jmir_cleaned_clusters.txt


% Load mir cluster data and report the mean profiles.
[nmirclus,mirclus]=load_clusters('jmir_cleaned_clusters.txt','cluster');

% Modify the cluster files such that singletons are also reported.
fp=fopen('jmir_cleaned_clusters2.txt','w');
zz=zeros(1,nmirs);
for n=1:nmirclus
zz(mirclus{n})=1;
end
for n=1:nmirclus
fprintf(fp,'cluster %d has %d genes\n',n,length(mirclus{n}));
for i=1:length(mirclus{n})
j=mirclus{n}(i);
fprintf(fp,'%d %s|%d|%d|%d\n',j,mirs{j}.genename,mirs{j}.chr,mirs{j}.start,mirs{j}.end);
end
fprintf(fp,'\n');
end
subset=find(zz==0);
for i=1:length(subset)
j=subset(i);
fprintf(fp,'cluster %d has 1 gene\n',nmirclus+i);
fprintf(fp,'%d %s|%d|%d|%d\n\n',j,mirs{j}.genename,mirs{j}.chr,mirs{j}.start,mirs{j}.end);
end
fclose(fp);

% Load updated mir clusters.
[nmirclus,mirclus]=load_clusters('jmir_cleaned_clusters2.txt','cluster');

% Report the mir cluster data.
fout=fopen('jmir_cleaned_cludata.csv','w');
fprintf(fout,'mir%c',9);
for i=1:nsamples
fprintf(fout,'%s',samples{i});
if (i==nsamples)
fprintf(fout,'\n');
else
fprintf(fout,'%c',9);
end
end
for m=1:nmirclus
tmp=mircdfs(mirclus{m},:);
if (length(tmp(:,1))>1)

tmp2=NaN*ones(1,nsamples);
for l=1:nsamples
vv=tmp(:,l);
val=mean(vv(find((vv>=0)&(vv<=1))));
tmp2(l)=val;
end
tmp=tmp2;

end

% Choose the first element of the cluster as a representative.
n=mirclus{m}(1);
fprintf(fout,'%s|%d|%d|%d%c',mirs{n}.genename,mirs{n}.chr,mirs{n}.start,mirs{n}.end,9);
for i=1:nsamples
fprintf(fout,'%f',tmp(i));
if (i==nsamples)
fprintf(fout,'\n');
else
fprintf(fout,'%c',9);
end
end
end
fclose(fout);


% Report the expanded TF data.

% Import the TF list.
fp=fopen('TFgenes.txt');
flag=1; nTFs=0; clear TFgenes;
while (flag==1)
s=fgets(fp);
if ((length(s)<=0)|(s==-1))
flag=0;
else
nTFs=nTFs+1;
item=getitemval3(s,0,9);
TFgenes{nTFs}=item(1:(length(item)-1));
end
end
fclose(fp);

% Mark the TF genes.
selected=zeros(1,ngenes);
for n=1:nTFs
str=TFgenes{n};
i=1; j=0;
while ((i<=ngenes)&(j==0))
if (strcmp(str,genes{i}.name)==1)
j=i;
end
i=i+1;
end
if (j>0)
selected(j)=1;
end
end

% Discard if the mRNA expression data is not valid.
for n=1:ngenes
if (selected(n)==1)
k=sum((jmrnadata(n,:)>=0)&(jmrnadata(n,:)<=1));
if (k<(nsamples*0.5))
selected(n)=0;
end
end
end
TFvalid=selected;

% TF genes.
aa=find(TFvalid==1); nTFs=length(aa);
clear TFs;
for n=1:nTFs
k=aa(n);
TFs{n}.genename=genes{k}.name;
end


% Report the expanded TFs.
fp=fopen('jexpanded_TF_mrna_cleaned.csv','w');
fprintf(fp,'gene%c',9);
for i=1:nsamples
str=samples{i};
if (i==nsamples)
str=str(1:(length(str)-1));
end
fprintf(fp,'%s',str);
if (i<nsamples)
fprintf(fp,'%c',9);
end
end
fprintf(fp,'\n');
for n=1:ngenes
if (selected(n)==1)
fprintf(fp,'%s|%d|%d|%d|%c',genes{n}.name,genes{n}.chr,genes{n}.start,genes{n}.end,9);
for i=1:nsamples
fprintf(fp,'%f',jmrnadata(n,i));
if (i<nsamples)
fprintf(fp,'%c',9);
end
end
fprintf(fp,'\n');
end
end
fclose(fp);


% Execute the following C programs to generate clusters of TF data.
% The cluster data are required for the procedures below.

./generate_cliques_clusters2 jexpanded_TF_mrna_cleaned.csv 0.4 jexpanded_TF_cleaned_cliques.txt 0.1 jexpanded_TF_cleaned_clusters.txt

% Load methylation cluster data and report the mean profiles.
[nTFclus,TFclus]=load_clusters('jexpanded_TF_cleaned_clusters.txt','cluster');
aa=find(TFvalid==1); nTFvalidgenes=length(aa);
for n=1:nTFclus
TFclus{n}=aa(TFclus{n});
end

% Modify the cluster files such that singletons are also reported.
fp=fopen('jexpanded_TF_cleaned_clusters2.txt','w');
%zz=zeros(1,nTFlvalidgenes);
zz=zeros(1,ngenes);
for n=1:nTFclus
zz(TFclus{n})=1;
end
for n=1:nTFclus
fprintf(fp,'cluster %d has %d genes\n',n,length(TFclus{n}));
for i=1:length(TFclus{n})
l=TFclus{n}(i); j=find(aa==l);
fprintf(fp,'%d %s|%d|%d|%d\n',j,genes{l}.name,genes{l}.chr,genes{l}.start,genes{l}.end);
end
fprintf(fp,'\n');
end
subset=find((zz==0)&(TFvalid==1));
for i=1:length(subset)
l=subset(i); j=find(aa==l);
fprintf(fp,'cluster %d has 1 gene\n',nTFclus+i);
fprintf(fp,'%d %s|%d|%d|%d\n\n',j,genes{l}.name,genes{l}.chr,genes{l}.start,genes{l}.end);
end
fclose(fp);

% Load updated TF clusters.
[nTFclus,TFclus]=load_clusters('jexpanded_TF_cleaned_clusters2.txt','cluster');
aa=find(TFvalid==1); nTFvalidgenes=length(aa);
for n=1:nTFclus
TFclus{n}=aa(TFclus{n});
end

% Report the TF cluster data.
fout=fopen('jexpanded_TF_cleaned_cludata.csv','w');
fprintf(fout,'gene%c',9);
for i=1:nsamples
fprintf(fout,'%s',samples{i});
if (i==nsamples)
fprintf(fout,'\n');
else
fprintf(fout,'%c',9);
end
end
aa=find(TFvalid==1); 
for m=1:nTFclus
tmp=jmrnadata(TFclus{m},:);
if (length(tmp(:,1))>1)

tmp2=NaN*ones(1,nsamples);
for l=1:nsamples
vv=tmp(:,l);
val=mean(vv(find((vv>=0)&(vv<=1))));
tmp2(l)=val;
end
tmp=tmp2;

end

% Choose the first element of the cluster as a representative.
n=TFclus{m}(1);
fprintf(fout,'%s|%d|%d|%d%c',genes{n}.name,genes{n}.chr,genes{n}.start,genes{n}.end,9);
for i=1:nsamples
fprintf(fout,'%f',tmp(i));
if (i==nsamples)
fprintf(fout,'\n');
else
fprintf(fout,'%c',9);
end
end
end
fclose(fout);



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Report selected methylated genes, mirs and TFs for associations.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

numthre=20;

% Methylated genes.
aa=find(methlvalid==1); nmethlvalidgenes=length(aa);
clear methlvalidgenes;
for n=1:nmethlvalidgenes
k=aa(n);
methlvalidgenes{n}.genename=genes{k}.name;
end

methgenecorrs=zeros(nmethlvalidgenes,ngenes);
jmethlvaliddata=jmethldata(find(methlvalid==1),:);
for ind=1:(floor(ngenes/100))
inds=((ind-1)*100+1):((ind-1)*100+100);
vals=new_missing_corrcoef(transpose([jmethlvaliddata;jmrnadata(inds,:)]));
methgenecorrs(:,inds)=vals(1:nmethlvalidgenes,(nmethlvalidgenes+1):length(vals(1,:)));
end
inds=(floor(ngenes/100)*100+1):ngenes;
vals=new_missing_corrcoef(transpose([jmethlvaliddata;jmrnadata(inds,:)]));
methgenecorrs(:,inds)=vals(1:nmethlvalidgenes,(nmethlvalidgenes+1):length(vals(1,:)));

% Select the methylated genes with sufficient negative associations with mRNA expressions.
fp=fopen('selmethls.txt','w');
for n=1:nmethlvalidgenes
k=sum(methgenecorrs(n,:)<=-0.3);
if ((k>=numthre)&(length(methlvalidgenes{n}.genename)>0))
fprintf(fp,'%d\n',n);
end
end
fclose(fp);


% Mirs.
mirgenecorrs=zeros(nmirs,ngenes);
for ind=1:(floor(ngenes/100))
inds=((ind-1)*100+1):((ind-1)*100+100);
vals=new_missing_corrcoef(transpose([mircdfs;jmrnadata(inds,:)]));
mirgenecorrs(:,inds)=vals(1:nmirs,(nmirs+1):length(vals(1,:)));
end
inds=(floor(ngenes/100)*100+1):ngenes;
vals=new_missing_corrcoef(transpose([mircdfs;jmrnadata(inds,:)]));
mirgenecorrs(:,inds)=vals(1:nmirs,(nmirs+1):length(vals(1,:)));

% Select the mirs with sufficient negative associations with mRNA expressions.
fp=fopen('selmirs.txt','w');
for n=1:nmirs
k=sum(mirgenecorrs(n,:)<=-0.3);
if ((k>=numthre)&(length(mirs{n}.genename)>0))
fprintf(fp,'%d\n',n);
end
end
fclose(fp);

% TFs.
jTFdata=jmrnadata(find(TFvalid==1),:);
TFgenecorrs=zeros(nTFs,ngenes);
for ind=1:(floor(ngenes/100))
inds=((ind-1)*100+1):((ind-1)*100+100);
vals=new_missing_corrcoef(transpose([jTFdata;jmrnadata(inds,:)]));
TFgenecorrs(:,inds)=vals(1:nTFs,(nTFs+1):length(vals(1,:)));
end
inds=(floor(ngenes/100)*100+1):ngenes;
vals=new_missing_corrcoef(transpose([jTFdata;jmrnadata(inds,:)]));
TFgenecorrs(:,inds)=vals(1:nTFs,(nTFs+1):length(vals(1,:)));

% Select the TFs with sufficient positive associations with mRNA expressions.
fp=fopen('selTFs.txt','w');
for n=1:nTFs
k=sum(TFgenecorrs(n,:)>=-0.3);
if ((k>=numthre)&(length(TFs{n}.genename)>0))
fprintf(fp,'%d\n',n);
end
end
fclose(fp);

