/* Utility functions for evaluating the p-values of co-occurrence and mutual exclusion.
 */

# include <stdio.h>
# include <stdlib.h>
# include <string.h>
# include <math.h>
# include <time.h>
# include "motif.h"
# define max(i,j) (i>j ? i : j)
# define min(i,j) (i<j ? i : j)
# define argmax(i,j) (j>i ? 2 : 1)

extern char * getitemval2(char *s, int ind, int *length);
extern int inlist(int Nv, int *v, int target);
double logchoose(int n, int k);
double hyper_geom_pval(int N, int K, int n0, int k0);
double inv_hyper_geom_pval(int N, int K, int n0, int k0);
double F(double theta, int n1, int n0);
double co_occur_pval(int ngenes, int nsamples, int ncooccur);
double binomial_pval(int n, int m, double p0);


// log of n choose k function.
double logchoose(int n, int k)
{
  int i;
  double val;

  if ((n<=0)||(k<=0)||(n<=k))
    return 0.0;

  val=0.0;
  for (i=(k+1); i<=n; i++)
    val+=log((double)i);
  for (i=1; i<=(n-k); i++)
    val-=log((double)i);
  return val;
}
  
// Hyper-geometric p-value.
double hyper_geom_pval(int N, int K, int n0, int k0)
{
  int k;
  double val, pval=0;

  if ((N<=0)||(K<0)||(n0<=0)||(k0<0)||(N<K)||(n0<k0))
    return 1.0;

  for (k=k0; k<=min(K,n0); k++)
    {
      val=logchoose(K,k)+logchoose(N-K,n0-k)-logchoose(N,n0);
      pval+=exp(val);      
    }

  return pval;
}

// Inverted hyper-geometric p-value.
double inv_hyper_geom_pval(int N, int K, int n0, int k0)
{
  int k;
  double val, pval=0;

  if ((N<=0)||(K<0)||(n0<=0)||(k0<0)||(N<K)||(n0<k0))
    return 0.0;
  
  for (k=0; k<=k0; k++)
    {
      if ((k<=K)&&((n0-k)<=(N-K)))
	{
	  val=logchoose(K,k)+logchoose(N-K,n0-k)-logchoose(N,n0);
	  pval+=exp(val);	  
	}

    }

  return pval;
}

// Evaluate the function int_{0}^{\theta} x^{n1}(1-x)^{n0} dx.
// F(\theta=1)=n1!n0!/(n1+n0+1)!.
double F(double theta, int n1, int n0)
{
  int i, j;
  double val, retval;

  if (theta<=0)
    return 0;

  if (n1==0)
    {
      val=(n0+1)*log(1-theta); val=exp(val);
      retval=1-val;
      retval/=(n0+1);
      return retval;
    }
  else if (n1<0)
    return 0;

  if (n0==0)
    {
      val=(n1+1)*log(theta); val=exp(val);
      retval=val;
      retval/=(n1+1);
      return retval;
    }
  else if (n0<0)
    return 0;

  if (theta>=1)
    {
      retval=0;
      for (i=1; i<=n0; i++)
	retval+=log((double)i);
      for (i=(n1+1); i<=(n1+n0+1); i++)
	retval-=log((double)i);
      retval=exp(retval);
      return retval;
    }

  retval=0;
  for (i=1; i<=(n0+1); i++)
    {
      val=0;
      for (j=(n0-i+2); j<=n0; j++)
	val+=log((double)j);
      for (j=(n1+1); j<=(n1+i); j++)
	val-=log((double)j);
      val=val+(n1+i)*log(theta)+(n0-i+1)*log(1-theta);
      retval+=exp(val);
    }

  return retval;
}

// Evaluate the p-value of co-occurrences.
// Null model: P(co-occurrence) is uniformly distributed on {P: P(co-occurrence)<=\hat{P}(co-occurrence)}.
// Test statistic: n(co-occurrence)>=empirical n(co-occurrence).
double co_occur_pval(int ngenes, int nsamples, int ncooccur)
{
  int i, j, k, nstates, *cnts, flag, curpos, carry, cnt, nk;
  double emp_p_cooccur, val=0, retval=0, val2=0;

  emp_p_cooccur=(double)ncooccur/(double)nsamples;

  // Debug
  emp_p_cooccur/=2;
  //emp_p_cooccur=0.01;

  if (ncooccur<=0)
    return 1.0;
  
  nstates=1;
  for (i=1; i<=ngenes; i++)
    nstates*=2;

  cnts=malloc(sizeof(int)*nstates);
  for (i=0; i<=(nstates-1); i++)
    *(cnts+i)=0;
  *(cnts+nstates-1)=ncooccur;
  flag=1; curpos=1; retval=0; cnt=0;

 
  retval=0;
  for (nk=ncooccur; nk<=nsamples; nk++)
    {
      val=0;
      for (i=1; i<=nsamples; i++)
	val+=log((double)i);
      for (i=1; i<=nk; i++)
	val-=log((double)i);
      for (i=1; i<=(nsamples-nk); i++)
	val-=log((double)i);
      val+=log((double)(nstates-1));
      if (emp_p_cooccur>=1)
	val2=1;
      else
	{
	  val2=(nstates-1)*log(1-emp_p_cooccur);
	  val2=1-exp(val2);
	}
      val=exp(val)/val2;
      val2=F(emp_p_cooccur,nk,nsamples-nk+nstates-2);
      val*=val2;
      retval+=val;      
    }        
  
  free(cnts);

  return retval;
}


// The pvalue of binomial distribution.
double binomial_pval(int n, int m, double p0)
{
  int i, j, k;
  double pval, val;

  pval=0;
  for (i=m; i<=n; i++)
    {
      val=logchoose(n,i)+i*log(p0)+(n-i)*log(1-p0);
      pval+=exp(val);
    }

  return pval;
}

